"""Module providing calculation functions provided by the user."""
# Import standard modules.
import sys

# Import own modules.
from src.tube_and_wing.empirical.operating_cost_estimation_tu_berlin.general.operatingcostestimationtuberlin \
    import operating_cost_estimation


def method_kerosene(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """Operating cost estimation method according to TU Berlin for kerosene-powered aircraft.

    This function performs the operating cost estimation according to the TU Berlin method for kerosene-powered aircraft
    configurations.
    [Add more information here...]
    The output dictionary 'kerosene_output_dict' contains the results of the cost estimation and is structured according
    to the following scheme:
        kerosene_output_dict = {'design':
                                {'group_name_1': {'parameter_1': value, 'parameter_2': value, ...},
                                 'group_name_2': {'parameter_1': value, 'parameter_2': value, ...},
                                 ...},
                                'study':
                                 {'group_name_1': {'parameter_1': value, 'parameter_2': value, ...},
                                  'group_name_2': {'parameter_1': value, 'parameter_2': value, ...},
                                  ...}}
    Note that if the study mission does not exist in aircraft exchange file, the 'kerosene_output_dict' does only
    contain the 'design' sub-dictionary.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict dict_ac_exchange: Dict containing parameters and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameters and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict kerosene_output_dict: Dictionary containing results from calculation for kerosene-powered aircraft
    """

    # Merge both data dictionaries.
    dict_complete_data = dict_ac_exchange | dict_mod_config
    # Extract general parameters from 'complete_data_dict'.
    dict_general_data = {key: value for key, value in dict_complete_data.items() if
                         not key.endswith('_design') and not key.endswith('_study')}

    """Design mission."""
    # Extract design mission related parameters from 'complete_data_dict' and save in 'design_data' (w/o ending).
    dict_design_data = {key.replace('_design', ''): value for key, value in dict_complete_data.items()
                        if key.endswith('_design')}
    # Iterate over dictionary.
    for key, value in dict_design_data.items():
        # Check, if any value is 'None'.
        if value is None:
            # If any value is 'None', the calculation of the costs is impossible and the program aborted.
            runtime_output.critical('Error: Operating cost estimation not possible due to missing parameters! '
                                    + 'Program aborted.')
            sys.exit(1)
    # Merge design and general data dictionary.
    dict_design = dict_general_data | dict_design_data
    # Calculate costs for design mission.
    runtime_output.print('----------------------------------------------------------')
    runtime_output.print('Operating cost estimation results for design mission.   ')
    runtime_output.print('----------------------------------------------------------')
    kerosene_output_dict_design = operating_cost_estimation(paths_and_names, routing_dict, dict_design, runtime_output)

    """Mission study."""
    # Extract study mission related parameters from 'complete_data_dict' and save in 'study_data' (w/o ending).
    dict_study_data = {key.replace('_study', ''): value for key, value in dict_complete_data.items()
                       if key.endswith('_study')}
    # Check, if study mission exists (Assumption: If any value is 'None', costs cannot be calculated for mission study).
    mission_study_exists = not any(value is None for value in dict_study_data.values())
    if mission_study_exists:
        # Merge dictionaries.
        dict_study = dict_general_data | dict_study_data
        # Calculate costs for mission study.
        runtime_output.print('Operating cost estimation results for study mission.   ')
        runtime_output.print('----------------------------------------------------------')
        kerosene_output_dict_study = operating_cost_estimation(paths_and_names, routing_dict, dict_study,
                                                               runtime_output)
        # Save both dictionaries in 'kerosene_output_dict'.
        kerosene_output_dict = {'design': kerosene_output_dict_design, 'study': kerosene_output_dict_study}
    else:
        runtime_output.warning('Warning: Operating cost estimation of study mission not possible due to missing data. '
                               + 'No operating costs calculated.')
        runtime_output.print('----------------------------------------------------------')
        kerosene_output_dict = {'design': kerosene_output_dict_design}

    return kerosene_output_dict
