/**
 * @file approach_step.cpp
 * @author Gerrit Pauls (gerrit.pauls@tuhh.de)
 * @brief 
 * @version 0.1
 * @date 2024-08-30
 * 
 * @copyright Copyright (c) 2024
 * 
 */
#include "approach_step.h"
#include <standardFiles/functions.h>

/* Constructor for creation of an approach step object and for the settings of the step parameters. */
approach_step::approach_step(const std::string &a_mode, const std::string &a_rating,
                           double a_derate, const std::string &a_config, double an_end_altitude,
                           double a_rate_of_climb_limit, double an_end_CAS, double a_ground_distance, double a_glide_path)
    :
    mode("mode", "Mode during this step"),
    rating("rating", "Sets thrust rating within climb/acceleration segments to Takeoff, Climb, Maximum continuous, Cruise"),
    derate("derate", "Derate during this step", 1., "1", 0., 1.5),
    configuration("configuration", "Configuration of the aircraft during this step"),
    rate_of_climb_limit("rate_of_climb_limit", "Rate of climb during this step (empty value = climb with full thrust)",
                    0., "m/s", 0., convertUnit(FOOTPERMINUTE, METERPERSECOND, 5000.)),
    end_CAS("calibrated_airspeed", "Calibrated airspeed at the end of this step", 0., "m/s", 0., 250.),
    end_altitude("altitude", "Altitude at the end of this step", 0., "m", 0., 20000.),
    ground_distance("ground_distance", "Distance to the ground", 0., "m", 0., 20000),
    glide_path("glide_path", "Angle between glide path and runway", 0., "rad", -0.35, 0.35) { // 0.35 rad = ~20 DEG
    this->derate.set_value(a_derate);
    this->mode.set_value(a_mode);
    this->rating.set_value(a_rating);
    this->configuration.set_value(a_config);
    this->end_altitude.set_value(an_end_altitude);
    this->rate_of_climb_limit.set_value(a_rate_of_climb_limit);
    this->end_CAS.set_value(an_end_CAS);
    this->ground_distance.set_value(a_ground_distance);
    this->glide_path.set_value(a_glide_path);
}

/* Destructor of the approach step object */
approach_step::~approach_step() {
}

void approach_step::write_approach_steps(std::string subpath, node* mission_file) {
    this->configuration.update((*mission_file)[subpath]);
    this->derate.update((*mission_file)[subpath]);
    this->mode.update((*mission_file)[subpath ]);
    this->rating.update((*mission_file)[subpath]);
    /* Catch empty endnodes and update nodes that shall be filled */
    if (!std::isnan(this->rate_of_climb_limit.value())) {
        this->rate_of_climb_limit.update((*mission_file)[subpath]);
    }
    if (!std::isnan(this->end_CAS.value())) {
        this->end_CAS.update((*mission_file)[subpath]);
    }
    if (!std::isnan(this->end_altitude.value())) {
        this->end_altitude.update((*mission_file)[subpath]);
    }
    if (!std::isnan(this->ground_distance.value())) {
        this->ground_distance.update((*mission_file)[subpath]);
    }
    if (!std::isnan(this->glide_path.value())) {
        this->glide_path.update((*mission_file)[subpath]);
    }
}
