/**
 * @file cruise_step.cpp
 * @author Gerrit Pauls (gerrit.pauls@tuhh.de)
 * @brief 
 * @version 0.1
 * @date 2024-08-30
 * 
 * @copyright Copyright (c) 2024
 * 
 */
#include "cruise_step.h"
#include <standardFiles/functions.h>

/* Constructor for creation of an cruise step object and for the settings of the step parameters. */
cruise_step::cruise_step(const std::string &a_mode, const std::string &a_rating,
                       double a_derate, const std::string &a_config, double an_end_altitude,
                       bool auto_select_initial_cruise_altitude, bool round_to_regular_flight_level, double auto_climb_step_altitude_change,
                       bool a_flight_management_system, double a_rate_of_climb_limit, double an_end_CAS,
                       double an_end_Mach, double a_relative_segment_length)
    :
    derate("derate", "Derate during this step", 1., "1", 0., 1.5),
    mode("mode", "Mode during this step"),
    rating("rating", "Sets thrust rating within climb/acceleration segments to Takeoff, Climb, Maximum continuous, Cruise"),
    configuration("configuration", "Configuration of the aircraft during this step"),
    rate_of_climb_limit("rate_of_climb_limit", "Rate of climb during this step (empty value = climb with full thrust)",
                    0., "m/s", 0., convertUnit(FOOTPERMINUTE, METERPERSECOND, 5000.)),
    end_CAS("calibrated_airspeed", "Calibrated airspeed at the end of this step", 0., "m/s", 0., 250.),
    end_Mach("mach_number", "Mach number at the end of this step", 0., "1", 0., 1.),
    end_altitude("altitude", "Altitude at the end of this step", 0., "m", 0., 20000.),
    flight_management_system("enabled",
        "Switch to indicate if a flight management system is equipped"), // called 'enabled' since the switch has a cost_index subnote
    cost_index("cost_index", "Cost index [kg/min], scaled 0 to 999 according to Sperry/Honeywell", 0., "1", 0., 999.),
    relative_segment_length("relative_segment_length", "The end of this cruise step relative to the whole cruise flight phase", 0., "1", 0., 1.),
    round_to_regular_flight_level("round_to_regular_flight_level", "Switch to round the flight level to multiples of 10", false),
    auto_select_optimum_flight_level("enabled",
        "Switch for automatic selection of the optimum flight level of the cruise step", false), // called 'enabled' since the switch has an auto_climb_step_height subnote
    auto_climb_step_height("auto_climb_step_height", "Height difference for an automatic altitude change step.", 0., "m", 0., 5000.) {
    this->derate.set_value(a_derate);
    this->mode.set_value(a_mode);
    this->rating.set_value(a_rating);
    this->configuration.set_value(a_config);
    this->rate_of_climb_limit.set_value(a_rate_of_climb_limit);
    this->end_CAS.set_value(an_end_CAS);
    this->end_Mach.set_value(an_end_Mach);
    this->end_altitude.set_value(an_end_altitude);
    this->flight_management_system.set_value(a_flight_management_system);
    this->relative_segment_length.set_value(a_relative_segment_length);
    if (auto_select_initial_cruise_altitude) {
        this->auto_select_optimum_flight_level.set_value(true);
        this->auto_climb_step_height.set_value(auto_climb_step_altitude_change);
    }
    if (round_to_regular_flight_level) {
        this->round_to_regular_flight_level.set_value(true);
    }
    if (flight_management_system.value()) {
        this->cost_index.set_value(30);
    }
}

/* Destructor of cruise_step object */
cruise_step::~cruise_step() {
}

void cruise_step::write_cruise_steps(std::string subpath, node* mission_file) {
    this->configuration.update((*mission_file)[subpath]);
    this->derate.update((*mission_file)[subpath]);
    this->mode.update((*mission_file)[subpath ]);
    this->rating.update((*mission_file)[subpath]);
    /* Catch empty endnodes and update nodes that shall be filled */
    if (!std::isnan(this->rate_of_climb_limit.value())) {
        this->rate_of_climb_limit.update((*mission_file)[subpath]);
    }
    if (!std::isnan(this->end_CAS.value())) {
        this->end_CAS.update((*mission_file)[subpath]);
    }
    if (!std::isnan(this->end_Mach.value())) {
        this->end_Mach.update((*mission_file)[subpath]);
    }
    if (!std::isnan(end_altitude.value())) {
        this->end_altitude.update((*mission_file)[subpath]);
    }
    if (!std::isnan(this->relative_segment_length.value())) {
        this->relative_segment_length.update((*mission_file)[subpath]);
    }
    this->auto_select_optimum_flight_level.update((*mission_file)[subpath + "/auto_select_optimum_flight_level"]);
    this->auto_climb_step_height.update((*mission_file)[subpath + "/auto_select_optimum_flight_level"]);
    mission_file->at(subpath + "/auto_select_optimum_flight_level").addAttrib("description", "Parameters to handle automatized flight_level changes");
    this->flight_management_system.update((*mission_file)[subpath + "/flight_management_system"]);
    mission_file->at(subpath + "/flight_management_system").addAttrib("description", "Flight management system settings");
    this->cost_index.update((*mission_file)[subpath + "/flight_management_system"]);
    this->round_to_regular_flight_level.update((*mission_file)[subpath]);
}
