/**
 * @file standard_mission.h
 * @author Gerrit Pauls (gerrit.pauls@tuhh.de)
 * @brief 
 * @version 0.1
 * @date 2024-08-30
 * 
 * @copyright Copyright (c) 2024
 * 
 */
#ifndef CREATE_MISSION_XML_SRC_STANDARD_MISSION_STANDARD_MISSION_STANDARD_MISSION_H_
#define CREATE_MISSION_XML_SRC_STANDARD_MISSION_STANDARD_MISSION_STANDARD_MISSION_H_

#include <string>
#include <memory>
#include <unordered_map>
#include <vector>
#include <moduleBasics/report.h>
#include <moduleBasics/strategySelector.h>
#include <moduleBasics/module.h>
#include <moduleBasics/modeSelector.h>
#include <aixml/endnode.h>
#include "../../libs/mission_steps/approach_step.h"
#include "../../libs/mission_steps/cruise_step.h"
#include "../../libs/mission_steps/departure_step.h"

class cruise_step;
class departure_step;
class approach_step;

class standard_mission : public Strategy {
 public:
    explicit standard_mission(const std::shared_ptr<RuntimeIO>& rtIO);
    ~standard_mission() = default;

    void initialize();
    void run();
    void update();
    void report();
    void save();

    const std::shared_ptr<RuntimeIO>& rtIO;

    /** \struct cruise_step_segment
    *   \brief data structure for to define the cruise flight phase
    */
    struct cruise_step_segment {
        double relative_segment_length; /**< Relative segment length of the step [-] */
        double altitude; /**< Altitude of the step [m] */
    };

    /** \class mission_parameters
    *   \brief gathers the attributes each mission needs for itself
    *   \param name: Name of the mission
    *   \param single_PAX_mass: Mass of a single passenger [m]
    */
    class mission_parameters {
     public:
        explicit mission_parameters(const std::string &name, const double single_PAX_mas);
        ~mission_parameters();
        /* Naming */
        std::string mission_type; /**< Name of the mission */
        std::string mission_xml_name; /**< Name of the generated mission XML file */
        /* Boolean Config settings */
        bool write_mission_file; /**< Switch for generating a mission file */
        bool auto_select_initial_cruise_altitude; /**< Switch for selecting an initial cruise altitude */
        bool auto_select_flight_level; /**< Switch for automatic selection of the flight level */
        bool round_to_regular_flight_level; /**< Switch for rounding the flight level */
        bool no_steps; /**< Switch for the number of steps */

        /* Mission values */
        Endnode<double> range; /**< Total range of the aircraft [m] */
        Endnode<double> payload; /**< Payload of the aircraft [kg] */
        Endnode<int> pax_number; /**< Number of the passengers [-] */
        Endnode<double> cargo_mass; /**< Mass of the cargo [kg] */
        Endnode<double> initial_cruise_mach; /**< Initial cruise Mach number [-] */
        Endnode<double> alternate_distance; /**< Range of the alternate mission [NM] */
        Endnode<double> taxi_time_origin; /**< Length of the taxiing time at start [s] */
        Endnode<double> taxi_time_destination; /**< Length of the taxiing time after landing [s] */
        Endnode<std::string> taxiing_procedure; /**< Switch to activate electric taxiing */
        Endnode<double> engine_warmup_time; /**< Engine warm up time before start [s] */
        Endnode<double> terminal_operation_time; /**< Time at terminal for intermediate landings [s] */
        Endnode<std::string> origin_airport; /**< ICAO code of origin airport */
        Endnode<std::string> destination_airport; /**< ICAO code of destination airport */
        double auto_climb_altitude_steps; /**< Altitude step at cruising flight if automatic altitude adaption [m] */
        double auto_rate_of_climb_steps; /**< Rate of climb for altitude step at cruising flight if automatic altitude adaption [m/s] */
        int cruise_step_number; /**< Number of the cruise steps [-] */
        double CAS_ATC_limit_climb; /**< Calibrated airspeed under flight level 100 during climb [m/s] */
        double CAS_ATC_limit_descent; /**< Calibrated airspeed under flight level 100 during descent [m/s] */
        double CAS_over_flight_level_100_climb; /**< Calibrated airspeed over flight level 100 during climb [m/s] */
        double CAS_over_flight_level_100_descent; /**< Calibrated airspeed over flight level 100 during descent [m/s] */
        double mach_climb; /**< Mach number for climb [-] */
        double initial_cruise_altitude; /**< Initial cruise altitude [FL] */
        double delta_temperature_ISA; /**< Temperature deviation from international standard atmosphere [K] */
        double delta_mach_climb_to_mach_cruise; /**< Difference between Mach number over crossover-altitude to cruise Mach number [-] */
        double maximum_rate_of_climb; /**< Limit for ROC in climb segments; if 0, full thrust is used for climb [m/s] */
        double approach_speed; /* Given approach speed [m/s] */
        std::vector<cruise_step_segment> cruise_step_segments; /* Vector of cruise speed segments to specify the cruise phases for a constant flight_level */

        /* Mission steps */
        std::vector<departure_step> departure_steps; /**< Vector for the flight steps on departure */
        std::vector<cruise_step> cruise_steps; /**< Vector for the flight steps on cruising flight */
        std::vector<approach_step> approach_steps; /**< Vector for the flight steps on approach */

        /* Modes */
        uint16_t takeoff_procedure; /* Definition of the takeoff procedure */
        uint16_t approach_procedure; /* Definition of the approach procedure */
        std::string climb_thrust_setting;  /* Definition of the climb procedure */
    };
    mission_parameters* design_mission; /* mission parameters for the design mission */
    mission_parameters* study_mission; /* mission parameters for the study mission */
    mission_parameters* requirements_mission; /* mission parameters for the requirement mission */

    /* Member functions */
    /* Initialize */
    /** \brief Function bundels the setter functions below to initialize the given mission
     * \details Function reads the range and payload depending on the mission type out of the requirements block of the aircraft XML.
     * \param mission: mission_parameters object which shall be filled
     */
    void initialize_mission(standard_mission::mission_parameters* mission);
    /** \brief Function sets the methods
     * \details Function gets settings like the switches for automatic flight_level change or step calculation + names and paths.
     * \param mission: mission_parameters object which shall be filled
     */
    void set_module_settings(standard_mission::mission_parameters* mission);
    /** \brief Function sets the range, payload, number of passengers and the cargo mass of the aircraft
     * \details Function reads the range and payload depending on the mission type out of the requirements block of the aircraft XML.
     * \param mission: mission_parameters object which shall be filled
     */
    void set_range_and_payload(standard_mission::mission_parameters* mission);
    /** \brief Function sets cruise parameters
     * \details Function fills the cruise_step_segments which again will be used for the cruise steps.
     * \param mission: mission_parameters object which shall be filled
     */
    void set_cruise(standard_mission::mission_parameters* mission);
    /** \brief Function sets climb and descent parameters
     * \details Function calculates transition altitudes and calibrated airspeeds for the different climb and descent steps.
     * \param mission: mission_parameters object which shall be filled
     */
    void set_climb_and_descent(standard_mission::mission_parameters* mission);
    /** \brief Function sets start and landing parameters
     * \details Function gathers input for start, landing and further ground operation 
     * \param mission: mission_parameters object which shall be filled
     */
    void set_start_and_landing(standard_mission::mission_parameters* mission);
    /** \brief Function to set performance limits
     * \details Function gathers climb_thrust_setting and maximum_rate_of_climb.
     * \param mission: mission_parameters object which shall be filled
     */
    void set_performance_limits(standard_mission::mission_parameters* mission);
    /** \brief Function to give shell output
     * \details Function displays Range, Payload, Cruise Mach number, Initial cruise altitude, Climb speed under FL100, Climb speed above FL100 and Climb Mach number
     * \param mission: mission_parameters object which shall be filled
     */
    void write_shell_output(standard_mission::mission_parameters* mission);

    /* Mission steps and profiles */
    /** \brief Function bundels the setter functions below to define the mission profiles for the given mission
     * \details set_cruise_profile, set_departure_profile and set_approach_profile are executed.
     * \param mission: mission_parameters object which shall be filled
     */
    void set_mission_profiles(standard_mission::mission_parameters* mission);
    /** \brief Function creates the cruise profile of the mission
     * \details It creates objects for the different cruise steps and saves them in a vector.
     * \param mission: mission_parameters object which shall be filled
     */
    void set_cruise_profile(standard_mission::mission_parameters* mission);
    /** \brief Function creates the departure profile.
     * \details It creates objects for the different departure steps and saves them in a vector.
     *          It checks the take off procedure and cancels the program if procedure in configuration file is unknown or not yet implemented.
     * \param mission: mission_parameters object which shall be filled
     */
    void set_departure_profile(standard_mission::mission_parameters* mission);
    /** \brief Function creates the approach profile of the mission.
     * \details It creates objects for the different approach steps and saves them in a vector.
     *          It checks the approach procedure and cancels the program if procedure in configuration file is unknown or not yet implemented.
     * \param mission: mission_parameters object which shall be filled
     */
    void set_approach_profile(standard_mission::mission_parameters* mission);

    /* Write mission file */
    /** \brief Function sets up global mission data for a given mission
     * \details It opens an aixml sets its first node up via an input stream.
     */
    void setup_mission_file(standard_mission::mission_parameters* mission);

    /** \brief Function fills mission files with the mission's endnodes
     * \details Endnodes for global mission parameters and cruise, approach and departure steps.
     */
    void fill_mission_file(standard_mission::mission_parameters* mission);


 private:
    /* Attributes */
    bool use_delta_max_operating_mach; /**< Switch for using the deviation between maximum and operating Mach number */
    bool adapt_climb_speed; /**< Switch for adapting climb speed schedule */
    bool use_design_mission_mach; /**< Switch for using Mach number in design mission for study mission */
    bool use_design_mission_ICO; /**< Switch for using initial cruise altitude in design mission for study mission */
    double delta_max_operating_mach; /**< Deviation between maximum and operating Mach number [-] */
    double crossover_altitude_specs; /**< Crossover altitude specifications [m] */


    /** \brief Function gets transition altitude
     * \param CAS: calibrated air speed [m/s]
     * \param Mach: Mach number [-]
     * \param atm: reference to standard atmosphere
     * \return atm.getAltitudeAtPressureRatio(pressureRatio) double: Altitude [m]
     */
    inline double get_transition_altitude(double CAS, double Mach, const atmosphere &atm) {
        double h_ref(0.);
        double exponent(ISENTROPIC_EXPONENT_DRY_AIR / (ISENTROPIC_EXPONENT_DRY_AIR - 1)); // gamma / (gamma - 1)
        double pressureRatio = (pow(1 + (ISENTROPIC_EXPONENT_DRY_AIR - 1) / 2. * pow(CAS / atm.getSpeedOfSoundISA(h_ref), 2.), exponent) - 1) /
                            (pow(1 + (ISENTROPIC_EXPONENT_DRY_AIR - 1) / 2. * pow(Mach, 2.), exponent) - 1);
        return atm.getAltitudeAtPressureRatio(pressureRatio);
    }
};

#endif // CREATE_MISSION_XML_SRC_STANDARD_MISSION_STANDARD_MISSION_STANDARD_MISSION_H_
