/**
 * \file schaeferPlot.cpp
 * \author Katrin Bistreck (katrin.bistreck@tuhh.de)
 * \brief Schaefer emissions plot implementations
 * \version 0.1
 * \date 2023-12-11
 *
 * @copyright Copyright (c) 2023
 *
 */

#include <string>
#include "../../../standard_strategy.h"
#include "../../../../toolinfo.h"
#include <moduleBasics/plot.h>

void StandardStrategy::generatePlots_schaefer() {
     std::string pltName(std::string(TOOL_NAME) + "_inflightEmissions");
     this->generateInflightEmissionsPlot_schaefer(pltName);
     pltName = std::string(TOOL_NAME) + "_lifeCycleEmissions";
     this->generateLifeCycleEmissionsPlot_schaefer(pltName);
}

void StandardStrategy::generateInflightEmissionsPlot_schaefer(const std::string& pltName) {
     /* Define data */
     double totalCO2 = theLifeCycleEmissionsSchaefer.theDevelopment.test_total.CO2 + theStudyMission.LTO_Total.CO2 + theStudyMission.Cruise_Total.CO2;
     double totalHC = theLifeCycleEmissionsSchaefer.theDevelopment.test_total.HC + theStudyMission.LTO_Total.HC + theStudyMission.Cruise_Total.HC;
     double totalNOx = theLifeCycleEmissionsSchaefer.theDevelopment.test_total.NOx + theStudyMission.LTO_Total.NOx + theStudyMission.Cruise_Total.NOx;
     double totalSO2 = theLifeCycleEmissionsSchaefer.theDevelopment.test_total.SO2 + theStudyMission.LTO_Total.SO2 + theStudyMission.Cruise_Total.SO2;
     double totalCO = theLifeCycleEmissionsSchaefer.theDevelopment.test_total.CO + theStudyMission.LTO_Total.CO + theStudyMission.Cruise_Total.CO;
     double totalH2O = theLifeCycleEmissionsSchaefer.theDevelopment.test_total.H2O + theStudyMission.LTO_Total.H2O + theStudyMission.Cruise_Total.H2O;
     double totalPM10 = theLifeCycleEmissionsSchaefer.theDevelopment.test_total.PM10 + theStudyMission.LTO_Total.soot + theStudyMission.Cruise_Total.soot;
     std::vector<double> cruise = {
          theStudyMission.Cruise_Total.CO2 / totalCO2 * 100,
          theStudyMission.Cruise_Total.H2O / totalH2O * 100,
          theStudyMission.Cruise_Total.SO2 / totalSO2 * 100,
          theStudyMission.Cruise_Total.HC / totalHC * 100,
          theStudyMission.Cruise_Total.CO / totalCO * 100,
          theStudyMission.Cruise_Total.NOx / totalNOx * 100,
          theStudyMission.Cruise_Total.soot / totalPM10 * 100
     };
     std::vector<double> LTO = {
          theStudyMission.LTO_Total.CO2 / totalCO2 * 100,
          theStudyMission.LTO_Total.H2O / totalH2O * 100,
          theStudyMission.LTO_Total.SO2 / totalSO2 * 100,
          theStudyMission.LTO_Total.HC / totalHC * 100,
          theStudyMission.LTO_Total.CO / totalCO * 100,
          theStudyMission.LTO_Total.NOx / totalNOx * 100,
          theStudyMission.LTO_Total.soot / totalPM10 * 100
     };
     std::vector<double> test = {
          theLifeCycleEmissionsSchaefer.theDevelopment.test_total.CO2 / totalCO2 * 100,
          theLifeCycleEmissionsSchaefer.theDevelopment.test_total.H2O / totalH2O * 100,
          theLifeCycleEmissionsSchaefer.theDevelopment.test_total.SO2 / totalSO2 * 100,
          theLifeCycleEmissionsSchaefer.theDevelopment.test_total.HC / totalHC * 100,
          theLifeCycleEmissionsSchaefer.theDevelopment.test_total.CO / totalCO * 100,
          theLifeCycleEmissionsSchaefer.theDevelopment.test_total.NOx / totalNOx * 100,
          theLifeCycleEmissionsSchaefer.theDevelopment.test_total.soot / totalPM10 * 100
     };
     std::vector<std::string> emissions = {"CO2", "H2O", "SO2", "HC", "CO", "NOx", "soot"};

     /* Plot with matplotplusplus */
     std::vector<std::vector<double>> data;
     if (theLifeCycleEmissionsSchaefer.schaeferIOData_->configuration.developmentEmissionSetting != "mode_0") {
          data = {test, LTO, cruise};
     } else {
          data = {LTO, cruise};
     }

     matplot::figure_handle fig = matplot::figure(true);
     matplot::axes_handle ax = fig->current_axes();
     ax->barstacked(data);
     ax->title("Flight Emissions");
     ax->ylabel("Emissions  [%]");
     ax->ylim({0, 120});
     ax->x_axis().ticklabels(emissions);
     std::vector<std::string> legend_labels;
     if (theLifeCycleEmissionsSchaefer.schaeferIOData_->configuration.developmentEmissionSetting != "mode_0") {
          legend_labels = {"Cruise", "LTO", "Test"}; // \todo check order in legend (definition needs to be in "wrong direction")
     } else {
          legend_labels = {"Cruise", "LTO"};
     }
     matplot::legend_handle lgd = ::matplot::legend(legend_labels);
     lgd->location(matplot::legend::general_alignment::top);
     lgd->num_columns(3);
     std::string plotFilePath = rtIO_->getPlotDir() + "/" + pltName + ".svg";
     matplot::save(plotFilePath);
     matplot::cla();
}

void StandardStrategy::generateLifeCycleEmissionsPlot_schaefer(const std::string& pltName) {
     std::vector<double> development_emissions = {
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.CO2,
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.H2O,
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.NOx,
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.CO,
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.SO2,
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.CH4,
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.PM10,
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.N2O,
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.PFC,
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.NMVOC,
          theLifeCycleEmissionsSchaefer.theDevelopment.development_total.HC
     };
     if (theLifeCycleEmissionsSchaefer.schaeferIOData_->configuration.producedAC != 0) {
          for (uint8_t i = 0; i < development_emissions.size(); i++) {
            development_emissions[i] /= theLifeCycleEmissionsSchaefer.schaeferIOData_->configuration.producedAC; //scale to one aircraft
          }
    } else {
          myRuntimeInfo->warn << "Number of produced aircraft is zero. Full development emissions will be shown in LCA plot" << std::endl;
    }
     std::vector<double> production_emissions = {
          theLifeCycleEmissionsSchaefer.theProduction.production_total.CO2,
          theLifeCycleEmissionsSchaefer.theProduction.production_total.H2O,
          theLifeCycleEmissionsSchaefer.theProduction.production_total.NOx,
          theLifeCycleEmissionsSchaefer.theProduction.production_total.CO,
          theLifeCycleEmissionsSchaefer.theProduction.production_total.SO2,
          theLifeCycleEmissionsSchaefer.theProduction.production_total.CH4,
          theLifeCycleEmissionsSchaefer.theProduction.production_total.PM10,
          theLifeCycleEmissionsSchaefer.theProduction.production_total.N2O,
          theLifeCycleEmissionsSchaefer.theProduction.production_total.PFC,
          theLifeCycleEmissionsSchaefer.theProduction.production_total.NMVOC,
          theLifeCycleEmissionsSchaefer.theProduction.production_total.HC
     };
     std::vector<double> operation_emissions = {
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.CO2,
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.H2O,
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.NOx,
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.CO,
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.SO2,
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.CH4,
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.PM10,
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.N2O,
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.PFC,
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.NMVOC,
          theLifeCycleEmissionsSchaefer.theOperation.operation_total.HC
     };
     std::vector<double> eol_emissions = {
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.CO2,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.H2O,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.NOx,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.CO,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.SO2,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.CH4,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.PM10,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.N2O,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.PFC,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.NMVOC,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLife_total.HC
     };
     std::vector<std::string> emissions = {"CO2", "H2O", "NOx", "CO", "SO2", "CH4", "PM10", "N2O", "PFC", "NMVOC", "HC"};
     matplot::figure_handle fig = matplot::figure(true);
     matplot::axes_handle ax = fig->current_axes();
     std::vector<std::vector<double>> emissions_data = {development_emissions, production_emissions, operation_emissions, eol_emissions};
     ax->bar(emissions_data);
     ax->title("Life Cycle Emissions");
     ax->ylabel("Emissions [kg]");
     ax->x_axis().ticklabels(emissions);
     ax->y_axis().scale(matplot::axis_type::axis_scale::log);
     auto legend = matplot::legend(ax, {"Development", "Production", "Operation", "End of Life"});// \todo Correct legend
     legend->location(matplot::legend::general_alignment::topright);
     std::string plotFilePath = rtIO_->getPlotDir() + "/" + pltName + ".svg";
     matplot::save(plotFilePath);
     matplot::cla();
}
