/**
 * \file AQI_schaefer.cpp
 * \author katrin.bistreck@tuhh.de
 * \brief Calculate air quality index according to Schaefer (2017)
 * \version 0.1
 * \date 2023-12-07
 *
 * @copyright Copyright (c) 2023
 *
 */
#include <iostream>
#include <memory>
#include <algorithm>
#include <string>
#include <moduleBasics/runtimeIO.h>
#include <aixml/node.h>
#include "AQI_schaefer.h"
#include "AQISchaeferIOData.h"

AQISchaefer::AQISchaefer() : Tolerance_CO(0.), Tolerance_HC(0.), Tolerance_NOx(0.), Tolerance_soot(0.), AQI(0.) {}

void AQISchaefer::run(const std::shared_ptr<RuntimeIO> &rtIO, std::shared_ptr<node> ecoAssXML) {
    myRuntimeInfo->out << "  Calculation of air quality index" << std::endl;
    AQISchaeferIOData_ = std::make_shared<AQISchaeferIOData>(rtIO, ecoAssXML);
    calculateAQISchaefer(rtIO->aircraft_energy_carrier_type());
}

void AQISchaefer::calculateAQISchaefer(const std::string &fuelType) {
    AQISchaeferIOData_->myEngine.set_same_engine_type_quantity(AQISchaeferIOData_->numberOfEngines);
    /*ICAO Emission Regulatory (subsonic AC), s. Annex 16 S.III-2-1*/
    /*for CO*/
    double DpF00_CO = 118.; //max. permitted output of CO [g] per kN thrust (per engine)
    /*for HC*/
    double DpF00_HC = 19.6; //max. permissible output at HC [g] per kN thrust (per engine)
    /*for NOx*/
    double DpF00_NOx;
    AQISchaeferIOData_->myEngine.calculate_N1_with_penalties(0, 0., myAtm, 1., "takeoff", 0., 0.);
    double thrustTakeoff = AQISchaeferIOData_->myEngine.get_thrust_aircraft();
    double P3t_TO = AQISchaeferIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
    double P2t_TO = AQISchaeferIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St2);
    double pressRatio = P3t_TO / P2t_TO; //s. Def. ICAO Annex 16 I-1-1 (Referencen Pressure ratio)
    double TOThrustPerEngine = convertUnit(NEWTON, KILO, NEWTON, thrustTakeoff) / AQISchaeferIOData_->numberOfEngines; // [kN]
    //note: Regulatory NOx for Engines with first production model after 31 Dec. 2007 (III-2-4 d); for preliminary design based on most recent TW
    if (pressRatio <= 30.) {
        if (TOThrustPerEngine > 89.) {
            DpF00_NOx = 16.72 + 1.4080 * pressRatio;
        } else if (TOThrustPerEngine > 26.7 && TOThrustPerEngine <= 89.) {
            DpF00_NOx = 38.5486 + 1.6823 * pressRatio - 0.2453 * TOThrustPerEngine - 0.00308 * pressRatio * TOThrustPerEngine;
        } else {
            // NOTE(franz#1#): Mantis ID: 0000318 if the worst comes to the worst: just set a value!
            myRuntimeInfo->err << "Thrust category of " << TOThrustPerEngine << " kN not in ICAO regulations" << std::endl;
            exit(1);
        }
    } else if (pressRatio > 30. && pressRatio < 82.6) {
        if (TOThrustPerEngine > 89.) {
            DpF00_NOx = -1.04 + 2.0 * pressRatio;
        } else if (TOThrustPerEngine > 26.7 && TOThrustPerEngine <= 89.) {
            DpF00_NOx = 46.16 + 1.4286 * pressRatio - 0.5303 * TOThrustPerEngine + 0.00642 * pressRatio * TOThrustPerEngine;
        } else {
            // NOTE(franz#1#): Mantis ID: 0000319 if the worst comes to the worst: just set a value!
            myRuntimeInfo->err << "Thrust category of " << TOThrustPerEngine << " kN not in ICAO regulations" << std::endl;
            exit(1);
        }
    } else {
        DpF00_NOx = 32.0 + 1.6 * pressRatio;
    }
    /*for soot*/
    double c_soot_ICAO(0.);
    if (fuelType == "kerosene") {
        double SN_1 = 83.6 * pow(TOThrustPerEngine, -0.274);
        double SN_2 = 50.;
        double SN_ICAO = std::min(SN_1, SN_2);
        c_soot_ICAO = 0.0000026156 * pow(SN_ICAO, 4.) - 0.00010998 * pow(SN_ICAO, 3.) + 0.0022367 * pow(SN_ICAO, 3.) + 0.10955 * SN_ICAO + 0.0012842; //mg/kg
    } else if (fuelType == "liquid_hydrogen") {
    } else {
        myRuntimeInfo->warn << "No smoke number defined for chosen fuel type!" << std::endl;
    }

    /*AQI Calculation*/
    // TODO(franz#1#): T139 Gewichtungsfaktoren einprogrammieren
    Tolerance_CO = AQISchaeferIOData_->CO_LTO / AQISchaeferIOData_->numberOfEngines * 1000. / (DpF00_CO * TOThrustPerEngine);
    Tolerance_HC = AQISchaeferIOData_->HC_LTO / AQISchaeferIOData_->numberOfEngines * 1000. / (DpF00_HC * TOThrustPerEngine);
    Tolerance_NOx = AQISchaeferIOData_->NOx_LTO / AQISchaeferIOData_->numberOfEngines * 1000. / (DpF00_NOx * TOThrustPerEngine);
    if (fabs(c_soot_ICAO) > ACCURACY_HIGH) {
     Tolerance_soot = AQISchaeferIOData_->c_soot_LTO_max / c_soot_ICAO;
    } else {
        Tolerance_soot = 0;
    }

    AQI = (Tolerance_CO + Tolerance_HC + Tolerance_NOx + Tolerance_soot) / 4.;
    myRuntimeInfo->info << "Tolerance Factor CO: " << Tolerance_CO << std::endl;
    myRuntimeInfo->info << "Tolerance Factor HC: " << Tolerance_HC << std::endl;
    myRuntimeInfo->info << "Tolerance Factor NOx: " << Tolerance_NOx << std::endl;
    myRuntimeInfo->info << "Tolerance Factor Soot: " << Tolerance_soot << std::endl;
    myRuntimeInfo->info << "Air Quality Index: " << AQI << std::endl;
}

