# Import standard modules.
import math


def cabin_length_calculation(dict_ac_exchange, dict_mod_config, payload_tube_description, runtime_output):
    """ Calculates minimum required cabin length per tube from given accommodation definition.
        References and calculation method:
            [1] M.Sc Andreas Gobbin.
                Numerische Modellierung des Auslegungsprozesses für Passagierkabinen von Verkehrsflugzeugen
                unter Berücksichtigung der wichtigsten Auslegungsforderungen in MatLab.
                Fachgebiet Luftfahrzeugbau und Leichtbau, TU Berlin. Berlin. 2015.

            [2] Prof. Dr.-Ing. Juergen Thorbeck TU-Berlin (2006), Script "Flugzeugentwurf I und II" chapter A

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict payload_tube_description: Dict containing parameter and values of current blended wing body design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict payload_tube_description: Dict containing parameter and values of current payload tube design
            dict interior_description: Dict containing the interior description of each passenger class and payload tube
    """
    ''' initialize local parameter '''
    internal_bulk_area_per_pax_and_tube = []
    minimum_required_door_space = 0
    required_cabin_length_per_tube = {}
    interior_description = payload_tube_description['interior_description']
    # ToDo: ersetzen mit constants from lib
    convert_feet_to_inch = 12

    # Read minimum required door space.
    for exit_type in \
            dict_ac_exchange['dict_cs_requirements']['fuselage_design_cs_requirements']['exit_definition']['exit']:
        minimum_required_door_space = max(minimum_required_door_space, round(float(exit_type['width']['value']), 2))

    # Read galley information and set full size trolleys per pax for selected range scenario.
    range_scenario = dict_ac_exchange['relevant_classes']['range_scenario']
    pax_per_full_size_trolley = (
        dict_mod_config['pax_per_full_size_trolley_regional_range'] if range_scenario == 'regional'
        else dict_mod_config['pax_per_full_size_trolley_short_range'] if range_scenario == 'short_range'
        else dict_mod_config['pax_per_full_size_trolley_medium_range'] if range_scenario == 'medium_range'
        else dict_mod_config['pax_per_full_size_trolley_long_range']
        )

    # calculate passenger class and tube depending accommodation in accordance to [1]
    i = 0
    while i < len(dict_ac_exchange['relevant_classes']['passenger_classes']['class_names']):
        # read passenger class names and corresponding distribution.
        class_name = dict_ac_exchange['relevant_classes']['passenger_classes']['class_names'][i]
        class_distribution = dict_ac_exchange['relevant_classes']['class_percentages'][i]
        # calculate passenger per class by given class distribution
        pax_per_class = math.ceil(class_distribution * dict_ac_exchange['total_number_of_passengers'])
        # initialize loop depending parameter
        total_class_abreast = 0
        auto_set_of_lavatory = False
        total_class_abreast_per_tube = []
        seat_rows_per_class_and_tube = []
        seat_width_per_class_and_tube = []
        class_distribution_per_class_and_tube = []
        number_of_pax_per_class_and_tube = []
        seat_pitch_per_class_and_tube = []
        required_seating_area_per_class_and_tube = []
        number_of_lavatories_per_class_and_tube = []
        service_area_lavatory_per_class_and_tube = []
        number_of_wardrobe_per_class_and_tube = []
        service_area_wardrobe_per_class_and_tube = []
        number_of_full_size_trolley_per_class_and_tube = []
        service_area_galley_per_class_and_tube = []
        required_service_area_per_class_and_tube = []
        required_service_area_per_class_pax_and_tube = []
        cabin_length_per_class_and_tube = []
        required_space_for_passenger_way_per_class_and_tube = []

        # estimate total class abreast of all tubes and the class abreast per tube
        for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
            seat_abreast_per_tube = dict_ac_exchange['relevant_classes']['passenger_classes'][class_name][
                'seat_abreast'][tube].split('_')
            total_class_abreast_per_tube.append(sum((int(abreast) for abreast in seat_abreast_per_tube)))
            total_class_abreast = total_class_abreast + sum((int(abreast) for abreast in seat_abreast_per_tube))

        # calculate number of seat rows per class and payload tube
        for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
            class_distribution_per_class_and_tube.append(total_class_abreast_per_tube[tube] / total_class_abreast)
            seat_rows_per_class_and_tube.append(
                math.ceil((pax_per_class * class_distribution_per_class_and_tube[tube])
                          / total_class_abreast_per_tube[tube]))
            number_of_pax_per_class_and_tube.append(seat_rows_per_class_and_tube[tube]
                                                    * total_class_abreast_per_tube[tube])

            # calculate average seat pitch per class and tube
            seat_pitch_counter = 0
            sum_of_seat_pitch_per_class_and_tube = 0
            for seat_side, seat_pitch in \
                    dict_ac_exchange['relevant_classes']['passenger_classes'][class_name]['seat_pitch'].items():
                if not seat_pitch:
                    continue
                else:
                    sum_of_seat_pitch_per_class_and_tube = sum_of_seat_pitch_per_class_and_tube + seat_pitch[tube]
                    seat_pitch_counter += 1
            seat_pitch_per_class_and_tube.append(sum_of_seat_pitch_per_class_and_tube / seat_pitch_counter)

            # calculate average seat width per class and tube
            seat_width_per_class_and_tube.append(
                round(sum(payload_tube_description['interior_description'][class_name][
                              'width_of_seat_rows_per_tube'][tube])
                      / total_class_abreast_per_tube[tube], 3))

            maximum_cabin_width_per_tube = 0
            for passenger_class, passenger_cabin_width in payload_tube_description['passenger_cabin_width'].items():
                maximum_cabin_width_per_tube = max(maximum_cabin_width_per_tube, passenger_cabin_width[tube])

            # calculate required seating area per class and tube in inch^2
            required_seating_area_per_class_and_tube.append(
                round(seat_pitch_per_class_and_tube[tube] * seat_rows_per_class_and_tube[tube]
                      * maximum_cabin_width_per_tube, 3))

            # calculate service area of lavatories per class and tube
            if not (dict_mod_config['use_specific_number_of_class_lavatories'][
                    'use_specific_number_of_class_lavatories_ID' + str(tube) + '_ID'
                    + str(dict_ac_exchange['relevant_classes']['class_indexes'][i])]):
                number_of_lavatories_per_class_and_tube.append(0)
                service_area_lavatory_per_class_and_tube.append(0)
                auto_set_of_lavatory = True
            else:
                number_of_lavatories_per_class_and_tube.append(
                    dict_mod_config['specific_number_of_class_lavatories'][
                          'specific_number_of_class_lavatories_ID' + str(tube) + '_ID' + str(
                              dict_ac_exchange['relevant_classes']['class_indexes'][i])])
                service_area_lavatory_per_class_and_tube.append(
                    round(number_of_lavatories_per_class_and_tube[tube]
                          * float(dict_ac_exchange['relevant_classes']['passenger_classes']
                                  [class_name]['class_lavatory']['lavatory'][tube]['depth']['value'])
                          * float(dict_ac_exchange['relevant_classes']['passenger_classes']
                                  [class_name]['class_lavatory']['lavatory'][tube]['width']['value']), 3))

            # calculate service area of wardrobes per class and tube
            if dict_ac_exchange['relevant_classes']['passenger_classes'][class_name][
                                'class_wardrobe']['use_wardrobe'][tube]:
                number_of_wardrobe_per_class_and_tube.append(
                    (math.ceil(dict_ac_exchange['relevant_classes']['passenger_classes'][
                                   class_name]['class_wardrobe']['wardrobe_space'][tube]
                               * number_of_pax_per_class_and_tube[tube]
                               / float(dict_ac_exchange['relevant_classes']['passenger_classes']
                                       [class_name]['class_wardrobe']['dimension'][tube]['width']['value']))))
                service_area_wardrobe_per_class_and_tube.append(
                    round(number_of_wardrobe_per_class_and_tube[tube]
                          * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name][
                                      'class_wardrobe']['dimension'][tube]['width']['value'])
                          * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name][
                                      'class_wardrobe']['dimension'][tube]['depth']['value']), 3))
            else:
                number_of_wardrobe_per_class_and_tube.append(0)
                service_area_wardrobe_per_class_and_tube.append(0)

            # calculate service area of galleys per class and tube
            number_of_full_size_trolley_per_class_and_tube.append(
                math.ceil(number_of_pax_per_class_and_tube[tube] / pax_per_full_size_trolley))
            service_area_galley_per_class_and_tube.append(
                round(number_of_full_size_trolley_per_class_and_tube[tube]
                      * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['trolley_definition'][
                                               'full_size_trolley']['depth']['value'])
                      * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['trolley_definition'][
                                               'full_size_trolley']['width']['value']), 3))

        # check if the auto set of lavatory flag is true: -> if true: -> estimate minimum required
        if auto_set_of_lavatory:
            maximum_pax_per_lavatory = max(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name][
                                               'class_lavatory']['pax_per_lavatory'])
            number_of_class_lavatories = (math.ceil(pax_per_class / maximum_pax_per_lavatory))
            # check if any lavatories exist for the current class
            #  -> if true: -> calculate missing number and distribute evenly to all tubes
            if sum(number_of_lavatories_per_class_and_tube) > 0:
                # check if the required number of class lavatories is greater than the number of currently placed
                #  -> if true: -> add missing number of lavatories to the tube with the lowest number of lavatories
                if number_of_class_lavatories > sum(number_of_lavatories_per_class_and_tube):
                    missing_number_of_class_lavatories = \
                        number_of_class_lavatories - sum(number_of_lavatories_per_class_and_tube)
                    # loop across the number of missing lavatories to add one of this to the tube with the lost number
                    for i in range(1, missing_number_of_class_lavatories + 1):
                        index_of_tube_with_lowest_number_of_lavatories = \
                            number_of_lavatories_per_class_and_tube.index(
                                min(number_of_lavatories_per_class_and_tube))
                        number_of_lavatories_per_class_and_tube[index_of_tube_with_lowest_number_of_lavatories] = \
                            number_of_lavatories_per_class_and_tube[
                                index_of_tube_with_lowest_number_of_lavatories] + 1
                        # add lavatory service area to the service area of tube with the lowest number of lavatories
                        service_area_lavatory_per_class_and_tube[index_of_tube_with_lowest_number_of_lavatories] = \
                            round(service_area_lavatory_per_class_and_tube[
                                      index_of_tube_with_lowest_number_of_lavatories]
                                  + (float(dict_ac_exchange['relevant_classes']['passenger_classes']
                                           [class_name]['class_lavatory']['lavatory'][
                                            index_of_tube_with_lowest_number_of_lavatories]['depth']['value'])
                                     * float(dict_ac_exchange['relevant_classes']['passenger_classes']
                                             [class_name]['class_lavatory']['lavatory'][
                                              index_of_tube_with_lowest_number_of_lavatories]['width']['value'])), 3)
            #  -> estimate minimum required number and distribute evenly to all tubes
            else:
                # loop across the number of missing lavatories to add one of this to the tube with the lost number
                for i in range(1, number_of_class_lavatories + 1):
                    index_of_tube_with_lowest_number_of_lavatories = \
                        number_of_lavatories_per_class_and_tube.index(
                            min(number_of_lavatories_per_class_and_tube))
                    number_of_lavatories_per_class_and_tube[index_of_tube_with_lowest_number_of_lavatories] = \
                        number_of_lavatories_per_class_and_tube[
                            index_of_tube_with_lowest_number_of_lavatories] + 1
                    # add lavatory service area to the service area of tube with the lowest number of lavatories
                    service_area_lavatory_per_class_and_tube[index_of_tube_with_lowest_number_of_lavatories] = \
                        round(service_area_lavatory_per_class_and_tube[
                                  index_of_tube_with_lowest_number_of_lavatories]
                              + (float(dict_ac_exchange['relevant_classes']['passenger_classes']
                                       [class_name]['class_lavatory']['lavatory'][
                                           index_of_tube_with_lowest_number_of_lavatories]['depth']['value'])
                                 * float(dict_ac_exchange['relevant_classes']['passenger_classes']
                                         [class_name]['class_lavatory']['lavatory'][
                                             index_of_tube_with_lowest_number_of_lavatories]['width']['value'])), 3)

        # calculate total service area per class and passenger
        for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
            # read size of internal bulk area per pax for regional aircraft
            internal_bulk_area_per_pax_and_tube.append(
                dict_mod_config['specific_internal_bulk_area']['use_internal_bulk_area_ID' + str(tube)]
                if dict_mod_config['use_internal_bulk_area']['use_internal_bulk_area_ID' + str(tube)] else 0.0)

            required_service_area_per_class_and_tube.append(
                round(service_area_lavatory_per_class_and_tube[tube] + service_area_wardrobe_per_class_and_tube[tube]
                      + service_area_galley_per_class_and_tube[tube], 3))
            required_service_area_per_class_pax_and_tube.append(
                required_service_area_per_class_and_tube[tube] / number_of_pax_per_class_and_tube[tube])

            # calculate cabin length per class in accordance to [2]
            # single-aisle aircraft
            if dict_ac_exchange['fuselage_type'] == "single_aisle":
                cabin_length_per_class_and_tube.append(
                    round(number_of_pax_per_class_and_tube[tube] / total_class_abreast_per_tube[tube]
                          * (seat_pitch_per_class_and_tube[tube]
                             + (required_service_area_per_class_pax_and_tube[tube] / seat_width_per_class_and_tube[
                                tube])
                             + (internal_bulk_area_per_pax_and_tube[tube]
                                / (dict_mod_config['passenger_aisle_width']
                                   / total_class_abreast + seat_width_per_class_and_tube[tube]))
                             + minimum_required_door_space
                             * (total_class_abreast
                                / pax_per_class + seat_pitch_per_class_and_tube[tube]
                                / (float(dict_ac_exchange['dict_cs_requirements']
                                         ['fuselage_design_cs_requirements']
                                         ['maximum_allowed_distance_between_exits']['value'])
                                   * convert_feet_to_inch))), 3))

                # calculate passenger way width per class and payload tube
                required_space_for_passenger_way_per_class_and_tube.append(
                    round(cabin_length_per_class_and_tube[tube] * dict_mod_config['passenger_aisle_width'], 3))

            # wide-body aircraft
            if dict_ac_exchange['fuselage_type'] == "wide_body":
                cabin_length_per_class_and_tube.append(
                    round(number_of_pax_per_class_and_tube[tube] / total_class_abreast_per_tube[tube]
                          * (seat_pitch_per_class_and_tube[tube]
                             + (required_service_area_per_class_pax_and_tube[tube] / seat_width_per_class_and_tube[
                                tube])
                             + (internal_bulk_area_per_pax_and_tube[tube]
                                / (dict_mod_config['passenger_aisle_width']
                                   / total_class_abreast + seat_width_per_class_and_tube[tube]))
                             + 2 * minimum_required_door_space
                             * (total_class_abreast
                                / pax_per_class + seat_pitch_per_class_and_tube[tube]
                                / (float(dict_ac_exchange['dict_cs_requirements']
                                         ['fuselage_design_cs_requirements']
                                         ['maximum_allowed_distance_between_exits']['value'])
                                   * convert_feet_to_inch))), 3))
                required_space_for_passenger_way_per_class_and_tube[tube] = \
                    2 * required_space_for_passenger_way_per_class_and_tube

        ''' set class dependent interior parameter '''
        interior_description[class_name]['total_class_abreast_per_tube'] = total_class_abreast_per_tube
        interior_description[class_name]['seat_rows_per_class_and_tube'] = seat_rows_per_class_and_tube
        interior_description[class_name]['seat_pitch_per_class_and_tube'] = seat_pitch_per_class_and_tube
        interior_description[class_name]['number_of_pax_per_class_and_tube'] = number_of_pax_per_class_and_tube
        interior_description[class_name]['number_of_wardrobe_per_class_and_tube'] \
            = number_of_wardrobe_per_class_and_tube
        interior_description[class_name]['seat_rows_per_class_and_tube'] = seat_rows_per_class_and_tube
        interior_description[class_name]['number_of_lavatories_per_class_and_tube'] \
            = number_of_lavatories_per_class_and_tube
        interior_description[class_name]['number_of_full_size_trolley_per_class_and_tube'] \
            = number_of_full_size_trolley_per_class_and_tube
        interior_description[class_name]['required_seating_area_per_class_and_tube'] \
            = required_seating_area_per_class_and_tube
        interior_description[class_name]['required_service_area_per_class_and_tube'] \
            = required_service_area_per_class_and_tube
        interior_description[class_name]['cabin_length_per_class_and_tube'] = cabin_length_per_class_and_tube
        interior_description[class_name]['required_space_for_passenger_way_per_class_and_tube'] \
            = required_space_for_passenger_way_per_class_and_tube

    i += 1

    # calculate total required cabin length per payload tube
    for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
        cabin_length_per_tube = 0
        for class_name in dict_ac_exchange['relevant_classes']['passenger_classes']['class_names']:
            cabin_length_per_tube = \
                cabin_length_per_tube + interior_description[class_name]['cabin_length_per_class_and_tube'][tube]
        if tube == 0:
            required_cabin_length_per_tube['payload_tube_' + str(tube)] = round(cabin_length_per_tube, 3)
        elif dict_ac_exchange['number_of_payload_tubes'] % 2 == 0 and tube % 2 != 0:
            required_cabin_length_per_tube['payload_tube_' + str(tube)] = \
                max(required_cabin_length_per_tube['payload_tube_' + str(tube - 1)], cabin_length_per_tube)
        elif dict_ac_exchange['number_of_payload_tubes'] % 2 == 0 and tube % 2 == 0:
            required_cabin_length_per_tube['payload_tube_' + str(tube)] = round(cabin_length_per_tube, 3)
        elif dict_ac_exchange['number_of_payload_tubes'] % 2 != 0 and tube % 2 != 0:
            required_cabin_length_per_tube['payload_tube_' + str(tube)] = round(cabin_length_per_tube, 3)
        elif dict_ac_exchange['number_of_payload_tubes'] % 2 != 0 and tube % 2 == 0:
            required_cabin_length_per_tube['payload_tube_' + str(tube)] = \
                max(required_cabin_length_per_tube['payload_tube_' + str(tube - 1)], cabin_length_per_tube)

    ''' debug print '''
    runtime_output.debug('Debug: The cabin_length_calculation function was successfully executed!')

    return required_cabin_length_per_tube, interior_description
