/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "aircraft.h"

#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>
#include <unitConversion/unitConversion.h>
#include "performance_assessment_settings.h"

#ifdef _WIN32
    #include <windows.h>
#endif  // _WIN32
#include <iostream>
#include <cmath>
#include <algorithm>
#include <vector>
#include <utility>
#include <string>

using std::endl;

aircraft::aircraft(const std::shared_ptr<RuntimeIO>& rtIO, const performance_assessment_settings &mySettings)
    :
    rtIO(rtIO),
    mySettingsPt(&mySettings),  // Pointer to the settings
    polarXML(aixml::openDocument(rtIO->getIODir() + "/aeroData/" + rtIO->acxml.at("/aircraft_exchange_file/analysis/aerodynamics/polar/polar_file/value"))),
    // acftName(rtIO->aircraft_type() + rtIO->aircraft_model()),
    // MTOM(EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass").read(rtIO->acxml).value()),
    // MLM(EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_landing_mass/mass_properties/mass").read(rtIO->acxml).value()),
    // OME(EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_zero_fuel_mass/mass_properties/mass").read(rtIO->acxml).value()),
    // MFM(EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_fuel_mass/mass_properties/mass").read(rtIO->acxml).value()),
    // maxPayload(EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_fuel_mass/mass_properties/mass").read(rtIO->acxml).value()),
    // MMO(EndnodeReadOnly<double>(
    //     "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_flight_envelopemaximum_operating_mach_number")
    //     .read(rtIO->acxml).value()),
    // VMO(EndnodeReadOnly<double>(
    //     "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_flight_envelope/maximum_operating_velocity")
    //     .read(rtIO->acxml).value()),
    // MDive(EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/speed/dive_mach_number").read(rtIO->acxml).value()),
    // VDive(EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/speed/dive_velocity").read(rtIO->acxml).value()),
    // MInitialCruise(EndnodeReadOnly<double>(
    //     "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission_requirements/initial_cruise_Mach_number")
    //     .read(rtIO->acxml).value()),
    // hInitialCruise(EndnodeReadOnly<double>(
    //     "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission_requirements/initial_cruise_altitude")
    //     .read(rtIO->acxml).value()),
    // hMaxOperating(EndnodeReadOnly<double>(
    //     "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_flight_envelope/maximum_operating_altitude")
    //     .read(rtIO->acxml).value()),
    // timeToClimb(EndnodeReadOnly<double>(
    //     "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission_requirements/time_to_climb")
    //     .read(rtIO->acxml).value()),
    // designRange(EndnodeReadOnly<double>(
    //     "/aircraft_exchange_file/analysis/mission/design_mission/range").read(rtIO->acxml).value()),
    // designPayload(EndnodeReadOnly<double>(
    //     "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/maximum_structrual_payload_mass")
    //     .read(rtIO->acxml).value()),
    // rangeMaxPayload(EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/range_max_payload_at_maximum_takeoff_mass").read(rtIO->acxml).value()),
    // rangeMaxFuel(EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/range_max_fuel_at_maximum_takeoff_mass").read(rtIO->acxml).value()),
    // payloadRangeMaxfuel(EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/payload_maximum_fuel_at_maximum_takeoff_mass").read(rtIO->acxml).value()),
    // rangeFerry(EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/range_maximum_fuel_empty").read(rtIO->acxml).value()),
    // numberEngines(rtIO->acxml.getVector(("/aircraft_exchange_file/component_design/propulsion/specific/propulsion"), 1).size()),
    // S_Wing(EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/aerodynamics/reference_values/S_ref").read(rtIO->acxml).value()),
    // optCL(EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/aerodynamics/lift_coefficients/C_LoptimumCruise").read(rtIO->acxml).value()),
    // frictionCoefficient(EndnodeReadOnly<double>(
    //     "/aircraft_exchange_file/requirements_and_specifications/requirements/additional_requirements/landing_gear/friction_coefficient").read(rtIO->acxml).value()),
    // brakingCoefficient(EndnodeReadOnly<double>(
    //     "/aircraft_exchange_file/requirements_and_specifications/requirements/additional_requirements/landing_gear/breaking_coefficient").read(rtIO->acxml).value()),
    // // myEngine(rtIO->getEngineDataDir()),
    myEngines(rtIO),
    aero(rtIO->acxml, polarXML),
    delta_isa(EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/" + this->getMissionType() + "/delta_ISA").read(rtIO->acxml).value()),
    // myMission(string(rtIO->getIODir(), rtIO->acxml.at("/aircraft_exchange_file/requirements_and_specifications/mission_files/" + this->getMissionType() + "/value")))
    myMission(rtIO->getIODir() + "/missionData/" + rtIO->acxml.at("/aircraft_exchange_file/requirements_and_specifications/mission_files/" + this->getMissionType() + "/value"),
        delta_isa)
    // engineInletD(checkBoundaries(rtIO->acxml.at("AcftExchangeFile/Propulsion/Engine@1/EngineDimensions/w_Engine"), 0., true, 20., true)), //Todo
    // factorVLOF(1.1),
    // factorV2(1.1)
    {

    // polarXML = aixml::openDocument(rtIO->getIODir() + "/aeroData/" + rtIO->acxml.at("/aircraft_exchange_file/analysis/aerodynamics/polar/polar_file/value"));
    acftName = rtIO->aircraft_type() + rtIO->aircraft_model();
    MTOM = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass").read(rtIO->acxml).value();
    MLM = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_landing_mass/mass_properties/mass").read(rtIO->acxml).value();
    OME = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_zero_fuel_mass/mass_properties/mass").read(rtIO->acxml).value();
    MFM =EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_fuel_mass/mass_properties/mass").read(rtIO->acxml).value();
    maxPayload = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_fuel_mass/mass_properties/mass").read(rtIO->acxml).value();
    MMO = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_flight_envelopemaximum_operating_mach_number")
        .read(rtIO->acxml).value();
    VMO = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_flight_envelope/maximum_operating_velocity")
        .read(rtIO->acxml).value();
    MDive = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/speed/dive_mach_number").read(rtIO->acxml).value();
    VDive = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/speed/dive_velocity").read(rtIO->acxml).value();
    MInitialCruise = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission_requirements/initial_cruise_Mach_number")
        .read(rtIO->acxml).value();
    hInitialCruise = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission_requirements/initial_cruise_altitude")
        .read(rtIO->acxml).value();
    hMaxOperating = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_flight_envelope/maximum_operating_altitude")
        .read(rtIO->acxml).value();
    timeToClimb = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission_requirements/time_to_climb")
        .read(rtIO->acxml).value();
    designRange = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/mission/design_mission/range").read(rtIO->acxml).value();
    designPayload = EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/maximum_structrual_payload_mass")
        .read(rtIO->acxml).value();
    rangeMaxPayload = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/range_max_payload_at_maximum_takeoff_mass").read(rtIO->acxml).value();
    rangeMaxFuel = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/range_max_fuel_at_maximum_takeoff_mass").read(rtIO->acxml).value();
    payloadRangeMaxfuel = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/payload_maximum_fuel_at_maximum_takeoff_mass").read(rtIO->acxml).value();
    rangeFerry = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/range_maximum_fuel_empty").read(rtIO->acxml).value();
    numberEngines = rtIO->acxml.getVector(("/aircraft_exchange_file/component_design/propulsion/specific/propulsion"), 1).size();
    S_Wing = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/aerodynamics/reference_values/S_ref").read(rtIO->acxml).value();
    optCL = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/aerodynamics/lift_coefficients/C_LoptimumCruise").read(rtIO->acxml).value();
    frictionCoefficient = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/additional_requirements/landing_gear/friction_coefficient").read(rtIO->acxml).value();
    brakingCoefficient = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/additional_requirements/landing_gear/breaking_coefficient").read(rtIO->acxml).value();
    // myEngine(rtIO->getEngineDataDir());
    // myEngines(rtIO);
    // aero = aerodynamics(rtIO->acxml, polarXML);
    // myMission = string(rtIO->getIODir(), rtIO->acxml.at("/aircraft_exchange_file/requirements_and_specifications/mission_files/" + this->getMissionType() + "/value"));
    engineInletD = checkBoundaries(rtIO->acxml.at("AcftExchangeFile/Propulsion/Engine@1/EngineDimensions/w_Engine"), 0., true, 20., true); //Todo
    factorVLOF = 1.1;
    factorV2 = 1.1;

    /* Constructor */
    /* Requirements */
    myReqs.OEICruiseAltitude = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_one_engine_operating_altitude")
        .read(rtIO->acxml).value();
    myReqs.TOFL = EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/takeoff_distance/value")
        .read(rtIO->acxml).value();
    myReqs.LDN = EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/landing_field_length/value")
        .read(rtIO->acxml).value();
    myReqs.approachSpeed = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_approach_speed/value")
        .read(rtIO->acxml).value();
    myReqs.setRequiredClimbGradients(rtIO->acxml.at("/AcftExchangeFile/Requirements/DesignMission/ClimbGradients/"), this->numberEngines);
    myChecks.TOFL = false;
    myChecks.LDN = false;
    myChecks.vApproach = false;
    myChecks.initCruiseAlt = false;
    myChecks.maxOptAlt = false;
    myChecks.OEICruiseAlt = false;
    myChecks.initCruiseM = false;
    myChecks.climbGradientSecondSegment = false;
    myChecks.climbGradientFinalSegment = false;
    if (OME + MFM > MTOM) {
        MFM = MTOM - OME;
    }
    // Energy provider
    myEnergyProviders.energyCarrierID_ground =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/ground")).value();
    myEnergyProviders.energyCarrierID_taxi =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/taxi")).value();
    myEnergyProviders.energyCarrierID_takeoff =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/takeoff")).value();
    myEnergyProviders.energyCarrierID_climb =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/climb")).value();
    myEnergyProviders.energyCarrierID_cruise =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/cruise")).value();
    myEnergyProviders.energyCarrierID_descent =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/descent")).value();
    myEnergyProviders.energyCarrierID_landing =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/landing")).value();


    //Determination of V2 for minimum climb gradients
    factorV2 = this->getMinimumFactorForV2();
    if (factorV2 <= 1.128) { //Par CS25.107: v2 !> 1.2 * vS or 1.128 * vSR (vS1g)
        factorVLOF = factorV2;
    } else {
        factorVLOF = 1.128;
    }
    /* Set atmosphere */
    this->atm.setAtmosphere(0., ISA_TEMPERATURE + myMission.deltaISA, ISA_PRESSURE);


}

aircraft::~aircraft() {
    // dtor
}

string aircraft::getMissionType() {
    string missionType;
    if (!mySettingsPt->designForMTOM && (mySettingsPt->mySwitches.doFuelPlanning && mySettingsPt->useStudyMissionForAnalysis)) {
        missionType = "study_mission_file";
    } else {
        missionType = "design_mission_file";
    }
    return missionType;
}

double aircraft::getReferenceStallSpeed(const double& aircraftMass, const double& flightAltitude, const double& maximumLiftCoefficient) {
    return sqrt((aircraftMass * G_FORCE) / (0.5 * this->atm.getDensity(flightAltitude) * maximumLiftCoefficient * this->S_Wing));
}

double aircraft::v1gS_Clean(double Mass, double Altitude) {
    return 1.0 * this->getReferenceStallSpeed(Mass, Altitude, this->aero.getCLmax(0., "Clean"));
}

double aircraft::v1gS_TO(double Mass, double Altitude) {
    return 1.0 * this->getReferenceStallSpeed(Mass, Altitude, this->aero.CLmaxTakeoff);
}

double aircraft::vLOF(double Mass, double Altitude) {
    return factorVLOF * this->getReferenceStallSpeed(Mass, Altitude, this->aero.CLmaxTakeoff);
}

double aircraft::v2(double Mass, double Altitude) {
    return factorV2 * this->getReferenceStallSpeed(Mass, Altitude, this->aero.CLmaxTakeoff);
}

double aircraft::vFTO(double Mass, double Altitude) {
    return 1.65 * this->getReferenceStallSpeed(Mass, Altitude, this->aero.CLmaxTakeoff);
}

double aircraft::vAPPR(double Mass, double Altitude) {
    return 1.23 * this->getReferenceStallSpeed(Mass, Altitude, this->aero.CLmaxLanding);
}

double aircraft::ROC(double Mass, double Alt, double Mach, double fACC, string Configuration, string Rating, double bleedOff, double shaftOff, bool OEI,
                     const aerodynamics& thisAERO) {
    // s. Scheiderer P.230 ff.
    // Determine engine carrier for climb phase
    int energy_carrier_id =  // TODO(hansmann): Consideration of varying energy carriers and operating strategies (shall be covered by engine.lib or multi_engine_prop.lib)
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/climb")).value();
    myEngines.calculate_N1_with_penalties(energy_carrier_id, Alt, Mach, atm, 1.0, Rating, bleedOff, shaftOff);  // Set operating point
    double ac_thrust = myEngines.get_thrust_aircraft(energy_carrier_id);
    double tempDrag(thisAERO.getCruiseDrag(Mach, Alt, 0.0, Mass, Configuration, atm));
    if (OEI == true) {
        ac_thrust = ac_thrust * static_cast<double>(numberEngines - 1) / static_cast<double>(numberEngines);
        tempDrag += calcWindMillingDrag(Mach, Alt);
    }
    double tempGamma = asin((ac_thrust - tempDrag) / (G_FORCE * Mass) / (1. + fACC));  // because not with const. TAS but with const. IAS or Ma is risen
                                                                                    //--> Introduction of facc: Commutation factor: facc = V/g * dV/dh; see table 11.1 p.232
    double tempLiftCoefficient = 2.*(G_FORCE * Mass * cos(tempGamma))
                          / (atm.getDensity(Alt) * pow(Mach * atm.getSpeedOfSound(Alt), 2) * S_Wing);
    tempDrag = thisAERO.getCLDrag(Mach, Alt, tempLiftCoefficient, Configuration, atm);
    double tempROC = ((ac_thrust - tempDrag) / (G_FORCE * Mass) * (Mach * atm.getSpeedOfSound(Alt))) / (1. + fACC);
    return tempROC;
}

double aircraft::calcWindMillingDrag(double Mach, double Altitude) {
    double tempCdWMJ(0.0);
    tempCdWMJ = 0.0785 * pow(engineInletD, 2.) / S_Wing
                + 2. / (1. + 0.16 * pow(Mach, 2.)) * 0.92 * (1.0 - 0.92) * (PI / 4. * pow(engineInletD, 2.)) / S_Wing;
    // tempCdWMJ = 0.75 * tempCdWMJ;
    return 0.5 * atm.getDensity(Altitude) * pow(Mach * atm.getSpeedOfSound(Altitude), 2.) * S_Wing * tempCdWMJ;
}

double aircraft::getMinimumFactorForV2() {
    double altitude(convertUnit(FOOT, METER, 400.));  // End of 2nd climb segment for standard departure according to FAR
    double factor(1.); // Start value
    double tempClimbGradient(NAN);
    do {
        factor += 0.0025;
        double Mach = factor * this->v1gS_TO(this->MTOM, altitude) / this->atm.getSpeedOfSound(altitude);
        double tempTAS = convertUnit(MACH, TRUEAIRSPEED, altitude, this->atm, Mach);
        //this->ROC(..., true, ...): OEI active, which is a requirement for minimum climb gradient in 2nd climb segment
        double tempROC = this->ROC(this->MTOM, altitude, Mach, 0.0, "Climb", "TO", this->myMission.bleedOffTO,  this->myMission.shaftOffTO, true, this->aero);
        tempClimbGradient = tempROC / tempTAS * 100.;
    } while (tempClimbGradient < myReqs.climbGradientSecondTOSegment);
    return factor;
}

void aircraft::Requirements::setRequiredClimbGradients(const node& climbGradientRequirement, const uint16_t& numberOfEngines) {
    if (numberOfEngines == 1) {
        myRuntimeInfo->err << "Climbing requirements are invalid for a single engine aircraft!" << endl;
        exit(1);
    } else if (numberOfEngines == 2) {
        this->climbGradientSecondTOSegment = checkBoundaries(climbGradientRequirement.at("secondTOsegment"), 2.4, true, 10., true);
        this->climbGradientFinalTOSegment = checkBoundaries(climbGradientRequirement.at("finalTOsegment"), 1.2, true, 10., true);
        this->climbGradientApproachOEI = checkBoundaries(climbGradientRequirement.at("approachOEI"), 2.1, true, 10., true);
    } else if (numberOfEngines == 3) {
        this->climbGradientSecondTOSegment = checkBoundaries(climbGradientRequirement.at("secondTOsegment"), 2.7, true, 10., true);
        this->climbGradientFinalTOSegment = checkBoundaries(climbGradientRequirement.at("finalTOsegment"), 1.5, true, 10., true);
        this->climbGradientApproachOEI = checkBoundaries(climbGradientRequirement.at("approachOEI"), 2.4, true, 10., true);
    } else if (numberOfEngines >= 4) {
        this->climbGradientSecondTOSegment = checkBoundaries(climbGradientRequirement.at("secondTOsegment"), 3.0, true, 10., true);
        this->climbGradientFinalTOSegment = checkBoundaries(climbGradientRequirement.at("finalTOsegment"), 1.7, true, 10., true);
        this->climbGradientApproachOEI = checkBoundaries(climbGradientRequirement.at("approachOEI"), 2.7, true, 10., true);
    }
    this->climbGradientApproachAEO = checkBoundaries(climbGradientRequirement.at("approachAEO"), 3.2, true, 10., true);
}
