/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "flightSimulator.h"

using std::endl;
using std::string;

flightSimulator::flightSimulator(aircraft *theAircraft)
    :
    theAcftPt(theAircraft),
    climbAltSteps(50.),
    cruiseSteps(10.) {
    //ctor
}

flightSimulator::~flightSimulator() {
    //dtor
}

double flightSimulator::calcTaxiFuel(double taxiTime) {
    double totalTaxiTime = convertUnit(MINUTE, SECOND, taxiTime); // in s
    theAcftPt->myEngines.calculate_N1_with_penalties(theAcftPt->myEnergyProviders.energyCarrierID_taxi, 0., 0., theAcftPt->atm, 1., "Idle", theAcftPt->myMission.bleedOffTO,
        theAcftPt->myMission.shaftOffTO);
    double theTaxiFuel = totalTaxiTime * theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_taxi);
    return theTaxiFuel;
}

/* Climb Fuel */
void   flightSimulator::getClimbFuel(double grossMass, double startAlt, double endAlt, double startMach, double endMach, const string &Config, const string &Rating,
                                     double bleedOff, double shaftOff, const aerodynamics &theAero, double *climbFuel, double *climbDistance, double *climbTime) {
    // linear Mach curve
    double machGradient = (endMach - startMach) / (endAlt - startAlt);
    double machConstant = endMach - machGradient * endAlt;
    double theAltitude = startAlt + climbAltSteps;
    while (theAltitude <= endAlt) {
        double segFuel(0.);
        double segTime(0.);
        double segDistance(0.);
        double M0 = machGradient * (theAltitude - climbAltSteps) + machConstant;
        double M1 = machGradient * theAltitude + machConstant;
        calcClimbFuel(grossMass, theAltitude - climbAltSteps, theAltitude,
                      M0, M1,
                      Config, Rating, bleedOff, shaftOff,
                      theAero,
                      &segTime, &segFuel, &segDistance);
        grossMass     -= segFuel;
        *climbFuel      += segFuel;
        *climbTime      += segTime;
        *climbDistance  += segDistance;
        theAltitude     += climbAltSteps;
    }
}

void flightSimulator::calcClimbFuel(double massAC, double H0, double H1, double Mach0, double Mach1, const string& config, string rating, double bleed, double shaft,
                                    aerodynamics AERO, double *theTime, double *theFuel, double *theDistance) {
    /* minimum climb performance */
    double a_minROC = convertUnit(FOOT, METER, theAcftPt->hInitialCruise) / theAcftPt->timeToClimb; //minimum ROC to maintain TTC requirement [m/s]
    double m_minROC = (a_minROC - convertUnit(FOOTPERMINUTE, METERPERSECOND, 500.)) / (0. - convertUnit(FOOT, METER, theAcftPt->hInitialCruise));

    double minimumROC0 = m_minROC * H0 + a_minROC;
    double minimumROC1 = m_minROC * H1 + a_minROC;
    // estimate according to Raymer, p.463 and equation 17.48-17.49
    double ROC0 = theAcftPt->ROC(massAC, H0, Mach0, 0.0, config, rating, bleed, shaft, false, AERO);
    double ROC1 = theAcftPt->ROC(massAC, H1, Mach1, 0.0, config, rating, bleed, shaft, false, AERO);
    //Increase the setting to "MaxCont"
    if (ROC0 <= minimumROC0 || ROC1 <= minimumROC1) {
        rating = "MaxCont";
        ROC0 = theAcftPt->ROC(massAC, H0, Mach0, 0.0, config, rating, bleed, shaft, false, AERO);
        ROC1 = theAcftPt->ROC(massAC, H1, Mach1, 0.0, config, rating, bleed, shaft, false, AERO);
        if (ROC0 <= 0. || ROC1 <= 0.) {
            myRuntimeInfo->err << "  No climb possible!" << endl;
            myRuntimeInfo->err << "  H0: " << H0 << " m    ->   H1: " << H1 << " m" << endl;
            myRuntimeInfo->err << "  M0: " << Mach0 << "      ->   M1: " << Mach1 << "" << endl;
            myRuntimeInfo->err << "  Configuration: " << config << endl;
            myRuntimeInfo->err << "  Setting:       " << rating << endl;
            myRuntimeInfo->err << "  ROC0: " << ROC0 << " m/s" << endl;
            myRuntimeInfo->err << "  ROC1: " << ROC1 << " m/s" << endl;
            myRuntimeInfo->err << "  Abort!" << endl;
            exit(1);
        }
        if (ROC0 <= minimumROC0 || ROC1 <= minimumROC1) {
            myRuntimeInfo->err << "The minimum climb requirement cannot be met!" << endl;
            myRuntimeInfo->err << "  H0: " << H0 << " m    ->   H1: " << H1 << " m" << endl;
            myRuntimeInfo->err << "  M0: " << Mach0 << "      ->   M1: " << Mach1 << "" << endl;
            myRuntimeInfo->err << "  Configuration: " << config << endl;
            myRuntimeInfo->err << "  Setting:       " << rating << endl;
            myRuntimeInfo->err << "  ROC0: " << ROC0 << " m/s" << endl;
            myRuntimeInfo->err << "  ROC0required: " << minimumROC0 << " m/s" << endl;
            myRuntimeInfo->err << "  ROC1: " << ROC1 << " m/s" << endl;
            myRuntimeInfo->err << "  ROC1required: " << minimumROC1 << " m/s" << endl;
            myRuntimeInfo->err << "  Abort!" << endl;
            exit(1);
        } else {
            ROC0 = minimumROC0;
            ROC1 = minimumROC1;
        }
    }
    /* Average Fuel Flow */
    double avgFuelFlow(0.0);
    // Point 0
    double Thrust = calcThrustForROC(massAC, H0, Mach0, ROC0, config, AERO);
    theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_climb, H0, Mach0, theAcftPt->atm, 1.0, rating, bleed, shaft, Thrust);
    if (theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_climb) < 0.99 * Thrust) {
        myRuntimeInfo->err << "Engines do not provide the necessary thrust for climbing!" << endl;
        myRuntimeInfo->err << "  H: " << H0 << endl;
        myRuntimeInfo->err << "  M: " << Mach0 << endl;
        myRuntimeInfo->err << "  Configuration: " << config << endl;
        myRuntimeInfo->err << "  Setting:       " << rating << endl;
        myRuntimeInfo->err << "  ROC: " << ROC0 << " m/s" << endl;
        myRuntimeInfo->err << "  Thrust_required:  " << Thrust << " kN" << endl;
        myRuntimeInfo->err << "  Thrust_available: " << theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_climb)/1000 << " kN" << endl;
        // myRuntimeInfo->err << "  N1: " << theAcftPt->myEngine.() << endl;  // TODO(hansmann): Currently no method available to get N1 afaik
        myRuntimeInfo->err << "Abort!" << endl;
        exit(1);
    }
    avgFuelFlow += theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_climb);
    // Point 1
    Thrust = calcThrustForROC(massAC, H1, Mach1, ROC1, config, AERO);
    theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_climb, H1, Mach1, theAcftPt->atm, 1.0, rating, bleed, shaft, Thrust);
    if (theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_climb) < 0.99 * Thrust) {
        myRuntimeInfo->err << "Engines do not provide the necessary thrust for climbing!" << endl;
        myRuntimeInfo->err << "  H: " << H1 << endl;
        myRuntimeInfo->err << "  M: " << Mach1 << endl;
        myRuntimeInfo->err << "  Configuration: " << config << endl;
        myRuntimeInfo->err << "  Setting:       " << rating << endl;
        myRuntimeInfo->err << "  ROC: " << ROC0 << " m/s" << endl;
        myRuntimeInfo->err << "  Thrust_required:  " << Thrust << " kN" << endl;
        myRuntimeInfo->err << "  Thrust_available: " << theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_climb)/1000 << " kN" << endl;
        // myRuntimeInfo->err << "  N1: " << theAcftPt->myEngine.getEngineN() << endl;
        myRuntimeInfo->err << "Abort!" << endl;
        exit(1);
    }
    avgFuelFlow += theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_climb);
    avgFuelFlow = 0.5 * avgFuelFlow;
    /* Average TAS */
    double avgTAS(0.);
    avgTAS = 0.5 * (Mach0 * theAcftPt->atm.getSpeedOfSound(H0) +  Mach1 * theAcftPt->atm.getSpeedOfSound(H1));
    /* Average ROC */
    double avgROC = 0.5 * (ROC0 + ROC1);
    *theTime = fabs(H1 - H0) / avgROC;
//    Raymer, P.463, eq.17.47
//    double a = (ROC1 - ROC0) / (H1 - H0); //Warning: denominator of ROC_constant = 0!!!!
//    *theTime = (-1./a) * log( ROC0/ROC1 ); //Warning: log(1) = 0 !!!!
    *theFuel = *theTime * avgFuelFlow;
    *theDistance = convertUnit(METER, KILO, METER, sqrt(pow((*theTime * avgTAS), 2) - pow((H1 - H0), 2)));
}

double flightSimulator::calcThrustForROC(double massOfAC, double H, double Mach, double theROC, string theConfiguration, aerodynamics theAERO) {
    double reqGamma = asin(theROC / (Mach * theAcftPt->atm.getSpeedOfSound(H)));
    double regLiftCoefficient = 2. * G_FORCE * massOfAC * cos(reqGamma) /
                                (theAcftPt->atm.getDensity(H) * pow(Mach * theAcftPt->atm.getSpeedOfSound(H), 2.) * theAcftPt->S_Wing);
    double reqDrag = theAERO.getCLDrag(Mach, H, regLiftCoefficient, theConfiguration, theAcftPt->atm);
    double reqThrust = reqDrag + (G_FORCE * massOfAC) * theROC / (Mach * theAcftPt->atm.getSpeedOfSound(H));
    return convertUnit(NEWTON, KILO, NEWTON, reqThrust); //kN
}

/* Accelerate Fuel */
void flightSimulator::getAccelerateFuel(double grossMass, double Altitude, double startMach, double endMach, string Config, string Rating, double bleedOff, double shaftOff,
                                        aerodynamics theAero, double *accFuel, double *accDistance, double *accTime) {
    if (startMach > endMach) {
        myRuntimeInfo->err << "An acceleration is to be calculated, but" << endl;
        myRuntimeInfo->err << "Mach0 = " << startMach << " > Mach1 = " << endMach << endl;
        myRuntimeInfo->err << "Abort!" << endl;
        exit(1);
    }
    // assumption of a constant acceleration -> linear increase in velocity
    double avgMach = 0.5 * (startMach + endMach);
    theAcftPt->myEngines.calculate_N1_with_penalties(theAcftPt->myEnergyProviders.energyCarrierID_cruise, Altitude, avgMach, theAcftPt->atm, 1.0, Rating, bleedOff, shaftOff);
    double avgAcc = (1. / grossMass) * (convertUnit(KILO, NEWTON, NEWTON, theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise))
                                          - theAero.getCruiseDrag(avgMach, Altitude, 0.0, grossMass, Config, theAcftPt->atm));
    if (avgAcc <= 0. && Rating == "Cruise") {
        Rating = "Climb";
        theAcftPt->myEngines.calculate_N1_with_penalties(theAcftPt->myEnergyProviders.energyCarrierID_cruise, Altitude, avgMach, theAcftPt->atm, 1.0, Rating, bleedOff, shaftOff);
        avgAcc = (1. / grossMass) * (convertUnit(KILO, NEWTON, NEWTON, theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise))
                                       - theAero.getCruiseDrag(avgMach, Altitude, 0.0, grossMass, Config, theAcftPt->atm));
    }
    if (avgAcc <= 0. && Rating == "Climb") {
        Rating = "MaxCont";
        theAcftPt->myEngines.calculate_N1_with_penalties(theAcftPt->myEnergyProviders.energyCarrierID_climb, Altitude, avgMach, theAcftPt->atm, 1.0, Rating, bleedOff, shaftOff);
        avgAcc = (1. / grossMass) * (convertUnit(KILO, NEWTON, NEWTON, theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_climb))
                                       - theAero.getCruiseDrag(avgMach, Altitude, 0.0, grossMass, Config, theAcftPt->atm));
    }
    if (avgAcc <= 0.) {
        myRuntimeInfo->err << "No acceleration possible!" << endl;
        myRuntimeInfo->err << "Altitude: " << convertUnit(METER, FOOT, Altitude) << " ft" << endl;
        myRuntimeInfo->err << "Mach0:    " << startMach << endl;
        myRuntimeInfo->err << "Mach1:    " << endMach << endl;
        myRuntimeInfo->err << "Config:   " << Config << endl;
        myRuntimeInfo->err << "Rating:   " << Rating << endl;
        myRuntimeInfo->err << "Abort!" << endl;
        exit(1);
    }
    *accTime = fabs(endMach - startMach) * theAcftPt->atm.getSpeedOfSound(Altitude) / avgAcc;
    *accDistance = convertUnit(METER, KILO, METER, *accTime * fabs(endMach - startMach) * theAcftPt->atm.getSpeedOfSound(Altitude));
    *accFuel = theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_cruise) * *accTime;;
}

bool flightSimulator::checkForAcceleration(double grossMass, double Altitude, double Mach, string Config, string Rating, double Derate, double bleedOff, double shaftOff,
                                           aerodynamics theAero) {
    double theDrag = theAero.getCruiseDrag(Mach, Altitude, 0., grossMass, Config, theAcftPt->atm);
    theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_cruise, Altitude, Mach, theAcftPt->atm, Derate, Rating, bleedOff, shaftOff,
        convertUnit(NEWTON, KILO, NEWTON, theDrag));  // TODO(hansmann): cuise is currently hardcoded here as energy carrier

    if (0.995 * convertUnit(NEWTON, KILO, NEWTON, theDrag) > theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise)) {
        return false;
    } else {
        return true;
    }
}

/* Cruise Flight */
void flightSimulator::getCruiseFuel(double grossMass, double Altitude, double Mach, string Config, string Rating, double bleedOff, double shaftOff, aerodynamics theAero,
                                    double *cruiseFuel, double cruiseDistance, double *cruiseTime) {
    *cruiseTime = 0.;
    *cruiseFuel = 0.;
    if (cruiseDistance > 0.) {
        double flownDistance(0.);
        do {
            double theDrag = theAero.getCruiseDrag(Mach, Altitude, 0., grossMass, Config, theAcftPt->atm);
            theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_cruise, Altitude, Mach, theAcftPt->atm, 1.0, Rating, bleedOff, shaftOff,
                convertUnit(NEWTON, KILO, NEWTON, theDrag));
            double theThrust = theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise);
            if (0.995 * convertUnit(NEWTON, KILO, NEWTON, theDrag) > theThrust) {
                myRuntimeInfo->err << "Error in cruise calculation! Abort!!!" << endl;
                myRuntimeInfo->err << "Thrust is not equal to resistance:" << endl;
                myRuntimeInfo->err << "-> Thrust: " << theThrust << " kN" << endl;
                myRuntimeInfo->err << "-> Drag:  " << convertUnit(NEWTON, KILO, NEWTON, theDrag) << " kN" << endl;
                exit(1);
            }
            double segTime = (cruiseSteps * 1000.) / (Mach * theAcftPt->atm.getSpeedOfSound(Altitude));
            double segFuel = segTime * theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_cruise);
            *cruiseTime += segTime;
            *cruiseFuel += segFuel;
            grossMass -= segFuel;
            flownDistance += cruiseSteps;
        } while (flownDistance <= cruiseDistance);
    } else {
        *cruiseTime = 0.;
        *cruiseFuel = 0.;
    }
}

bool flightSimulator::checkForSteadyLevelFlight(double grossMass, double Altitude, double Mach, string Config, string Rating, double bleedOff, double shaftOff,
                                                aerodynamics theAero) {
    double theDrag = theAero.getCruiseDrag(Mach, Altitude, 0., grossMass, Config, theAcftPt->atm);
    theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_cruise, Altitude, Mach, theAcftPt->atm, 1.0, Rating, bleedOff, shaftOff,
        convertUnit(NEWTON, KILO, NEWTON, theDrag));
    double theThrust = theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise);
    if (0.995 * convertUnit(NEWTON, KILO, NEWTON, theDrag) > theThrust) {
        return false;
    } else {
        return true;
    }
}

/* Descend */
void flightSimulator::getDescendFuel(double startAltitude, double endAltitude, double startMach, double endMach, double bleedOff, double shaftOff, double *descendFuel,
                                     double *descendDistance, double *descendTime) {
    double deltaAltitude = fabs(startAltitude - endAltitude);
    double avgTAS = 0.5 * (startMach * theAcftPt->atm.getSpeedOfSound(startAltitude)
                           + endMach * theAcftPt->atm.getSpeedOfSound(endAltitude));
    *descendDistance = convertUnit(METER, KILO, METER, 1. / tan(convertUnit(DEGREE, RADIAN, 5.)) * deltaAltitude);
    *descendTime = sqrt(pow(convertUnit(KILO, METER, METER, *descendDistance), 2.) + pow(deltaAltitude, 2)) / avgTAS;
    theAcftPt->myEngines.calculate_N1_with_penalties(theAcftPt->myEnergyProviders.energyCarrierID_descent, startAltitude, startMach, theAcftPt->atm, 1., "Idle", bleedOff,
        shaftOff);
    *descendFuel = *descendTime * theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_descent);
}
