/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef performance_assessment_SRC_TAW_DEFAULTMETHODS_FUELPLANNING_LOW_FIDELITY_FLIGHTSIMULATOR_H_
#define performance_assessment_SRC_TAW_DEFAULTMETHODS_FUELPLANNING_LOW_FIDELITY_FLIGHTSIMULATOR_H_

#include "../../aircraft.h"

#include <aerodynamics/aerodynamics.h>
#include <runtimeInfo/runtimeInfo.h>
#include <unitConversion/unitConversion.h>

#include <string>

class aircraft;
class aerodynamics;

using std::string;

/**\class flightSimulator class that contains the amounts of fuel required for each flight phase
*
*/
class flightSimulator {
 public:
    /**\brief Function to calculate the taxi fuel
    *\param taxi time (min)
    * \return taxi fuel (kg)
    */
    double calcTaxiFuel(double taxiTime);

    /**\brief Function to compute the necessary climb fuel
    *\param grossMass Aircraft gross mass (kg)
    *\param startAlt Altitude at the starting point of the climb segment (m)
    *\param endAlt Altitude at the end of the climb segment (m)
    *\param startMach Mach number at the start of the climb segment (-)
    *\param endMach Mach number at the end of the climb segment (-)
    *\param Config Flaps configuration (-)
    *\param Rating Engine rating (-)
    *\param bleedOff Current bleed air offtake (kg/s)
    *\param shaftOff Shaft power offtake (kW)
    *\param theAero Object of the aerodynamics class
    *\param climbFuel Fuel needed during climb phase (kg)
    *\param climbDistance Distance covered during climb phase (km)
    *\param climbTime Time spent during climb phase (s)
    */
    void   getClimbFuel(double grossMass,
                        double startAlt, double endAlt,
                        double startMach, double endMach,
                        const string &Config, const string &Rating,
                        double bleedOff, double shaftOff,
                        const aerodynamics &theAero,
                        double *climbFuel, double *climbDistance, double *climbTime);


    /**\brief Function to compute the fuel needed to accelerate from one Mach number to another Mach number
    *\param grossMass Aircraft gross mass (kg)
    *\param Altitude Aircraft altitude (m)
    *\param startMach Mach number at the starting point of the acceleration segment (-)
    *\param endMach Mach number at the ending point of the acceleration segment (-)
    *\param Config Flaps configuration (-)
    *\param Rating Engine rating (-)
    *\param bleedOff Current bleed air offtake (kg/s)
    *\param shaftOff Current shaft power offtake (kW)
    *\param theAero Object of the aerodynamics class
    *\param accFuel Fuel needed during acceleration phase (kg)
    *\param accDistance Distance covered during acceleration phase (km)
    *\param accTime Time spent during acceleration phase (s)
    */
    void getAccelerateFuel(double grossMass,
                           double Altitude,
                           double startMach, double endMach,
                           string Config, string Rating,
                           double bleedOff, double shaftOff,
                           aerodynamics theAero,
                           double *accFuel, double *accDistance, double *accTime);

    /**\brief Function checks if enough thrust is provided by the engine
    *\param grossMass Aircraft gross mass (kg)
    *\param Altitude Aircraft altitude (m)
    *\param Mach Aircraft Mach number (-)
    *\param Config Flaps configuration (-)
    *\param Rating Engine rating (-)
    *\param Derate Artificial constant to reduce the engine spool speed (between 0 and 1)
    *\param bleedOff Current bleed air offtake (kg/s)
    *\param shaftOff Current shaft power offtake (kW)
    *\param theAero Object of the aerodynamics class
    *\return True or false
    */
    bool checkForAcceleration(double grossMass,
                              double Altitude, double Mach,
                              string Config, string Rating, double Derate,
                              double bleedOff, double shaftOff,
                              aerodynamics theAero);

    /**\brief Function to compute the cruise fuel needed
    *\param grossMass Aircraft gross mass (kg)
    *\param Altitude Aircraft altitude (m)
    *\param Mach Aircraft Mach number (-)
    *\param Config Flaps configuration (-)
    *\param Rating Engine rating (-)
    *\param bleedOff Current bleed air offtake (kg/s)
    *\param shaftOff Current shaft power offtake (kW)
    *\param theAero Object of the aerodynamics class
    *\param cruiseFuel Fuel needed during the cruise phase (kg)
    *\param cruiseDistance Distance covered during cruise phase (km)
    *\param cruiseTime Time spent for the cruise phase (s)
    */
    void getCruiseFuel(double grossMass,
                       double Altitude, double Mach,
                       string Config, string Rating,
                       double bleedOff, double shaftOff,
                       aerodynamics theAero,
                       double *cruiseFuel, double cruiseDistance, double *cruiseTime);

    /**\brief Function checks if enough engine thrust is provided
    *\param grossMass Aircraft gross mass (kg)
    *\param Altitude Aircraft altitude (m)
    *\param Mach Aircraft Mach number (-)
    *\param Config Flap configuration (-)
    *\param Rating Engine rating (-)
    *\param bleedOff Current bleed air offtake (kg/s)
    *\param shaftOff Current shaft power offtake (kW)
    *\param theAero Object of the aerodynamics class
    *\return True or false
    */
    bool checkForSteadyLevelFlight(double grossMass,
                                   double Altitude, double Mach,
                                   string Config, string Rating,
                                   double bleedOff, double shaftOff,
                                   aerodynamics theAero);

    /**\brief Function to compute the descend fuel needed
     *\param startAltitude Altitude at the starting point of the descend segment (m)
     *\param endAltitude Altitude at the end of the descend segment (m)
     *\param startMach Mach number at the start of the descend segment (-)
     *\param endMach Mach number at the end of the descend segment (-)
     *\param bleedOff Current bleed air offtake (kg/s)
     *\param shaftOff Shaft power offtake (kW)
     *\param descendFuel Fuel needed during descend phase (kg)
     *\param descendDistance Distance covered during descend phase (km)
     *\param descendTime Time spent for the descend phase (s)
     */
    void getDescendFuel(double startAltitude, double endAltitude,
                        double startMach, double endMach,
                        double bleedOff, double shaftOff,
                        double *descendFuel, double *descendDistance, double *descendTime);

    /**\brief Constructor flightSimulator
     *\param theAircraft Address to the aircraft object
     */
    explicit flightSimulator(aircraft *theAircraft);

    /**\brief Destructor flightSimulator object
    *
    */
    virtual ~flightSimulator();

 private:
    aircraft *theAcftPt; /**< Pointer to the aircraft object */

    double climbAltSteps; /**< Climb altitude steps (m) */
    double cruiseSteps; /**< Cruise steps (km) */

    /**\brief Function to compute the climb fuel needed
    *\param massAC Aircraft mass (kg)
    *\param H0 Initial aircraft height (m)
    *\param H1 Final aircraft height (m)
    *\param Mach0 Initial aircraft Mach number (-)
    *\param Mach1 Final aircraft Mach number (-)
    *\param config Flap configuration (-)
    *\param rating Engine rating (-)
    *\param bleed Current bleed air (kg/s)
    *\param shaft Shaft power (kW)
    *\param AERO Object of the aerodynamics class
    *\param theTime Time spent in the climb phase (s)
    *\param theFuel Average fuel needed (kg)
    *\param theDistance Distance covered (km)
    */
    void calcClimbFuel(double massAC, double H0, double H1, double Mach0, double Mach1,
                       const string& config, string rating, double bleed, double shaft,
                       aerodynamics AERO,
                       double *theTime, double *theFuel, double *theDistance);

    /**\brief Function to compute the thrust needed for the given rate of climb (theROC)
    *\param massOfAC Aircraft mass (kg)
    *\param H Height (m)
    *\param Mach Mach number (-)
    *\param theROC Rate of climb (m/s)
    *\param theConfiguration Flap configuration (-)
    *\param AERO Object of the aerodynamics class
    *\return The required thrust for the given rate of climb (kN)
    */
    double calcThrustForROC(double massOfAC, double H, double Mach, double theROC, string theConfiguration, aerodynamics theAERO);
};

#endif  // performance_assessment_SRC_TAW_DEFAULTMETHODS_FUELPLANNING_LOW_FIDELITY_FLIGHTSIMULATOR_H_
