/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "performance_assessment_factory.h"
#include "performance_assessment_output.h"  // Resolving circular dependency

namespace taw {

performance_assessment_factory::performance_assessment_factory(const std::shared_ptr<RuntimeIO>& rtIOin)
    :
    rtIO(rtIOin),
    mySettings(rtIOin),
    mySettingsPt(std::make_shared<performance_assessment_settings>(mySettings)),
    myAcft(rtIO, mySettings),
    myAcftPt(std::make_shared<aircraft>(myAcft)),
    missionXML(aixml::openDocument(static_cast<std::string>(rtIO->acxml.at(
        "/aircraft_exchange_file/requirements_and_specifications/mission_files/study_mission_file/value")))) {
}

void performance_assessment_factory::initialize() {
    /* Get desired fidelity level of methods*/
    throwError<std::string>(__FILE__, __func__, __LINE__, "Module is not finished yet.");
    fidelityLevelFuelPlanning = static_cast<std::string>(rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/PayloadRangeDiagram/fidelityLevel"));  // PFADE ÜBERPRÜFEN
    fidelityLevelOptimizeMission = static_cast<std::string>(rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/PayloadRangeDiagram/fidelityLevel"));
    fidelityLevelPayloadRangeDiagram = static_cast<std::string>(rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/PayloadRangeDiagram/fidelityLevel"));
    fidelityLevelEnginePerformance = static_cast<std::string>(rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/EnginePerformance/fidelityLevel"));
    fidelityLevelFlightEnvelopePerformance = static_cast<std::string>(rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/FlightEnvelopePerformance/fidelityLevel"));
    fidelityLevelClimbPerformance = static_cast<std::string>(rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/ClimbPerformance/fidelityLevel"));
    fidelityLevelTakeOffPerformance = static_cast<std::string>(rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/TOPerformance/fidelityLevel"));
    fidelityLevelLandingPerformance = static_cast<std::string>(rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/LandingPerformance/fidelityLevel"));
    fidelityLevelVnDiagram = static_cast<std::string>(rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/VnDiagram/fidelityLevel"));
    fidelityLevelSARPerformance = static_cast<std::string>(rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/SARPerformance/fidelityLevel"));
    /* Other settings*/
    // doFuelPlanning = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/Modes/FuelPlanning");
    // doOptimizeMission = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/Mission/optimizeMissionProfile");
    // doPayloadRange              = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/PayloadRangeDiagram");
    // doEnginePerformance         = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/EnginePerformance");
    // doFlightEnvelopePerformance = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/FlightEnvelopePerformance");
    // overwriteInitialValues      = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/FlightEnvelopePerformance").getIntAttrib("OverwriteInitialValues");
    // doClimbPerformance          = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/ClimbPerformance");
    // doTOPerformance             = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/TOPerformance");
    // if (doTOPerformance) {
    //     doBLFLPerformance       = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/TOPerformance").getBoolAttrib("CalculateBLFLPerformance");
    // } else {
    //     doBLFLPerformance       = false;
    // }
    // doLandingPerformance        = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/LandingPerformance");
    // doVnDiagram                 = rtIO->moduleConfig.at("/ConfigFile/ProgramSettings/PerformanceChecks/VnDiagram");

    // doFuelPlanning = mySettingsPt->mySwitches.doFuelPlanning;




    /* Initialization of performance_assessment objects */
    /* Fuel Planning */
    if (fidelityLevelFuelPlanning == "high") {  // high fidelity fuel Planning
        // Custom implementation here, e.g.:
        // abstractFuelPlanningPt = std::make_shared<high::fuelPlanning>(someArgs);  // Create Object and puts it in interface class pointer
    } else if (fidelityLevelFuelPlanning == "low") {  // low fidelity fuelPlanning
        abstractFuelPlanningPt = std::make_shared<low::fuelPlanning>(&myAcft, &mySettings, rtIO);  // Create Object and puts it in interface class pointer
        abstractFuelPlanningPt->setFactory(shared_from_this());  // Enables Factory-Products to access the factory to check what fidelity levels other products are
        // TODO(Hansmann): Currently unused
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for FuelPlanning selected! Exiting"<< std::endl;
        exit(1);
    }
    /* MissionOptimization */
        if (fidelityLevelOptimizeMission == "high") {  // high fidelity optimizeMission
        // ...
    } else if (fidelityLevelFuelPlanning == "low") {  // low fidelity optimizeMission
        abstractOptimizeMissionProfilePt = std::make_shared<low::optimizeMissionProfile>(&myAcft);  // Create Object and puts it in interface class pointer
    } else {  // default case: low
        myRuntimeInfo->warn << "Unavailable fidelity level for 'MissionOptimization' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* PayloadRange */
    if (fidelityLevelPayloadRangeDiagram == "high") {
        // ...
    } else if (fidelityLevelPayloadRangeDiagram == "low") {
        abstractPayloadRangePt = std::make_shared<low::payloadRange>(&myAcft, mySettings);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'PayloadRange' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* EnginePerformance */
    if (fidelityLevelEnginePerformance == "high") {
        // ...
    } else if (fidelityLevelEnginePerformance == "low") {
        abstractEnginePerformancePt = std::make_shared<low::enginePerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'EnginePerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* ClimbPerformance */
    if (fidelityLevelClimbPerformance == "high") {
        // ...
    } else if (fidelityLevelClimbPerformance == "low") {
        abstractCeilingPerformancePt = std::make_shared<low::ceilingPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'ClimbPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* SARPerformance */
    if (fidelityLevelSARPerformance == "high") {
        // ...
    } else if (fidelityLevelSARPerformance == "low") {
        abstractSARPerformancePt = std::make_shared<low::SARPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'SARPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Take-Off performance */
    if (fidelityLevelTakeOffPerformance == "high") {
        // ...
    } else if (fidelityLevelTakeOffPerformance == "low") {
        abstractTakeOffPerformancePt = std::make_shared<low::takeOffPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'takeOffPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Landing Performance */
    if (fidelityLevelLandingPerformance == "high") {
        // ...
    } else if (fidelityLevelLandingPerformance == "low") {
        theLandingPerformance = std::make_shared<low::landingPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'landingPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Vn-Diagram */
    if (fidelityLevelVnDiagram == "high") {
        // ...
    } else if (fidelityLevelVnDiagram == "low") {
        theVnDiagram = std::make_shared<low::VnDiagram>(myAcftPt, mySettingsPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'VnDiagram' selected! Exiting"<< std::endl;
        exit(1);
    }
}  // initialize()

void performance_assessment_factory::run() {
/* Run the selected modules */
if (mySettingsPt->mySwitches.doFuelPlanning == true) {
    myRuntimeInfo->out << "  -> Fuel Planning" << std::endl;
    try {
        theFuelPlanning->doFuelPlanning();
    } catch (const int& errorCode) {
        throwError(__FILE__, __func__, __LINE__,
                "Error in fuel planning. Possibly thrust rating of engine could not be set correctly. See logLevel=2 (Info) for more details.");
    }
}
/* Optimize Mission */
if (mySettingsPt->optimizeMission == true) {
    try {
        theOptimizeMission->doOptimizeMission(theFuelPlanning, rtIO, mySettingsPt->designForMTOM);
        // theOptimizeMission->doOptimizeMission(&theFuelPlanning, rtIO, mySettingsPt->designForMTOM);
        } catch (const int& errorCode) {
        throwError(__FILE__, __func__, __LINE__,
                    "Error in mission optimization. Possibly thrust rating of engine could not be set correctly. See logLevel=2 (Info) for more details.");
        }
}

/* PayloadRange */
if (mySettingsPt->mySwitches.doPayloadRange == true) {
    if (!mySettingsPt->designForMTOM && mySettingsPt->useStudyMissionForAnalysis) {
        myRuntimeInfo->out << "Attention: Payload range diagram incorrect when MTOM Design off or StudyMission is used!" << std::endl;
        theFuelPlanning->doPayloadRangeOLD();
    } else {
        myRuntimeInfo->out << "      ... Payload Range using OLD method" << std::endl;
        theFuelPlanning->doPayloadRangeOLD();
        theFuelPlanning->doFuelPayloadRangeOLD();
    }
} else {
    theFuelPlanning->payloadRangeFromAcftXML();
}
if (mySettingsPt->mySwitches.doPayloadRange == true && mySettingsPt->mySwitches.doFuelPlanning == false) {
    myRuntimeInfo->out << "      ... Payload Range" << std::endl;
    try {
        abstractPayloadRangePt->calcPayloadRange();
    } catch (const int& errorCode) {
        throwError(__FILE__, __func__, __LINE__,
                    "Error in payload-range calculation. Possibly thrust rating of engine could not be set correctly. See logLevel=2 (Info) for more details.");
    }
}

/* EnginePerformance */
if (mySettingsPt->mySwitches.doEnginePerformance == true) {
    myRuntimeInfo->out << "  -> Engine Performance" << std::endl;
    abstractEnginePerformancePt->doEngineCalculation(mySettingsPt->enginePerformance.Rating);
}
/* Climb Performance */
if (mySettingsPt->mySwitches.doClimbPerformance == true) {
    myRuntimeInfo->out << "  -> Ceiling Performance" << std::endl;
    abstractCeilingPerformancePt->doCeilingCalculation();
    if (mySettingsPt->mySwitches.doFlightEnvelopePerformance == true) {
        myRuntimeInfo->out << "  -> Flight Envelopes" << std::endl;
        if (fabs(myAcftPt->MDive - 0) < ACCURACY_HIGH || fabs(myAcftPt->VDive - 0.) < ACCURACY_HIGH || mySettingsPt->mySwitches.overwriteInitialValues) {
            abstractFlightEnvelopesPt->doDiveSpeedLimits();
        }
        abstractFlightEnvelopesPt->doFlightEnvelope(abstractCeilingPerformancePt);
        myRuntimeInfo->out << "  -> SAR Performance" << std::endl;
        abstractSARPerformancePt->doSARCalculation(abstractFlightEnvelopesPt);
    }
}

/* Take-Off Performance*/
if (mySettingsPt->mySwitches.doTOPerformance == true) {
    myRuntimeInfo->out << "  -> Takeoff Performance" << std::endl;
    abstractTakeOffPerformancePt->doTakeOffPerformance(mySettingsPt->myField.runwaySlope,
            myAcftPt->frictionCoefficient,
            myAcftPt->brakingCoefficient,
            mySettingsPt->myField.headWind);
    if (mySettingsPt->mySwitches.doBLFLPerformance == true) { //BLFL performance cannot be calculated without TOperformance
    // Calculation of Balance Field Length
    myRuntimeInfo->out << "  -> Calculation of BLFL" << std::endl;
    abstractTakeOffPerformancePt->calcBLFL(mySettingsPt->myField.runwaySlope,
                                    myAcftPt->frictionCoefficient,
                                    myAcftPt->brakingCoefficient,
                                    mySettingsPt->myField.headWind);
    }
}
/* Landing Performance */
if (mySettingsPt->mySwitches.doLandingPerformance == true) {
    myRuntimeInfo->out << "  -> Landing Performance" << std::endl;
    theLandingPerformance->doLandingPerformance(mySettingsPt->myField.runwaySlope,
            myAcftPt->brakingCoefficient,
            mySettingsPt->myField.headWind);
}
/* Vn-Diagram */
if (mySettingsPt->mySwitches.doVnDiagram == true) {
    myRuntimeInfo->out << "  -> V-n Diagram" << std::endl;
    theVnDiagram->doVnDiagram();
}
}  // end of run()

void performance_assessment_factory::update() {
  // update
}

void performance_assessment_factory::report() {
    performance_assessment_output theOutput(shared_from_this(), shared_from_this()->myAcftPt);

    theOutput.generateHtmlBody();
    theOutput.writeXmlOutput();
    // (...)
}

void performance_assessment_factory::save() {
  // save data if necessary
}

}  // namespace taw
