/*  Copyright (C) 2023 Chair of Aircraft Design, Technical University Munich
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * @file test_default_turbofan.cpp
 * @author Sebastian Oberschwendtner (sebastian.oberschwendtner@tum.de)
 * @brief Test the default pylon geometry designer.
 * @version 3.0.0
 * @date 2023-12-20
 */

/* === Includes === */
#include <filesystem>
#include <gtest/gtest.h>
#include <aircraftGeometry2/geometry/builder.h>

/* Unit Under Test */
#include "pylon/default/default_pylon.h"

/* === Fixtures === */
class PylonDefault : public ::testing::Test
{
  protected:
    void SetUp() override
    {
        /* Setup the aircraft geometry */
        geom2::Polygon_2 poly = geom2::build::ellipse(1.0, 1.0).get_contour(false);
        nacelle.sections.emplace_back(poly);
        nacelle.sections.emplace_back(poly);
        nacelle.sections.back().origin = geom2::Point_3(0, 0, 2.0);
        nacelle.normal = geom2::Direction_3(1, 0, 0);
        wing.sections.emplace_back(poly);
        wing.sections.emplace_back(poly);
        wing.sections.at(0).set_chord_length(2.0);
        wing.sections.at(1).set_chord_length(1.0);
        wing.sections.at(1).origin = geom2::Point_3(0, 0, -5.0);
        wing.origin = geom2::Point_3(3, 0, 2);
        wing.normal = geom2::Direction_3(0, -1, 0);
        wing.is_symmetric = true;
        aircraft->wings.emplace_back(wing);
        aircraft->fuselages.emplace_back(nacelle);
    }
    std::shared_ptr<node> config = aixml::openDocument(std::filesystem::path{CMAKE_TEST_CONFIG_FILE});
    geom2::MultisectionSurface<geom2::PolygonSection> nacelle;
    geom2::MultisectionSurface<geom2::AirfoilSection> wing;
    std::shared_ptr<geometry::Aircraft> aircraft = std::make_shared<geometry::Aircraft>();
};

/* === Tests === */
/**
 * @brief Test designing the pylon of a kerosene turbofan engine when the engine is under the wing.
 */
TEST_F(PylonDefault, DesignTurbofanKeroseneUnderWing)
{
    /* Create the pylon designer */
    geometry::pylon::Default designer{config, aircraft, {CMAKE_TEST_STUBS_DIR}};
    designer.initialize();

    /* Setup the engine */
    Turbofan<EnergyCarrier::Kerosene> engine{0, {Parent::Wing, Lateral::Right, Longitudinal::Front, Vertical::Under}};
    engine.set_dimension({0.8, 0.8, 1.5});
    engine.set_fan({0.9});
    engine.set_position({2.0, 3.0, -1.0});
    engine.set_nacelle(nacelle);

    /* Design the right engine */
    designer(engine);

    /* Check the result */
    auto result = engine.pylon();
    ASSERT_EQ(result.sections.size(), 2);
    EXPECT_EQ(result.normal, geom2::Direction_3(0, 0, -1));
    EXPECT_EQ(result.origin, geom2::Point_3(0, 0, 0));
    EXPECT_EQ(result.sections.at(0).name, "pylon");
    EXPECT_NEAR(result.sections.at(0).origin.x(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(0).origin.y(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(0).origin.z(), -0.5, 1e-3);
    EXPECT_NEAR(result.sections.at(0).get_chord_length(), 1.5, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.x(), 1.0, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.y(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.z(), -3.0, 1e-3);
    EXPECT_NEAR(result.sections.at(1).get_chord_length(), 1.5, 1e-3);

    /* Design the left engine */
    engine.set_position({2.0, -3.0, -1.0});
    designer(engine);

    /* Check the result */
    result = engine.pylon();
    ASSERT_EQ(result.sections.size(), 2);
    EXPECT_EQ(result.normal, geom2::Direction_3(0, 0, -1));
    EXPECT_EQ(result.origin, geom2::Point_3(0, 0, 0));
    EXPECT_NEAR(result.sections.at(0).origin.x(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(0).origin.y(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(0).origin.z(), -0.5, 1e-3);
    EXPECT_NEAR(result.sections.at(0).get_chord_length(), 1.5, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.x(), 1.0, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.y(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.z(), -3.0, 1e-3);
    EXPECT_NEAR(result.sections.at(1).get_chord_length(), 1.5, 1e-3);
}
/**
 * @brief Test designing the pylon of a kerosene turbofan engine when the engine is on the fuselage side.
 */
TEST_F(PylonDefault, DesignTurbofanKeroseneMidFuselage)
{
    /* Create the pylon designer */
    geometry::pylon::Default designer{config, aircraft, {CMAKE_TEST_STUBS_DIR}};
    designer.initialize();

    /* Setup the engine */
    Turbofan<EnergyCarrier::Kerosene> engine{0, {Parent::Fuselage, Lateral::Right, Longitudinal::Rear, Vertical::Mid}};
    engine.set_dimension({0.8, 0.8, 1.5});
    engine.set_fan({0.9});
    engine.set_position({1.0, 2.0, 0.1});
    nacelle.sections.at(0).set_width(1.5);
    nacelle.sections.at(1).set_width(1.5);
    engine.set_nacelle(nacelle);

    /* Design the right engine */
    designer(engine);

    /* Check the result */
    auto result = engine.pylon();
    ASSERT_EQ(result.sections.size(), 2);
    EXPECT_EQ(result.normal, geom2::Direction_3(0, 1, 0));
    EXPECT_EQ(result.origin, geom2::Point_3(0, 0, 0));
    EXPECT_EQ(result.name, "pylon_0");
    EXPECT_NEAR(result.sections.at(0).origin.x(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(0).origin.y(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(0).origin.z(), -0.75, 1e-3);
    EXPECT_NEAR(result.sections.at(0).get_chord_length(), 1.5, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.x(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.y(), 0.1, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.z(), -1.5, 1e-3);
    EXPECT_NEAR(result.sections.at(1).get_chord_length(), 1.5, 1e-3);

    /* Design the left engine */
    engine.set_position({1.0, -2.0, 0.1});
    designer(engine);

    /* Check the result */
    result = engine.pylon();
    ASSERT_EQ(result.sections.size(), 2);
    EXPECT_EQ(result.normal, geom2::Direction_3(0, -1, 0));
    EXPECT_EQ(result.origin, geom2::Point_3(0, 0, 0));
    EXPECT_NEAR(result.sections.at(0).origin.x(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(0).origin.y(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(0).origin.z(), -0.75, 1e-3);
    EXPECT_NEAR(result.sections.at(0).get_chord_length(), 1.5, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.x(), 0.0, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.y(), -0.1, 1e-3);
    EXPECT_NEAR(result.sections.at(1).origin.z(), -1.5, 1e-3);
    EXPECT_NEAR(result.sections.at(1).get_chord_length(), 1.5, 1e-3);
}

/**
 * @brief When the parent location is Vertical::In no pylon is necessary.
 */
TEST_F(PylonDefault, DesignTurbofanKeroseneIn)
{
    /* Create the pylon designer */
    geometry::pylon::Default designer{config, aircraft, {CMAKE_TEST_STUBS_DIR}};
    designer.initialize();

    /* Setup the engine */
    Turbofan<EnergyCarrier::Kerosene> engine{0, {Parent::Fuselage, Lateral::Right, Longitudinal::Rear, Vertical::In}};
    engine.set_dimension({0.8, 0.8, 1.5});
    engine.set_fan({0.9});
    engine.set_position({1.0, 2.0, 0.1});
    nacelle.sections.at(0).set_width(1.5);
    nacelle.sections.at(1).set_width(1.5);
    engine.set_nacelle(nacelle);

    /* Design the right engine */
    designer(engine);

    /* Check the result */
    auto result = engine.pylon();
    ASSERT_TRUE(result.sections.empty());
}

/**
 * @brief Test designing unsupported pylons.
 */
TEST_F(PylonDefault, DesignTurbofanKeroseneUnsupported)
{
    /* Create the pylon designer */
    geometry::pylon::Default designer{config, aircraft, {CMAKE_TEST_STUBS_DIR}};
    designer.initialize();

    /* Setup the engine */
    Turbofan<EnergyCarrier::Kerosene> engine{0, {Parent::Fuselage, Lateral::Mid, Longitudinal::Rear, Vertical::Over}};
    engine.set_dimension({0.8, 0.8, 1.5});
    engine.set_fan({0.9});
    engine.set_position({1.0, 2.0, 0.1});
    nacelle.sections.at(0).set_width(1.5);
    nacelle.sections.at(1).set_width(1.5);
    engine.set_nacelle(nacelle);

    /* Design the right engine */
    ASSERT_THROW(designer(engine), std::runtime_error);
}

/**
 * @brief Test designing a pylon when no aircraft geometry exists yet.
 */
TEST_F(PylonDefault, DesignTurbofanKeroseneNoGeometry)
{
    /* Create the pylon designer */
    geometry::pylon::Default designer{config, nullptr, {CMAKE_TEST_STUBS_DIR}};
    designer.initialize();

    /* Setup the engine */
    Turbofan<EnergyCarrier::Kerosene> engine{0, {Parent::Fuselage, Lateral::Right, Longitudinal::Rear, Vertical::Mid}};
    engine.set_dimension({0.8, 0.8, 1.5});
    engine.set_fan({0.9});
    engine.set_position({1.0, 2.0, 0.1});
    nacelle.sections.at(0).set_width(1.5);
    nacelle.sections.at(1).set_width(1.5);
    engine.set_nacelle(nacelle);

    /* Design the right engine */
    ASSERT_NO_THROW(designer(engine));
    EXPECT_TRUE(engine.pylon().sections.empty());
}
