/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA24_H_
#define SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA24_H_

/* The following statements import necessary libraries */
#include <string>                                                   // Import c++ standard library for using string
#include <vector>                                                   // Import c++ standard library for using vector
#include <memory>

/* The follwing statements import necessary header files */
#include <svl/SVL.h>                                // Import of header file from src folder for SVL
#include "../aircraftSystem.h"                      // Import of header file from src folder for aircraftSystem
#include "../powerProfile.h"                        // Import of header file from src folder for powerProfile
#include "../systemsIOData.h"                       // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                    // aircraftSystem.h because that would lead to circular dependencies
#include <aircraftGeometry2/processing/measure.h>
/* Forward declarations */
class fuselage;
class geometricalFunctions;
class nacelle;

/** \class conventionalATA24
 * \brief class that contains all properties of the electric system (ATA24)
 */
class conventionalATA24 : public aircraftSystem {
 public:
    double cableMass; /**< [kg] mass of the cables of the electric system */
    double otherMass; /**< [kg] other (not cables) mass of the electric system */
    double efficiencyFactorElectricSystem; /**< [-] efficiency factor of the electric system */
    double maxRelPower; /**< [-] Ratio of maximum permanent power and maximum required power of all generators */
    double specificCableMass; /** < [kg/m] specific cable mass */

    int numberOfElectricCircuits;/**< [-] Number of electric circuits*/
    double percentageElectricPower/**< [-] Percentage of electrical energy generated by the engine */;
    double lengthCable; /** < [m] cable length */

    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;
    void checkUserInput() override;
    void calculateConductingPower(const powerProfile& totalBleed,
                                  const std::vector<powerProfile>& totalElectric,
                                  const std::vector<powerProfile>& totalHydraulic) override;
    void calculateConductorMass(const powerProfile& sourceBleed,
                                const std::vector<powerProfile>& sourceElectric,
                                const std::vector<powerProfile>& sourceHydraulic) override;

    void getPower(std::vector<powerProfile> theElectric);

    double getMass(double maxPower);
    Vec3 getRefPoint();

    conventionalATA24(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO);
    virtual ~conventionalATA24();

 private:
     /** \brief calculation of the length of the electric cables based on the engine position
      * \param Fuselage Fuselage geometry
      * \param Wing Wing geometry
      * \param Nacelles Vector containing Nacelle geometries
     */
     double getCableLength(const geom2::MultisectionSurface<geom2::PolygonSection>& Fuselage,
                          const systemsIOData::Wing& Wing,
                          const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& Nacelles);
};

#endif // SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA24_H_
