/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA49_H_
#define SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA49_H_

/* The following statements import necessary libraries */
#include <string>                                                   // Import c++ standard library for using string
#include <vector>                                                   // Import c++ standard library for using vector
#include <memory>                                                   // Import c++ standard library for using memory

/* The following statements import necessary header files */
#include <svl/SVL.h>                                                // Import of header file from src folder for SVL
#include "../aircraftSystem.h"                                      // Import of header file from src folder for aircraft system
#include "../powerProfile.h"                                        // Import of header file from src folder for power profile
#include "../systemsIOData.h"                                       // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                                    // aircraftSystem.h because that would lead to circular dependencies

/** \class conventionalATA49
 * \brief class that contains all properties of the auxiliary power unit (ATA49)
 */
class conventionalATA49: public aircraftSystem {
 public:
    double posXFuselage;/**< [-] Relative position of the APU in reference to the total fuselage length **/

    double percentageBleedAir; /**< [-] Percentage of Bleed Air generated by the APU (during mission) **/
    double percentageElectricPower; /**< [-] Percentage of Electric Power generated by the APU (during mission)**/
    double percentageHydraulicPower; /**< [-] Percentage of Hydraulic Power generated by the APU (during mission)**/

    double designBleedAir; /**< [-] Percentage of Bleed Air for which the APU is designed. **/
    double designElectricPower; /**< [-] percentage of electrical energy for which the APU is designed **/
    double designHydraulicPower; /**< [-] Percentage of hydraulic energy for which the APU is designed. **/

    double efficiencyFactorBleedAir; /**< [-] Efficiency of Bleed Air extraction **/

    double installationFactor;/**< [-] Installation factor for attached parts such as fire protection, noise protection, etc. [Tor13] p. 288 2.0-2.5*/

    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;

    void calculateSourcePower(const powerProfile& sourceBleed,
                              const std::vector<powerProfile>& sourceElectric,
                              const std::vector<powerProfile>& sourceHydraulic) override;
    void calculateSourceMass(const powerProfile& sourceBleed,
                             const std::vector<powerProfile>& sourceElectric,
                             const std::vector<powerProfile>& sourceHydraulic) override;

    /** \brief calculates the power of the APU (ATA49)
     * \param theBleed powerProfile
     * \param theElectric vector<powerProfile>
     * \param theHydraulic vector<powerProfile>
     * \return void
     */
    void getPower(powerProfile theBleed, std::vector<powerProfile> theElectric, std::vector<powerProfile> theHydraulic);

    /** \brief calculates the mass of the APU (ATA49)
     * \param theBleed powerProfile
     * \param theElectric vector<powerProfile>
     * \param theHydraulic vector<powerProfile>
     * \return double: mass of the APU [kg]
     */
    double getMass(const powerProfile& theBleed,
                   const std::vector<powerProfile>& theElectric,
                   const std::vector<powerProfile>& theHydraulic);

    /** \brief calculates the reference point of the APU (ATA49)
     * \return Vec3: APU reference point in global x, y, z-coordinates [m]
     */
    Vec3 getRefPoint();

    /** \brief Constructor
     * \param systemName string
     * \param systemIsOperating bool: switch whether the system is operating (1) or not(0)
     * \param data pointer to IOData
     */
    conventionalATA49(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO);
    virtual ~conventionalATA49();

 private:
    /** \brief negates all values in the power profile
     * \param Power powerProfile*: pointer to the power profile
     * \return void
     */
    void powerNegation(powerProfile *Power);
};

#endif // SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA49_H_
