/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "electricATA30.h"
#include <unordered_map>
#include <atmosphere/atmosphere.h>
#include <unitConversion/constants.h>
#include <unitConversion/unitConversion.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>
#include "../aircraftSystems/wingAntiIcing/liquidWaterContent.h"
#include "../aircraftSystems/wingAntiIcing/wingAntiIcing.h"
#include "../geometricalFunctions.h"

electricATA30::electricATA30(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO) {
    ATA = "ATA-30";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

electricATA30::~electricATA30() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void electricATA30::readParameters(const node& config) {
    std::string unit;
    std::string ATA30config = "/module_configuration_file/program_settings/systems_constants/ice_rain_protection";
    unit = std::string(config.at(ATA30config + "/top_operating_altitude/unit"));
    if (unit == "ft") {
        operatingAltitude   = config.at(ATA30config + "/top_operating_altitude/value");
    } else if (unit == "m") {
        operatingAltitude = convertUnit(METER, FOOT,
                                                static_cast<double>(config.at(ATA30config + "/top_operating_altitude/value")));
    } else if (unit == "km") {
        operatingAltitude   = convertUnit(KILO, METER, FOOT,
                                                  static_cast<double>(config.at(ATA30config + "/top_operating_altitude/value")));
    } else {
        myRuntimeInfo->err << "The unit of the IceRain.operatingAltitude is unknown: " << unit << std::endl;
        exit(1);
    }
    data_->configuration.checkForValidity(operatingAltitude, 0.0, 45000., "IceRain.operatingAltitude");
    engineAntiIce   = config.at(ATA30config + "/engine_anti_ice/value");
    data_->configuration.checkForValidity(engineAntiIce, 0.0, 5.0, "IceRain.engineAntiIce");
    elecPowerDeparture  = config.at(ATA30config + "/elec_power_departure/value");
    data_->configuration.checkForValidity(elecPowerDeparture, 0.0, 300000., "IceRain.elecPowerDeparture");
    elecPowerCruise  = config.at(ATA30config + "/elec_power_cruise/value");
    data_->configuration.checkForValidity(elecPowerCruise, 0.0, 300000., "elecPowerCruise");
    elecPowerApproach  = config.at(ATA30config + "/elec_power_approach/value");
    data_->configuration.checkForValidity(elecPowerApproach, 0.0, 300000., "elecPowerApproach");
    elecPowerLand  = config.at(ATA30config + "/elec_power_land/value");
    data_->configuration.checkForValidity(elecPowerLand, 0.0, 300000., "elecPowerLand");
    wingAntiIcingSwitch = config.at(ATA30config + "/switch_off_wing_anti_icing/value");
    wingSkinThickness = config.at(ATA30config + "/skin_thickness/value");
    data_->configuration.checkForValidity(wingSkinThickness, 0.000001, 1., "skinThickness");
    wingRelSpanAntiIcingStart_str = std::string(config.at(ATA30config + "/rel_span_anti_icing_start/value"));
    wingSkinHeatConductivity = config.at(ATA30config + "/heat_conductivity_wing/value");
    data_->configuration.checkForValidity(wingSkinHeatConductivity, 50, 500., "heatConductivity");
    efficiencyOfElectroThermalSystem = config.at(ATA30config + "/efficiency_factor_electro_thermic_anti_icing/value");
    data_->configuration.checkForValidity(efficiencyOfElectroThermalSystem, 0.1, 1.0, "efficiencyOfElectroThermalSystem");
    designDropDiameter = config.at(ATA30config + "/drop_diameter/value");
    data_->configuration.checkForValidity(designDropDiameter, 15., 40., "dropDiameter");
    numberOfPowerSources = config.at(ATA30config + "/shaft_power_sources/number_of_power_sources/value");
    double operationFactor_gesamt(0.);
    for (int i(0); i < numberOfPowerSources; i++) {
        powerSources.push_back(PowerSource());
        powerSources.back().type =
            std::string(config.at(ATA30config + "/shaft_power_sources/power_source@" + num2Str(i) + "/type/value"));
        powerSources.back().sourceID =
            config.at(ATA30config + "/shaft_power_sources/power_source@" + num2Str(i) + "/source_ID/value");
        powerSources.back().operationFactor =
            config.at(ATA30config + "/shaft_power_sources/power_source@" + num2Str(i) + "/operation_factor/value");
        operationFactor_gesamt += Rounding(powerSources.back().operationFactor, 10);
    }
    if (operationFactor_gesamt != 1.) {
        myRuntimeInfo->err << "The work distribution of the \"PowerSources\" for Ice and Rain Protection does not add up to 100% ("
                           << operationFactor_gesamt * 100 << "%). adjust in config. Abort!"
                           << std::endl;
        exit(1);
    }
    percentageOfOME = checkBoundaries(config.at(ATA30config + "/percentageOfOME"), 0, true, 1, true);
}

/* Updating system specific parameters in aircraft xml file */
void electricATA30::updateXML() {
    updateMass("de_icing", Mass, ATA);
}

void electricATA30::calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
        const std::vector<powerProfile>& allPreviousSystemsHydraulic,
        const std::vector<powerProfile>& allPreviousSystemsElectric,
        const powerProfile& allPreviousSystemsHeat) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(allPreviousSystemsBleed);
    UNUSED(allPreviousSystemsHydraulic);
    UNUSED(allPreviousSystemsElectric);
    UNUSED(allPreviousSystemsHeat);
    /* Calculate Systems Power */
    initializePowerProfile();
    if (Operating == true) {
        getPower();
    }
    calculateMaxMissionPower();
}
void electricATA30::calculateSinkMass(const powerProfile& sourceBleed,
                                      const std::vector<powerProfile>& sourceElectric,
                                      const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceHydraulic);
    UNUSED(sourceElectric);
    /* Calculate Systems Mass */
    Mass = data_->configuration.SystemsFactor * getMass();
    refPoint = getRefPoint();
}
void electricATA30::getPower() {
    //Initialisieren der Hilfsvariablen
    int nrEngines(data_->data.Propulsion.numberEngines);
    //Wing Anti-Icing
    std::unordered_map<std::string, double> theRelSpanPositionsAntiIcing;
    // Get relative start and end positions for ice protection system
    geometricalFunctions myGeoFunctions;
    theRelSpanPositionsAntiIcing = myGeoFunctions.getRelSpanAntiIcing(data_->theWing.at(0), wingRelSpanAntiIcingStart_str);

    wingAntiIcing theWingAntiIcing(wingAntiIcing(data_->theWing.at(0).aerodynamicSurface, wingSkinThickness, theRelSpanPositionsAntiIcing["start"],
                                                  theRelSpanPositionsAntiIcing["end"], wingSkinHeatConductivity, data_->theWing.at(0).frontSparPos));
    liquidWaterContent theLiquidWaterContent;
    double designElectricPower(0.), designAltitude(TRANSITION_ALTITUDE);
    double theLWC = theLiquidWaterContent.getLiquidWaterContentForMaximumIcing(Rounding(convertTemperature(KELVIN, CELCIUS,
                                                                                                           data_->mission.myISA.getTemperature(designAltitude)), 2),
                                                                                        designDropDiameter, 5.5);
    // calibratedAirspeedUpToTransAltitude is used instead of CAS over transition altitude,
    // because TAS approaching velocities as high as VMO would not be used as design point for IPS system (not designed for worst conditions for each parameter)
    double trueAirspeed_SI(convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, designAltitude, data_->mission.myISA, convertUnit(KNOTS, METERPERSECOND,
                                                                                                             data_->calibratedAirspeedUpToTransAltitude)));
    double designMach = convertUnit(TRUEAIRSPEED, MACH, designAltitude, data_->mission.myISA, trueAirspeed_SI);
    theWingAntiIcing.setPowerConsumption(designMach, designAltitude, &data_->mission.myISA, theLWC);
    designElectricPower = theWingAntiIcing.getElectricConsumption(efficiencyOfElectroThermalSystem);
    bleedAir.designPower = 0.0;
    for (size_t n(0); n < Electric.size(); n++) {
        Electric.at(n).designPower = 0.0;
        double designPower_tmp(0.);
        if (designPower_tmp < elecPowerDeparture) {
            designPower_tmp = elecPowerDeparture;
        } else if (designPower_tmp < elecPowerCruise) {
            designPower_tmp = elecPowerCruise;
        } else if (designPower_tmp < elecPowerApproach) {
            designPower_tmp = elecPowerApproach;
        }
        designPower_tmp += designElectricPower;
        for (size_t j(0); j < powerSources.size(); j++) {
            double operationFactor_tmp(powerSources.at(j).operationFactor);
            if (powerSources.at(j).type == "Electric" && powerSources.at(j).sourceID == static_cast<int>(n + 1)) {
                Electric.at(n).designPower += designPower_tmp * operationFactor_tmp;
            }
        }
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        bleedAir.missionPower.at(i).baseLoad = 0.0;
        heatLoad.missionPower.at(i).baseLoad = 0.0;
        /*Loop over hydraulic circuits*/
        for (size_t n(0); n < Hydraulic.size(); n++) {
            Hydraulic.at(n).missionPower.at(i).baseLoad = 0.0;
        }
        /*Loop over electric circuits**/
        for (size_t n(0); n < Electric.size(); n++) {
            // Demand of the wing anti-icing
            Electric.at(n).missionPower.at(i).baseLoad = 0.0;
            double baseLoad_tmp(0.);
            if (data_->mission.Steps.at(i).Point.Altitude < operatingAltitude && wingAntiIcingSwitch == false) {
                baseLoad_tmp += designElectricPower;
            } else {
                baseLoad_tmp = 0.0;
            }
            /* Electric power demand */
            if (data_->mission.Steps.at(i).missionPhase == "Departure") {
                baseLoad_tmp += elecPowerDeparture;
            } else if (data_->mission.Steps.at(i).missionPhase == "Cruise") {
                baseLoad_tmp += elecPowerCruise;
            } else if (data_->mission.Steps.at(i).missionPhase == "Approach") {
                baseLoad_tmp += elecPowerApproach;
            } else {
                myRuntimeInfo->err << "Mission segment to determine the electric demand of the ice and rain protection system ins unknown" << std::endl;
                myRuntimeInfo->err << data_->mission.Steps.at(i).missionPhase << std::endl;
                myRuntimeInfo->err << "Abort!" << std::endl;
                exit(1);
            }
            for (size_t j(0); j < powerSources.size(); j++) {
                double operationFactor_tmp(powerSources.at(j).operationFactor);
                if (powerSources.at(j).type == "Electric" && powerSources.at(j).sourceID == static_cast<int>(n + 1)) {
                    Electric.at(n).missionPower.at(i).baseLoad += baseLoad_tmp * operationFactor_tmp;
                }
            }
        }
        // Demand of the engines
        bleedAir.missionPower.at(i).baseLoad += static_cast<double>(nrEngines) * engineAntiIce;
    }
}

double electricATA30::getMass() {
    return percentageOfOME * data_->data.Mass.OME;
}

Vec3 electricATA30::getRefPoint() {
    /* Assumption: reference point at 45% fuselage length **/
    double fuselage_length = geom2::measure::length(data_->theFuselage.back());
    return Vec3(data_->theFuselage.back().origin[0] + 0.45 * fuselage_length, data_->theFuselage.back().origin[1], data_->theFuselage.back().origin[2]);
}
