/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "liquidWaterContent.h"

#include <cmath>

#include <runtimeInfo/runtimeInfo.h>

liquidWaterContent::liquidWaterContent() {
    //ctor
}

liquidWaterContent::~liquidWaterContent() {
    //dtor
}

// *** LWC according to CS-25 Ammd. 12 Appendix C Fig.1 ***
// ************************************************
double liquidWaterContent::getLiquidWaterContentForMaximumIcing(double TemperatureCelsius, double dropDiameterMikroMeter, double cloudExtensionNauticalMiles) {
    bool LWC_Calc(true);
    //Check of the temperature range
    if (TemperatureCelsius < -30. || TemperatureCelsius > 0.) {
        myRuntimeInfo->err << "Temperature outside the allowed range for determining the LWC according to CS-25 for max. continous icing conditions!" << std::endl;
        myRuntimeInfo->err << "Allowed range: -30 deg C < T < 0 deg C" << std::endl;
        myRuntimeInfo->err << "Temperature: T = " << TemperatureCelsius << " deg C" << std::endl;
//        myRuntimeInfo->err << "Abort!" << std::endl;
        LWC_Calc = false;
        //exit(1);
    }
    //Checking the water droplet diameter range
    if (dropDiameterMikroMeter < 15. || dropDiameterMikroMeter > 40.) {
        myRuntimeInfo->err << "Droplet diameter outside the allowed range for determining the LWC according to CS-25 for max. continous icing conditions!" << std::endl;
        myRuntimeInfo->err << "Allowed range: 15 micro m < d_drop < 40 micro m" << std::endl;
        myRuntimeInfo->err << "Diameter: d_drop = " << dropDiameterMikroMeter << "micro meter" << std::endl;
//        myRuntimeInfo->err << "Abort!" << std::endl;
        LWC_Calc = false;
        //exit(1);
    }
    //check horizontal cloud extension
    if (cloudExtensionNauticalMiles < 5. || cloudExtensionNauticalMiles > 310.) {
        myRuntimeInfo->err << "Horizontal cloud extenison outside the allowed range to determine the LWC according to CS-25 for max. continous icing conditions!" << std::endl;
        myRuntimeInfo->err << "Allowed range: 5 NM < d_cloud < 310 NM" << std::endl;
        myRuntimeInfo->err << "Cloud extenison: d_cloud = " << cloudExtensionNauticalMiles << " NM" << std::endl;
//        myRuntimeInfo->err << "Abort!" << std::endl;
        LWC_Calc = false;
        //exit(1);
    }
    double X0(0.);
    double X1(0.);
    double Y0(0.);
    double Y1(0.);
    if (LWC_Calc == true) {
        if (TemperatureCelsius <= 0. && TemperatureCelsius > -10.) {
            X0 = 0.;
            X1 = -10.;
            Y0 = -1.254 * pow(dropDiameterMikroMeter, 0.2401)  + 3.198;
            Y1 = 7.627  * pow(dropDiameterMikroMeter, -0.7699) - 0.3432;
        } else if (TemperatureCelsius <= -10. && TemperatureCelsius > -20.) {
            X0 = -10.;
            X1 = -20.;
            Y0 = 7.627  * pow(dropDiameterMikroMeter, -0.7699) - 0.3432;
            Y1 = 4.322  * pow(dropDiameterMikroMeter, -0.8501) - 0.1293;
        } else if (TemperatureCelsius <= -20. && TemperatureCelsius >= -30.) {
            X0 = -20.;
            X1 = -30.;
            Y0 = 4.322  * pow(dropDiameterMikroMeter, -0.8501) - 0.1293;
            Y1 = 9.976 * pow(dropDiameterMikroMeter, -1.388)   - 0.02726;
        }
    }
    //LWC multiplicator according to CS-25 Ammd. 12 Appendix C Fig.1
    double f_cloud = -0.27622 * log(cloudExtensionNauticalMiles) + 1.7846;
    if (LWC_Calc == true) {
        return f_cloud * ((Y1 - Y0) / (X1 - X0) * (TemperatureCelsius - X0) + Y0);
    } else {
        return 0.;
    }
}
// *** LWC according to CS-25 Ammd. 12 Appendix C Fig.4 ***
// ************************************************
double liquidWaterContent::getLiquidWaterContentForIntermittentIcing(double TemperatureCelsius, double dropDiameterMikroMeter, double cloudExtensionNauticalMiles) {
    //Checking the temperature range
    if (TemperatureCelsius < -30. || TemperatureCelsius > 0.) {
        myRuntimeInfo->err << "Temperature outside the allowed range for determining the LWC according to CS-25 for intermittent icing conditions!" << std::endl;
        myRuntimeInfo->err << "allowed range: -30 deg C < T < 0 deg C" << std::endl;
        myRuntimeInfo->err << "Temperature: T = " << TemperatureCelsius << " deg C" << std::endl;
        myRuntimeInfo->err << "Abort!" << std::endl;
        exit(1);
    }
    //Checking the water droplet diameter range
    if (dropDiameterMikroMeter < 15. || dropDiameterMikroMeter > 50.) {
        myRuntimeInfo->err << "Droplet diameter outside the allowed range to determine the LWC according to CS-25 for intermittent icing conditions!" << std::endl;
        myRuntimeInfo->err << "Allowerd range: 15 micro m < d_drop < 50 mikro m" << std::endl;
        myRuntimeInfo->err << "Diameter: d_drop = " << dropDiameterMikroMeter << "micro meter" << std::endl;
        myRuntimeInfo->err << "Abort!" << std::endl;
        exit(1);
    }
    //check horizontal cloud extension
    if (cloudExtensionNauticalMiles < 0.2 || cloudExtensionNauticalMiles > 5.21) {
        myRuntimeInfo->err << "Horizontal cloud extension outside the allowed range to determine the LWC according to CS-25 for intermittent icing conditions!" << std::endl;
        myRuntimeInfo->err << "Allower range: 0.2 NM < d_cloud < 5.21 NM" << std::endl;
        myRuntimeInfo->err << "Cloud extension: d_cloud = " << cloudExtensionNauticalMiles << " NM" << std::endl;
        myRuntimeInfo->err << "Abort!" << std::endl;
        exit(1);
    }
    double X0(0.);
    double X1(0.);
    double Y0(0.);
    double Y1(0.);
    if (TemperatureCelsius <= 0. && TemperatureCelsius > -10.) {
        double p1, p2, p3, p4, p5, p6;
        X0 = 0.;
        X1 = -10.;
        p1 = 6.547e-007;
        p2 = -0.0001123;
        p3 = 0.00745;
        p4 = -0.2353;
        p5 = 3.394;
        p6 = -14.65;
        Y0 =   p1 * pow(dropDiameterMikroMeter, 5.)
               + p2 * pow(dropDiameterMikroMeter, 4.)
               + p3 * pow(dropDiameterMikroMeter, 3.)
               + p4 * pow(dropDiameterMikroMeter, 2.)
               + p5 * pow(dropDiameterMikroMeter, 1.) + p6;
        p1 = 5.742e-007;
        p2 = -0.0001003;
        p3 = 0.006774;
        p4 = -0.2176;
        p5 = 3.185;
        p6 = -14.18;
        Y1 =   p1 * pow(dropDiameterMikroMeter, 5.)
               + p2 * pow(dropDiameterMikroMeter, 4.)
               + p3 * pow(dropDiameterMikroMeter, 3.)
               + p4 * pow(dropDiameterMikroMeter, 2.)
               + p5 * pow(dropDiameterMikroMeter, 1.) + p6;
    } else if (TemperatureCelsius <= -10. && TemperatureCelsius > -20.) {
        double p1, p2, p3, p4, p5, p6;
        p1 = 5.742e-007;
        p2 = -0.0001003;
        p3 = 0.006774;
        p4 = -0.2176;
        p5 = 3.185;
        p6 = -14.18;
        Y0 =   p1 * pow(dropDiameterMikroMeter, 5.)
               + p2 * pow(dropDiameterMikroMeter, 4.)
               + p3 * pow(dropDiameterMikroMeter, 3.)
               + p4 * pow(dropDiameterMikroMeter, 2.)
               + p5 * pow(dropDiameterMikroMeter, 1.) + p6;
        p1 = 2.927e-007;
        p2 = -5.205e-005;
        p3 = 0.003584;
        p4 = -0.1171;
        p5 = 1.717;
        p6 = -6.9;
        Y1 =   p1 * pow(dropDiameterMikroMeter, 5.)
               + p2 * pow(dropDiameterMikroMeter, 4.)
               + p3 * pow(dropDiameterMikroMeter, 3.)
               + p4 * pow(dropDiameterMikroMeter, 2.)
               + p5 * pow(dropDiameterMikroMeter, 1.) + p6;
    } else if (TemperatureCelsius <= -20. && TemperatureCelsius >= -30.) {
        double p1, p2, p3, p4, p5, p6;
        p1 = 2.927e-007;
        p2 = -5.205e-005;
        p3 = 0.003584;
        p4 = -0.1171;
        p5 = 1.717;
        p6 = -6.9;
        Y0 =   p1 * pow(dropDiameterMikroMeter, 5.)
               + p2 * pow(dropDiameterMikroMeter, 4.)
               + p3 * pow(dropDiameterMikroMeter, 3.)
               + p4 * pow(dropDiameterMikroMeter, 2.)
               + p5 * pow(dropDiameterMikroMeter, 1.) + p6;
        p1 = 1.529e-007;
        p2 = -2.689e-005;
        p3 = 0.001849;
        p4 = -0.0608;
        p5 = 0.8978;
        p6 = -3.501;
        Y1 =   p1 * pow(dropDiameterMikroMeter, 5.)
               + p2 * pow(dropDiameterMikroMeter, 4.)
               + p3 * pow(dropDiameterMikroMeter, 3.)
               + p4 * pow(dropDiameterMikroMeter, 2.)
               + p5 * pow(dropDiameterMikroMeter, 1.) + p6;
    }
    //LWC Multiplikator nach CS-25 Ammd. 12 Appendix C Fig.6
    double f_cloud = -0.177 * log(cloudExtensionNauticalMiles) + 1.1615;
    return f_cloud * ((Y1 - Y0) / (X1 - X0) * (TemperatureCelsius - X0) + Y0);
}
// *** LWC according to CS-25 Ammd. 12 Appendix C Text ***
// ************************************************
double liquidWaterContent::getLiquidWaterContentForMaxTakeOff() {
    //double d_drop = 20.0; //micro meter
    //double Temp = -9.; //deg C
    double LWC = 0.35; //g/m3
    return LWC;
}
