/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "wingAntiIcing.h"

#include <numeric>

#include <atmosphere/atmosphere.h>
#include <runtimeInfo/runtimeInfo.h>
#include <unitConversion/unitConversion.h>
//#include <CGAL/Polygon_2.h>

wingAntiIcing::wingAntiIcing(const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing, double skinThickness, double wingRelSpanAntiIcingStart,
                             const double& wingRelSpanAntiIcingEnd, double theHeatConductivity, double frontSparPos)
    :
    totalExternalHeatFlux(0.),
    theSkinThickness(skinThickness),
    heatConductivity(theHeatConductivity),
    nrWingSegments(25) {
    /* Creating the individual wing segments */
    double wingSpan = geom2::measure::span(theWing);
    double segmentSpan(wingSpan / 2 * (wingRelSpanAntiIcingEnd - wingRelSpanAntiIcingStart) / static_cast<double>(nrWingSegments));
    for (int i = 0; i < nrWingSegments; i++) {
        //Initialize the segment
        theWingSegments.push_back(wingSegment());
        //Determination of the individual parameters
        theWingSegments.back().segSpan   = segmentSpan; //m
        theWingSegments.back().yPos      = (geom2::measure::span(theWing) / 2 * wingRelSpanAntiIcingStart
                                           + (static_cast<double>(i) + 0.5) * theWingSegments.back().segSpan); //m
                                           //check wether extrusion direction is negative
        // check if extrusion direction is negative and if so set y position accordingly
        bool negativeExtrusion = theWing.sections.front().origin.z() > theWing.sections.back().origin.z();
        if (negativeExtrusion) {
            theWingSegments.back().yPos = -theWingSegments.back().yPos;
        }
        theWingSegments.back().segThickness    = getMaxSegThickness(theWing, theWingSegments.back().yPos); //m
        theWingSegments.back().segThickness    = convertUnit(NOPREFIX, METER, NOPREFIX, FOOT,  theWingSegments.back().segThickness);  //ft
        theWingSegments.back().segFrontSpar    = frontSparPos * geom2::measure::chord(theWing, theWingSegments.back().yPos); //m
        theWingSegments.back().segFrontSpar    = convertUnit(NOPREFIX, METER, NOPREFIX, FOOT,  theWingSegments.back().segFrontSpar);  //ft
        theWingSegments.back().segIcingSurface = convertUnit(AREA, NOPREFIX, METER, NOPREFIX, FOOT, getIcingSurface(theWingSegments.back().yPos, theWingSegments.back().segSpan,
                                                                                                                    convertUnit(NOPREFIX, FOOT, NOPREFIX, METER,
                                                                                                                                theWingSegments.back().segFrontSpar),
                                                                                                                    theWing));//[ft2]
    }
    //ctor
}

wingAntiIcing::~wingAntiIcing() {
    //dtor
}

double wingAntiIcing::getMaxSegThickness(const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing, double yPos) {
    /* find sections before the yPos */
    auto section_index = geom2::detail::get_section_index(theWing.sections, yPos);
    size_t ID = section_index.value_or(-1); //read value from std::optional<size_t>
    /* get the maximum thickness of thes sections before and after the yPos */
    double first_thickness = geom2::measure::thickness_max(theWing.sections.at(ID));
    double second_thickness = geom2::measure::thickness_max(theWing.sections.at(ID + 1));
    /* linearly interpolate the maximum thickness between the sections */
    double interpolated_thickness = Interpolation(theWing.sections.at(ID).origin.z(), theWing.sections.at(ID + 1).origin.z(), first_thickness, second_thickness, yPos);
    return interpolated_thickness;
}

void wingAntiIcing::setPowerConsumption(double Mach, double Alt, atmosphere *theISA,
                                        double liquidWaterContent) {
    bool runCalc(false);
    //Pruefen ob nach SAE Vereisungsbedingungen vorliegen -> [Fre11]
    if (Alt <= convertUnit(NOPREFIX, FOOT, NOPREFIX, METER, 12000.) &&
            theISA->getTemperature(Alt) >= 243.15 && theISA->getTemperature(Alt) <= STANDARD_TEMPERATURE
            && liquidWaterContent > 0.) {
        runCalc = true;
    } else if (Alt >= convertUnit(NOPREFIX, FOOT, NOPREFIX, METER, 12000.) && Alt <= convertUnit(NOPREFIX, FOOT, NOPREFIX, METER, 22000.) &&
               theISA->getTemperature(Alt) >= 243.15 &&
               theISA->getTemperature(Alt) <= STANDARD_TEMPERATURE - 0.002 * convertUnit(NOPREFIX, METER, NOPREFIX, FOOT,
                                                                                         Alt - convertUnit(NOPREFIX, FOOT, NOPREFIX, METER, 12000.))
               && liquidWaterContent > 0.) {
        runCalc = true;
    } else {
        myRuntimeInfo->out << "There are no icing conditions!" << std::endl;
        runCalc = false;
    }
    externalHeatFlux.clear();
    TinnerSkin.clear();
    if (runCalc == true) {
        // Bestimmen der True Airspeed
        double TAS = convertUnit(NOPREFIX, METERPERSECOND, NOPREFIX, KNOTS,  Mach * theISA->getSpeedOfSound(Alt));  // kts
        double externalHeatTransferCoefficient(0.); /**< Thermal conductivity coefficient for heat transfer Profile surface -> air **/ //[Btu/(h * ft2 * Fahrenheit)]
        double Tskin = (theISA->getTemperature(Alt)); //[K]
        externalHeatFlux.clear();
        TinnerSkin.clear();
        //Determination of heat flows etc. for the individual wing segments
        for (size_t i = 0; i < theWingSegments.size(); i++) {
            // Determining the water quantity per wing segment
            double totalWaterCatch = getWaterCatch(theWingSegments.at(i).segThickness,
                                                   TAS,
                                                   liquidWaterContent);/**< Water quantity per wing segment, from [SAE04] page 12 Graph**/ //[lb/h]
            //Iteration for Tskin
            double tempTskin(0.);
            do {
                tempTskin = Tskin;
                // Determination of the thermal conduction coefficient Tair -> surface
                externalHeatTransferCoefficient = getExternalHeatTransferCoeff(theISA->getDensity(Alt), TAS, theWingSegments.at(i).segFrontSpar,
                                                  theISA->getTemperature(Alt), Tskin);
                // Determination of the saturated steam pressure at surface temperature
                double vapourPressureSkin = getVapourPressureSkin(totalWaterCatch, externalHeatTransferCoefficient,
                                     (theWingSegments.at(i).segIcingSurface /
                                      convertUnit(NOPREFIX, METER, NOPREFIX, FOOT, theWingSegments.at(i).segSpan)),
                                     theISA->getPressure(Alt));
                // Determining the surface temperature from the vapour pressure table
                Tskin = getskinTFromVapourPressure(convertUnit(NOPREFIX, PSI, NOPREFIX, PASCAL, vapourPressureSkin));
            } while (fabs(1. - tempTskin / Tskin) > 0.0001);
            double enthalpyVapourization = getEnthalpyVapourization(Tskin);/**< Evaporation enthalpy Water**/    //[Btu/lb]
            externalHeatFlux.push_back(convertUnit(NOPREFIX, METER, NOPREFIX, FOOT, theWingSegments.at(i).segSpan)
                                       * getHeatFluxSkinToAir(totalWaterCatch, externalHeatTransferCoefficient,
                                               theWingSegments.at(i).segIcingSurface / convertUnit(NOPREFIX, METER, NOPREFIX, FOOT, theWingSegments.at(i).segSpan),
                                               Tskin, theISA->getTemperature(Alt), Mach, theISA->getViscosity(Alt), enthalpyVapourization));
            //Source: Baehr and Stephan Heat and mass transfer
            TinnerSkin.push_back(((0.293 * externalHeatFlux.back()) * theSkinThickness)
                                 / (convertUnit(AREA, NOPREFIX, FOOT, NOPREFIX, METER, theWingSegments.at(i).segIcingSurface) * heatConductivity)
                                 + Tskin);
        }
        //conversion in kW and for both wing halves
        totalExternalHeatFlux = 2. * 0.293 / 1000. * accumulate(externalHeatFlux.begin(), externalHeatFlux.end(), 0.);
    } else {
        totalExternalHeatFlux = 0.;
    }
}

double wingAntiIcing::getBleedAirConsumption(double bleedAirTemp, double efficiencyBleed) {
    std::vector<double> bleedAirMassFlow; //kg/s
    if (static_cast<int>(TinnerSkin.size()) == 0 || static_cast<int>(externalHeatFlux.size()) == 0) {
        //Bleed Air laeuft nicht!!!
        bleedAirMassFlow.push_back(0.);
    } else {
        if (static_cast<int>(theWingSegments.size()) != static_cast<int>(TinnerSkin.size()) ||
                static_cast<int>(theWingSegments.size()) != static_cast<int>(externalHeatFlux.size())) {
            myRuntimeInfo->err << "Vector lengths are not equal! Calculation Bleed Air de-icing not possible! Abort!" << std::endl;
            exit(1);
        } else {
            for (size_t i = 0; i < theWingSegments.size(); i++) {
                bleedAirMassFlow.push_back(getMassFlow(TinnerSkin.at(i), bleedAirTemp, efficiencyBleed, 0.293 * externalHeatFlux.at(i)));
            }
        }
    }
    return  2. * accumulate(bleedAirMassFlow.begin(), bleedAirMassFlow.end(), 0.);
}

double wingAntiIcing::getElectricConsumption(double efficiencyElectroThermic) {
    std::vector<double> electricPower;
    if (static_cast<int>(externalHeatFlux.size()) == 0) {
        //Bleed Air does not run!!!
        electricPower.push_back(0.);
    } else {
        if (static_cast<int>(theWingSegments.size()) != static_cast<int>(externalHeatFlux.size())) {
            myRuntimeInfo->err << "Vector lengths are not equal! Calculation electrical de-icing not possible! Abort!" << std::endl;
            exit(1);
        } else {
            for (size_t i = 0; i < theWingSegments.size(); i++) {
                electricPower.push_back(0.293 * externalHeatFlux.back() / efficiencyElectroThermic);
            }
        }
    }
    return 2. * accumulate(electricPower.begin(), electricPower.end(), 0.);
}

double wingAntiIcing::getWaterCatch(double theThickness, double theTAS, double theWaterContent) {
    /** [SAE04] page 12 Graph **/
    double X0(0.);
    double X1(0.);
    double Y0(0.);
    double Y1(0.);
    if (theTAS > 550.) {
        myRuntimeInfo->err << "True Air Speed too high" << std::endl;
        myRuntimeInfo->err << "No Water Catch can be calculated. Abort!" << std::endl;
        exit(1);
    } else if (theTAS <= 550. && theTAS > 500.) {
        X0 = 550.;
        X1 = 500.;
        Y0 = 41.28 * pow(theThickness, 0.427);
        Y1 = 35.78 * pow(theThickness, 0.4164);
    } else if (theTAS <= 500. && theTAS > 450.) {
        X0 = 500.;
        X1 = 450.;
        Y0 = 35.78 * pow(theThickness, 0.4164);
        Y1 = 31.29 * pow(theThickness, 0.3897);
    } else if (theTAS <= 450. && theTAS > 400.) {
        X0 = 450.;
        X1 = 400.;
        Y0 = 31.29 * pow(theThickness, 0.3897);
        Y1 = 26.73 * pow(theThickness, 0.3700);
    } else if (theTAS <= 400. && theTAS > 350.) {
        X0 = 400.;
        X1 = 350.;
        Y0 = 26.73 * pow(theThickness, 0.3700);
        Y1 = 22.28 * pow(theThickness, 0.3369);
    } else if (theTAS <= 350. && theTAS > 300.) {
        X0 = 350.;
        X1 = 300.;
        Y0 = 22.28 * pow(theThickness, 0.3369);
        Y1 = 17.38 * pow(theThickness, 0.3388);
    } else if (theTAS <= 300. && theTAS > 250.) {
        X0 = 300.;
        X1 = 250.;
        Y0 = 17.38 * pow(theThickness, 0.3388);
        Y1 = 12.57 * pow(theThickness, 0.3471);
    } else if (theTAS <= 250. && theTAS > 200.) {
        X0 = 250.;
        X1 = 200.;
        Y0 = 12.57 * pow(theThickness, 0.3471);
        Y1 = 9.327 * pow(theThickness, 0.3141);
    } else if (theTAS <= 200. && theTAS > 150.) {
        X0 = 200.;
        X1 = 150.;
        Y0 = 9.327 * pow(theThickness, 0.3141);
        Y1 = 6.02 * pow(theThickness, 0.3136);
    } else if (theTAS <= 150. && theTAS >= 100.) {
        X0 = 150.;
        X1 = 100.;
        Y0 = 6.02 * pow(theThickness, 0.3136);
        Y1 = 3.175 * pow(theThickness, 0.2967);
    } else {
        myRuntimeInfo->err << "True Air Speed too low!" << std::endl;
        myRuntimeInfo->err << "No Water Catch can be calculated. Abort!" << std::endl;
        exit(1);
    }
    return Interpolation(X0, X1, Y0, Y1, theTAS) * theWaterContent;
}

double wingAntiIcing::getExternalHeatTransferCoeff(double theDensity, double theTAS, double Lx, double airT, double skinT) {
    //  [SAE04] page 17, eq (3F-9)
    return 0.64 * pow(0.5 * convertUnit(KELVIN, FAHRENHEIT, airT + skinT) + 459.67, 0.3)
           * pow(1.69 * convertUnit(KILO, GRAM, NOPREFIX, POUND, theDensity)
                 / convertUnit(VOLUME, NOPREFIX, METER, NOPREFIX, FOOT, 1.)
                 * theTAS, 0.8) / pow(Lx, 0.2);
}

double wingAntiIcing::getskinTFromVapourPressure(double theVapourPressureSkin) {
    // Dampfdruckteballe aus Lucas Thermodynamik
    return 273.15 + (26.63 * pow(theVapourPressureSkin, 0.1632) - 74.55);
}

double wingAntiIcing::getVapourPressureSkin(double theWaterCatch, double theHeatTransCoefSkin, double theSurface, double thePressure) {
    // [SAE04] page 21, eq (3F-13)
    return  pow(1. + 1.74 * theHeatTransCoefSkin * theSurface / theWaterCatch, -1.)
            * convertUnit(NOPREFIX, PASCAL, NOPREFIX, PSI, thePressure);
}

double wingAntiIcing::getEnthalpyVapourization(double theSkinTemp) {
    return (50.05 - 0.9298 * theSkinTemp / 1000. - 65.19 * pow(theSkinTemp / 1000., 2)) * 1000. / 18.02 * 0.4299;
}

double wingAntiIcing::getHeatFluxSkinToAir(double theWaterCatch, double theHeatTransferCoefficient, double theIcingSurafe,
        double theTskin, double theTair, double theMach, double dynamicViscosity, double theEnthalpyVapourization) {
    double prandtlNumber = dynamicViscosity * getHeatCapacityOfAir(theTair) / getHeatConductivityOfAir(theTair);
    double totalAirTemperature = theTair * (1. + sqrt(prandtlNumber) * (1.4 - 1.) / 2. * pow(theMach, 2.));
    //[SAE04]
    return theHeatTransferCoefficient * theIcingSurafe * (convertUnit(NOPREFIX, KELVIN, NOPREFIX, FAHRENHEIT, theTskin)
            - convertUnit(NOPREFIX, KELVIN, NOPREFIX, FAHRENHEIT, totalAirTemperature))
           + theWaterCatch * 0.23885 * (HEAT_CAPACITY_WATER / 1000.) * (convertUnit(NOPREFIX, KELVIN, NOPREFIX, FAHRENHEIT, theTskin)
                   - convertUnit(NOPREFIX, KELVIN, NOPREFIX, FAHRENHEIT, totalAirTemperature))
           + theWaterCatch * theEnthalpyVapourization;
}

double wingAntiIcing::getHeatCapacityOfAir(double theAirTemp) {
    return (1.003 + 0.00001 * theAirTemp) * 1000.;
}

double wingAntiIcing::getHeatConductivityOfAir(double theAirTemp) {
    return 0.02427 + 7.13 * pow(10., -5.) * theAirTemp;
}

double wingAntiIcing::getMassFlow(double theInnerSkinTemp, double theBleedAirTemp, double effiencyFactor, double theHeatFlux) {
    double tempHeatCapacityAir = getHeatCapacityOfAir(0.5 * (theBleedAirTemp + theInnerSkinTemp));
    return 0.293 * theHeatFlux / (tempHeatCapacityAir * fabs(theBleedAirTemp - theInnerSkinTemp) * effiencyFactor);
}

double wingAntiIcing::getIcingSurface(double theYPOS, double theSegWidth, double frontSpar,
                           const geom2::MultisectionSurface<geom2::AirfoilSection>& Wing) {
    double contour_length(0.); //length of the relevant contour for anti icing
    /* Determine section */
    auto section_index = geom2::detail::get_section_index(Wing.sections, theYPOS);
    //convert section_index to an int
    int ID = section_index.value_or(-1);
    geom2::Polygon_2 contour = Wing.sections.at(ID).get_contour();
    /* go over the sections contour and get the contour length infront of the front spar */
    for (auto edge = contour.edges_begin(); edge != contour.edges_end(); ++edge) {
        geom2::Point_2 source = edge->source();
        geom2::Point_2 target = edge->target();
        // Check if both endpoints are in front of the boundary
        if (source.x() <= frontSpar && target.x() <= frontSpar) {
            // Compute the length of the edge and accumulate
            double length = CGAL::squared_distance(source, target);
            contour_length += std::sqrt(length);
        } else if (source.x() <= frontSpar && target.x() > frontSpar) {
            // Edge crosses the boundary, accumulate portion in front of the boundary
            double lengthInFront = std::abs(target.x() - frontSpar);
            contour_length += lengthInFront;
        } else if (source.x() > frontSpar && target.x() <= frontSpar) {
            // Edge crosses the boundary, accumulate portion in front of the boundary
            double lengthInFront = std::abs(source.x() - frontSpar);
            contour_length += lengthInFront;
        }
        // If both endpoints are to the right of the boundary, ignore the edge
    }
    return contour_length * theSegWidth;
}

double wingAntiIcing::Interpolation(double X0, double X1, double Y0, double Y1, double theX) {
    return (Y1 - Y0) / (X1 - X0) * (theX - X0) + Y0;
}
