/**
 * \file standardSystemsDesign.cpp
 * \author
 * \brief standard systems design strategy
 * \version 0.1
 * \date 2024-06-27
 * 
 * @copyright Copyright (c) 2024
 * 
 */
#include <iostream>
#include <memory>
#include <string>
#include <cstdlib>
#include "standardSystemsDesign.h"
#include "systemsIOData.h"
#include "standardData.h"
#include "systemsDesignDefines.h"

// Models of the individual aircraft systems
#include "aircraftSystems/conventionalATA21.h"
#include "aircraftSystems/conventionalATA24.h"
#include "aircraftSystems/conventionalATA25.h"
#include "aircraftSystems/conventionalATA26.h"
#include "aircraftSystems/conventionalATA27.h"
#include "aircraftSystems/conventionalATA28.h"
#include "aircraftSystems/conventionalATA29.h"
#include "aircraftSystems/conventionalATA30.h"
#include "aircraftSystems/electricATA30.h"
#include "aircraftSystems/conventionalATA32.h"
#include "aircraftSystems/conventionalATA33.h"
#include "aircraftSystems/conventionalATA35.h"
#include "aircraftSystems/conventionalATA36.h"
#include "aircraftSystems/conventionalATA49.h"
#include "aircraftSystems/conventionalATA70.h"
#include "aircraftSystems/ATAXX.h"

void standardSystemsDesign::initialize() {
    const node& config = rtIO_->moduleConfig;
    /* Initializing IOData */
    data_ = std::make_shared<systemsIOData>(*rtIO_, config, rtIO_->getMissionDataDir());
    /* Read geometry components from aircraft xml */
    data_->readFuselages(rtIO_->acxml, rtIO_->getIODir()+"/");
    data_->readNacelles(rtIO_->acxml, rtIO_->getIODir()+"/");
    data_->readWing(rtIO_->acxml, rtIO_->getIODir()+"/");
    data_->readEmpennage(rtIO_->acxml, rtIO_->getIODir()+"/");

    /* Initializing values with data from the aircraft xml */
    data_->readPropulsion(rtIO_->acxml, rtIO_->getEngineDataDir());
    data_->readMassesFromXML(rtIO_->acxml);
    data_->readPerformance(rtIO_->acxml);
    data_->readAccomodation(rtIO_->acxml);

    /* Initializing individual systems */
    initializeSystems();

    /* Checking inputs from xml and config for consistency */
    checkUserInputForConsistency();
    checkGeometryForConsistency();
    //checkArchitectureForConsistency(); //TODO(seabrooke): re-implement once interfaceSystem logic is finished
}

void standardSystemsDesign::run() {
    myRuntimeInfo->out << "Run systemsDesign standard strategy" << std::endl;
    /* Run modes */
    if (data_->configuration.designCase) {
        myRuntimeInfo->out << "Run - SizingMode" << std::endl;
        size();
    } else {
        myRuntimeInfo->out << "Run - StudyMode" << std::endl;
        study();
    }
}

void standardSystemsDesign::update() {
    if (data_->configuration.designCase == true) {
        myRuntimeInfo->out << "Update systems" << std::endl;
        data_->updateMassProperties();
            /* Update system specific data in the acxml */
        for (size_t i = 0; i < data_->Systems.allSystems.size(); i++) {
            data_->Systems.allSystems.at(i)->updateXML();
        }
    }
    /* Update mission file */
    data_->updateMissionXML(rtIO_->getMissionDataDir()); //this updates all values that are calculated with the mission steps, including values in the acxml
}

void standardSystemsDesign::report() {
    if (rtIO_->plotOn) {
        myRuntimeInfo->out << "Generate plots ..." << std::endl;
        generatePlotData();
    }
    myRuntimeInfo->out << "Write reports ..." << std::endl;
    setHtmlBody();
    setTexBody();
    report_.generateReports();
}

void standardSystemsDesign::save() {
    //not used on this level
}

/* METHODS */

void standardSystemsDesign::size() {
    myRuntimeInfo->out << "Start calculation ..." << std::endl;
    /* Calculating performance profile */
    calculatePerformanceProfile();
    /* Calculating the masses for the design case */
    myRuntimeInfo->out << "Calculate system masses ..." << std::endl;
    getSystemsMass();
    myWeightsPt = new weightsAndCGs(data_);
    myWeightsPt->getWeights();
    myRuntimeInfo->out << "... calculation finished!" << std::endl;
}

void standardSystemsDesign::study() {
    myRuntimeInfo->out << "Start calculation ..." << std::endl;
    /* Calculating performance profile */
    calculatePerformanceProfile();
    myRuntimeInfo->out << "No system masses are determined for the analysis calculation.!" << std::endl;
    bool reqMissionOfftakesWritten(rtIO_->acxml.at("AcftExchangeFile/Performance/RequirementsChecks/MissionFile").getIntAttrib("OfftakesWritten"));
    if (!reqMissionOfftakesWritten && data_->configuration.reqMission) {
        myRuntimeInfo->out << "Calculation for requirement mission ..." << std::endl;
        calculatePerformanceProfile();
        rtIO_->acxml.at("AcftExchangeFile/Performance/RequirementsChecks/MissionFile").setAttrib("OfftakesWritten", 1);
    }
}

void standardSystemsDesign::initializeSystems() {
    //Initialization of energy sinks
    for (size_t i = 0; i < data_->configuration.Sinks.size(); i++) {
        if (data_->configuration.Sinks.at(i).systemName == "remainingConsumers") {
            data_->Systems.energySink.push_back(
                new ATAXX(data_->configuration.Sinks.at(i).systemName, data_->configuration.Sinks.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Sinks.at(i).systemName == "conventionalFurnishing") { //ATA-25
            data_->Systems.energySink.push_back(
                new conventionalATA25(data_->configuration.Sinks.at(i).systemName, data_->configuration.Sinks.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Sinks.at(i).systemName == "conventionalFireProtection") { //ATA-26
            data_->Systems.energySink.push_back(
                new conventionalATA26(data_->configuration.Sinks.at(i).systemName, data_->configuration.Sinks.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Sinks.at(i).systemName == "conventionalFlightControl") { //ATA-27
            data_->Systems.energySink.push_back(
                new conventionalATA27(data_->configuration.Sinks.at(i).systemName, data_->configuration.Sinks.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Sinks.at(i).systemName == "conventionalFuel") { //ATA-28
            data_->Systems.energySink.push_back(
                new conventionalATA28(data_->configuration.Sinks.at(i).systemName, data_->configuration.Sinks.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Sinks.at(i).systemName == "conventionalIceRainProtection") { //ATA-30
            data_->Systems.energySink.push_back(
                new conventionalATA30(data_->configuration.Sinks.at(i).systemName, data_->configuration.Sinks.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Sinks.at(i).systemName == "conventionalGear") { //ATA-32
            data_->Systems.energySink.push_back(
                new conventionalATA32(data_->configuration.Sinks.at(i).systemName, data_->configuration.Sinks.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Sinks.at(i).systemName == "conventionalLighting") { //ATA-33
            data_->Systems.energySink.push_back(
                new conventionalATA33(data_->configuration.Sinks.at(i).systemName, data_->configuration.Sinks.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Sinks.at(i).systemName == "conventionalOxygenSystem") { // ATA-35
            data_->Systems.energySink.push_back(
                new conventionalATA35(data_->configuration.Sinks.at(i).systemName, data_->configuration.Sinks.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Sinks.at(i).systemName == "electricIceRainProtection") { //ATA-30
            data_->Systems.energySink.push_back(
                new electricATA30(data_->configuration.Sinks.at(i).systemName, data_->configuration.Sinks.at(i).systemOperating, data_, *rtIO_));
        } else {
            myRuntimeInfo->err << "The system is unknown!     " << data_->configuration.Sinks.at(i).systemName << std::endl;
            exit(1);
        }
    }
    //Initialization of the ECS
    data_->Systems.energyECS.push_back(new conventionalATA21(data_->configuration.theECS.systemName, data_->configuration.theECS.systemOperating, data_, *rtIO_));
    //Initializing the Virtual Systems
    for (int i = 0; i < data_->configuration.numberOfVirtualSystems; i++) {
        if (data_->configuration.virtualSystem.at(i).systemName == "virtualSystem") {
            std::stringstream nrATA;
            data_->Systems.virtualSystem.push_back(new aircraftSystem(data_->configuration.virtualSystem.at(i).systemName,
                                                                      data_->configuration.virtualSystem.at(i).systemOperating, data_, *rtIO_));
            nrATA << "ATA-virtual" << i + 1;
            data_->Systems.virtualSystem.back()->ATA = nrATA.str();
        } else {
            myRuntimeInfo->err << "The system is unknown!     " << data_->configuration.virtualSystem.at(i).systemName << std::endl;
            exit(1);
        }
    }
    //Initialization of energy sources
    for (int i = 0; i < data_->configuration.numberOfSources; i++) {
        if (data_->configuration.Sources.at(i).systemName == "conventionalAPU") {
            data_->Systems.energySource.push_back(new conventionalATA49(data_->configuration.Sources.at(i).systemName,
                                                                     data_->configuration.Sources.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Sources.at(i).systemName == "conventionalPropulsion") {
            data_->Systems.energySource.push_back(new conventionalATA70(data_->configuration.Sources.at(i).systemName,
                                                                     data_->configuration.Sources.at(i).systemOperating, data_, *rtIO_));
        } else {
            myRuntimeInfo->err << "The system is unknown!     " << data_->configuration.Sources.at(i).systemName << std::endl;
            exit(1);
        }
    }
    //Initialization of the power distributors
    for (int i = 0; i < data_->configuration.numberOfConductors; i++) {
        if (data_->configuration.Conductors.at(i).systemName == "ElectricSystem") {
            data_->Systems.energyConductor.push_back(new conventionalATA24(data_->configuration.Conductors.at(i).systemName,
                                                                        data_->configuration.Conductors.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Conductors.at(i).systemName == "HydraulicSystem") {
            data_->Systems.energyConductor.push_back(new conventionalATA29(data_->configuration.Conductors.at(i).systemName,
                                                                        data_->configuration.Conductors.at(i).systemOperating, data_, *rtIO_));
        } else if (data_->configuration.Conductors.at(i).systemName == "BleedAirSystem") {
            data_->Systems.energyConductor.push_back(new conventionalATA36(data_->configuration.Conductors.at(i).systemName,
                                                                        data_->configuration.Conductors.at(i).systemOperating, data_, *rtIO_));
        } else {
            myRuntimeInfo->err << "The system is unknown!     " << data_->configuration.Conductors.at(i).systemName << std::endl;
            exit(1);
        }
    }
    /* Group all systems into one vector */
    data_->Systems.getAllSystems();
    /* Call read function of each system for reading system specific parameters from config */
    for (size_t i = 0; i < data_->Systems.allSystems.size(); i++) {
        data_->Systems.allSystems.at(i)->readParameters(data_->configFile);
    }
    /* since ECS is not yet in the allSystems vector: */
    data_->Systems.energyECS.back()->readParameters(data_->configFile);
}

void standardSystemsDesign::calculatePerformanceProfile() {
    /* Getting energy consumption */
    getSinkEnergyConsumption();
    //Transfer of electric <--> hydraulic power through the use of electric pumps and hydraulic generators
    transferElec2HydPower(&data_->data.sinkConsumption.Electric, &data_->data.sinkConsumption.Hydraulic, &data_->data.transfSinkComnsumption.Hydraulic);
    transferHyd2ElecPower(&data_->data.sinkConsumption.Hydraulic, &data_->data.sinkConsumption.Electric,
                          &data_->data.sinkConsumption.heatLoad, &data_->data.transfSinkComnsumption.Electric);
    //power ECS und iterativ bis konvergenz!!!
    double avgECSBleedDeviation(1.);
    std::vector<double> avgECSBleedLoad;
    int Run(1);
    do {
        //Initialization of the accumulated performances of the ECS
        if (static_cast<int>(data_->data.ECSConsumption.Bleed.missionPower.size()) <= 0) {
            clearInitializePowerConsumption(&data_->data.ECSConsumption);
        }
        //Initialization of the accumulated performances of the Conductor if not already done
        if (static_cast<int>(data_->data.conductorConsumption.Bleed.missionPower.size()) <= 0) {
            clearInitializePowerConsumption(&data_->data.conductorConsumption);
        }
        //initialize the individual Conductor services
        for (size_t i(0); i < data_->Systems.energyConductor.size(); i++) {
            data_->Systems.energyConductor.at(i)->initializePowerProfile();
        }
        //Calculating the energy consumption of the air conditioning system
        getECSEnergyConsumption();
        //Calculation of energy losses by the distribution elements
        getConductorEnergyConsumption(data_->data.sinkConsumption, data_->data.ECSConsumption);
        //Transfer of electric <--> hydraulic power through the use of electric pumps and hydraulic generators
        transferElec2HydPower(&data_->data.conductorConsumption.Electric, &data_->data.conductorConsumption.Hydraulic, &data_->data.transfConductorConsumption.Hydraulic);
        transferHyd2ElecPower(&data_->data.conductorConsumption.Hydraulic, &data_->data.conductorConsumption.Electric,
                              &data_->data.conductorConsumption.heatLoad, &data_->data.transfConductorConsumption.Electric);
        //Cleaning up the peaks
        for (size_t n(0); n < data_->data.totalConsumption.Hydraulic.size(); n++) {
            cleanPowerPeaks(&data_->data.totalConsumption.Hydraulic.at(n));
        }
        for (size_t n(0); n < data_->data.totalConsumption.Electric.size(); n++) {
            cleanPowerPeaks(&data_->data.totalConsumption.Electric.at(n));
        }
        cleanPowerPeaks(&data_->data.totalConsumption.heatLoad);
        cleanPowerPeaks(&data_->data.totalConsumption.Bleed);
        avgECSBleedLoad.push_back(getAvgBleedECS(data_->data.totalConsumption.Bleed));
        Run = static_cast<int>(avgECSBleedLoad.size()) - 1;
        if (Run > 1) {
            avgECSBleedDeviation = fabs(1. - avgECSBleedLoad.at(Run) / avgECSBleedLoad.at(Run - 1));
        } else {
            avgECSBleedDeviation = 1.0;
        }
    } while (avgECSBleedDeviation > 0.0001 && Run <= 50);
    //Power calculation of the sources
    getSourceEnergyConsumption();
    // Moving the ECS into the energy sink and adding it to the all systems vector
    data_->Systems.energySink.insert(data_->Systems.energySink.end(), data_->Systems.energyECS.back());
    data_->Systems.allSystems.insert(data_->Systems.allSystems.end(), data_->Systems.energyECS.back());
}

/* Performance Profile Methods */
double standardSystemsDesign::getAvgBleedECS(powerProfile theBleed) {
    double avgBleedAir(0.0);
    for (size_t i = 0; i < theBleed.missionPower.size(); i++) {
        avgBleedAir += theBleed.missionPower.at(i).baseLoad;
    }
    return avgBleedAir / (static_cast<double>(theBleed.missionPower.size()) - 1.);
}

void standardSystemsDesign::getSinkEnergyConsumption() {
    //Initialization of the accumulated sink performances
    clearInitializePowerConsumption(&data_->data.sinkConsumption);
    /* Energy Sinks **/
    for (size_t i = 0; i < data_->configuration.Sinks.size(); i++) {
        myRuntimeInfo->debug << "Calculate system power for sink: " << data_->Systems.energySink.at(i)->Name << std::endl;
        data_->Systems.energySink.at(i)->calculateSystemsPower(data_->data.sinkConsumption.Bleed, data_->data.sinkConsumption.Hydraulic,
                                                               data_->data.sinkConsumption.Electric, data_->data.sinkConsumption.heatLoad);
        getAccumulatedPower(*data_->Systems.energySink.at(i), &data_->data.sinkConsumption);
        myRuntimeInfo->debug << "Design Power (el, hyd, bleed): " << data_->Systems.energySink.at(i)->Electric.at(0).designPower << " , "
                                                                  << data_->Systems.energySink.at(i)->Hydraulic.at(0).designPower << " , "
                                                                  << data_->Systems.energySink.at(i)->bleedAir.designPower << " , " << std::endl;
    }
    /* Virtual Systems are defined by Config **/
    for (size_t i = 0; i < data_->configuration.virtualSystem.size(); i++) {
        calculateVirtualSystemPower(i);
        getAccumulatedPower(*data_->Systems.virtualSystem.at(i), &data_->data.sinkConsumption);
    }
    //Cleaning up the peaks
    for (size_t n(0); n < data_->data.sinkConsumption.Hydraulic.size(); n++) {
        cleanPowerPeaks(&data_->data.sinkConsumption.Hydraulic.at(n));
    }
    for (size_t n(0); n < data_->data.sinkConsumption.Electric.size(); n++) {
        cleanPowerPeaks(&data_->data.sinkConsumption.Electric.at(n));
    }
    cleanPowerPeaks(&data_->data.sinkConsumption.heatLoad);
    cleanPowerPeaks(&data_->data.sinkConsumption.Bleed);
    //Create the transferred consumption. Here first equal to the normal sinkConsumption
    clearInitializePowerConsumption(&data_->data.transfSinkComnsumption);
}

void standardSystemsDesign::getECSEnergyConsumption() {
    data_->Systems.energyECS.back()->calculateECSPower(data_->data.sinkConsumption.heatLoad, data_->data.ECSConsumption.heatLoad, data_->data.conductorConsumption.heatLoad);
    clearInitializePowerConsumption(&data_->data.ECSConsumption);
    getAccumulatedPower(*data_->Systems.energyECS.back(), &data_->data.ECSConsumption);
    //Cleaning up the peaks
    for (size_t n(0); n < data_->data.ECSConsumption.Hydraulic.size(); n++) {
        cleanPowerPeaks(&data_->data.ECSConsumption.Hydraulic.at(n));
    }
    for (size_t n(0); n < data_->data.ECSConsumption.Electric.size(); n++) {
        cleanPowerPeaks(&data_->data.ECSConsumption.Electric.at(n));
    }
    cleanPowerPeaks(&data_->data.ECSConsumption.heatLoad);
    cleanPowerPeaks(&data_->data.ECSConsumption.Bleed);
    //Create the transferred consumption
    clearInitializePowerConsumption(&data_->data.transfECSConsumption);
}

void standardSystemsDesign::getConductorEnergyConsumption(const standardData::accumulatedPowerConsumption& aSinkConsumption,
                                                          const standardData::accumulatedPowerConsumption& anECSConsumption) {
    clearInitializePowerConsumption(&data_->data.totalConsumption);
    standardData::accumulatedPowerConsumption dummyConsumption;
    clearInitializePowerConsumption(&dummyConsumption);
    getTotalConsumption(&data_->data.totalConsumption, aSinkConsumption, anECSConsumption, dummyConsumption);
    //Cleaning up the peaks
    for (size_t n(0); n < data_->data.totalConsumption.Hydraulic.size(); n++) {
        cleanPowerPeaks(&data_->data.totalConsumption.Hydraulic.at(n));
    }
    for (size_t n(0); n < data_->data.totalConsumption.Electric.size(); n++) {
        cleanPowerPeaks(&data_->data.totalConsumption.Electric.at(n));
    }
    cleanPowerPeaks(&data_->data.totalConsumption.heatLoad);
    cleanPowerPeaks(&data_->data.totalConsumption.Bleed);
    for (size_t i = 0; i < data_->configuration.Conductors.size(); i++) {
        data_->Systems.energyConductor.at(i)->calculateConductingPower(data_->data.totalConsumption.Bleed,
                                                                    data_->data.totalConsumption.Electric, data_->data.totalConsumption.Hydraulic);
    }
    clearInitializePowerConsumption(&data_->data.conductorConsumption);
    for (size_t j = 0; j < data_->configuration.Conductors.size(); j++) {
        getAccumulatedPower(*data_->Systems.energyConductor.at(j), &data_->data.conductorConsumption);
    }
    //Cleaning up the peaks
    for (size_t n(0); n < data_->data.ECSConsumption.Hydraulic.size(); n++) {
        cleanPowerPeaks(&data_->data.ECSConsumption.Hydraulic.at(n));
    }
    for (size_t n(0); n < data_->data.ECSConsumption.Electric.size(); n++) {
        cleanPowerPeaks(&data_->data.ECSConsumption.Electric.at(n));
    }
    cleanPowerPeaks(&data_->data.ECSConsumption.heatLoad);
    cleanPowerPeaks(&data_->data.ECSConsumption.Bleed);
    //Create the transferred consumption
    clearInitializePowerConsumption(&data_->data.transfConductorConsumption);
}

//rename to something like: getRequiredSourcePower
void standardSystemsDesign::getSourceEnergyConsumption() {
    clearInitializePowerConsumption(&data_->data.totalConsumption);
    /* Get the total power consumption of all systems */
    getTotalConsumption(&data_->data.totalConsumption,
                        data_->data.sinkConsumption, data_->data.ECSConsumption, data_->data.conductorConsumption);
    /* Cleaning up the peaks */
    for (size_t n(0); n < data_->data.totalConsumption.Hydraulic.size(); n++) {
        cleanPowerPeaks(&data_->data.totalConsumption.Hydraulic.at(n));
    }
    for (size_t n(0); n < data_->data.totalConsumption.Electric.size(); n++) {
        cleanPowerPeaks(&data_->data.totalConsumption.Electric.at(n));
    }
    cleanPowerPeaks(&data_->data.totalConsumption.heatLoad);
    cleanPowerPeaks(&data_->data.totalConsumption.Bleed);
    // calling source power functions also calls the addEngineEfficiencies function
    for (size_t i = 0; i < data_->configuration.Sources.size(); i++) {
        data_->Systems.energySource.at(i)->calculateSourcePower(data_->data.totalConsumption.Bleed, data_->data.totalConsumption.Electric, data_->data.totalConsumption.Hydraulic);
    }
}

void standardSystemsDesign::getTotalConsumption(standardData::accumulatedPowerConsumption *theTotal,
        standardData::accumulatedPowerConsumption Sinks,
        standardData::accumulatedPowerConsumption ECS,
        standardData::accumulatedPowerConsumption Conductors) {
    /* Maximum performance depending on the mission */
    theTotal->Bleed.maxMissionPower     =   Sinks.Bleed.maxMissionPower
                                            + ECS.Bleed.maxMissionPower
                                            + Conductors.Bleed.maxMissionPower;
    for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
        theTotal->Electric.at(n).maxMissionPower  =   Sinks.Electric.at(n).maxMissionPower
                + ECS.Electric.at(n).maxMissionPower
                + Conductors.Electric.at(n).maxMissionPower;
    }
    /*aigner: Loop over the hydraulic circuit*/
    for (size_t n(0); n < data_->configuration.Hydraulic.circuits.size(); n++) {
        theTotal->Hydraulic.at(n).maxMissionPower =   Sinks.Hydraulic.at(n).maxMissionPower
                + ECS.Hydraulic.at(n).maxMissionPower
                + Conductors.Hydraulic.at(n).maxMissionPower;
    }
    /* Design Power **/
    theTotal->Bleed.designPower         =   Sinks.Bleed.designPower
                                            + ECS.Bleed.designPower
                                            + Conductors.Bleed.designPower;
    for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
        theTotal->Electric.at(n).designPower  =   Sinks.Electric.at(n).designPower
                + ECS.Electric.at(n).designPower
                + Conductors.Electric.at(n).designPower;
    }
    /*aigner: Loop over the hydraulic circuit*/
    for (size_t n(0); n < data_->configuration.Hydraulic.circuits.size(); n++) {
        theTotal->Hydraulic.at(n).designPower     =   Sinks.Hydraulic.at(n).designPower
                + ECS.Hydraulic.at(n).designPower
                + Conductors.Hydraulic.at(n).designPower;
    }
    /* Summation of the performances per mission section **/
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        /* Bleed Air **/
        theTotal->Bleed.missionPower.at(i).baseLoad =     Sinks.Bleed.missionPower.at(i).baseLoad
                + ECS.Bleed.missionPower.at(i).baseLoad
                + Conductors.Bleed.missionPower.at(i).baseLoad;
        for (size_t j = 0; j < Sinks.Bleed.missionPower.at(i).peakLoad.size(); j++) {
            theTotal->Bleed.missionPower.at(i).peakLoad.push_back(Sinks.Bleed.missionPower.at(i).peakLoad.at(j));
            theTotal->Bleed.missionPower.at(i).timePeak.push_back(Sinks.Bleed.missionPower.at(i).timePeak.at(j));
        }
        for (size_t j = 0; j < ECS.Bleed.missionPower.at(i).peakLoad.size(); j++) {
            theTotal->Bleed.missionPower.at(i).peakLoad.push_back(ECS.Bleed.missionPower.at(i).peakLoad.at(j));
            theTotal->Bleed.missionPower.at(i).timePeak.push_back(ECS.Bleed.missionPower.at(i).timePeak.at(j));
        }
        for (size_t j = 0; j < Conductors.Bleed.missionPower.at(i).peakLoad.size(); j++) {
            theTotal->Bleed.missionPower.at(i).peakLoad.push_back(Conductors.Bleed.missionPower.at(i).peakLoad.at(j));
            theTotal->Bleed.missionPower.at(i).timePeak.push_back(Conductors.Bleed.missionPower.at(i).timePeak.at(j));
        }
        /* Electric **/
        for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
            theTotal->Electric.at(n).missionPower.at(i).baseLoad =   Sinks.Electric.at(n).missionPower.at(i).baseLoad
                    + ECS.Electric.at(n).missionPower.at(i).baseLoad
                    + Conductors.Electric.at(n).missionPower.at(i).baseLoad;
            for (size_t j = 0; j < Sinks.Electric.at(n).missionPower.at(i).peakLoad.size(); j++) {
                theTotal->Electric.at(n).missionPower.at(i).peakLoad.push_back(Sinks.Electric.at(n).missionPower.at(i).peakLoad.at(j));
                //            std::cout << "theTotal->Electric.missionPower.at(i).peakLoad.at(j) is: " << theTotal->Electric.at(n).missionPower.at(i).peakLoad.at(j) << std::endl;
                theTotal->Electric.at(n).missionPower.at(i).timePeak.push_back(Sinks.Electric.at(n).missionPower.at(i).timePeak.at(j));
            }
            for (size_t j = 0; j < ECS.Electric.at(n).missionPower.at(i).peakLoad.size(); j++) {
                theTotal->Electric.at(n).missionPower.at(i).peakLoad.push_back(ECS.Electric.at(n).missionPower.at(i).peakLoad.at(j));
                theTotal->Electric.at(n).missionPower.at(i).timePeak.push_back(ECS.Electric.at(n).missionPower.at(i).timePeak.at(j));
            }
            for (size_t j = 0; j < Conductors.Electric.at(n).missionPower.at(i).peakLoad.size(); j++) {
                theTotal->Electric.at(n).missionPower.at(i).peakLoad.push_back(Conductors.Electric.at(n).missionPower.at(i).peakLoad.at(j));
                theTotal->Electric.at(n).missionPower.at(i).timePeak.push_back(Conductors.Electric.at(n).missionPower.at(i).timePeak.at(j));
            }
        }
        /* Hydraulic **/
        for (size_t n(0); n < data_->configuration.Hydraulic.circuits.size(); n++) {
            theTotal->Hydraulic.at(n).missionPower.at(i).baseLoad =     Sinks.Hydraulic.at(n).missionPower.at(i).baseLoad
                    + ECS.Hydraulic.at(n).missionPower.at(i).baseLoad
                    + Conductors.Hydraulic.at(n).missionPower.at(i).baseLoad;
            for (size_t j = 0; j < Sinks.Hydraulic.at(n).missionPower.at(i).peakLoad.size(); j++) {
                theTotal->Hydraulic.at(n).missionPower.at(i).peakLoad.push_back(Sinks.Hydraulic.at(n).missionPower.at(i).peakLoad.at(j));
                theTotal->Hydraulic.at(n).missionPower.at(i).timePeak.push_back(Sinks.Hydraulic.at(n).missionPower.at(i).timePeak.at(j));
            }
            for (size_t j = 0; j < ECS.Hydraulic.at(n).missionPower.at(i).peakLoad.size(); j++) {
                theTotal->Hydraulic.at(n).missionPower.at(i).peakLoad.push_back(ECS.Hydraulic.at(n).missionPower.at(i).peakLoad.at(j));
                theTotal->Hydraulic.at(n).missionPower.at(i).timePeak.push_back(ECS.Hydraulic.at(n).missionPower.at(i).timePeak.at(j));
            }
            for (size_t j = 0; j < Conductors.Hydraulic.at(n).missionPower.at(i).peakLoad.size(); j++) {
                theTotal->Hydraulic.at(n).missionPower.at(i).peakLoad.push_back(Conductors.Hydraulic.at(n).missionPower.at(i).peakLoad.at(j));
                theTotal->Hydraulic.at(n).missionPower.at(i).timePeak.push_back(Conductors.Hydraulic.at(n).missionPower.at(i).timePeak.at(j));
            }
        }
    }
}

void standardSystemsDesign::clearInitializePowerConsumption(standardData::accumulatedPowerConsumption *theConsumption) {
    theConsumption->Bleed.missionPower.clear();
    theConsumption->Bleed.designPower = 0.;
    theConsumption->Bleed.maxMissionPower = 0.;
    theConsumption->Electric.clear();
    for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
        theConsumption->Electric.push_back(powerProfile());
        theConsumption->Electric.back().missionPower.clear();
        theConsumption->Electric.back().designPower = 0.;
        theConsumption->Electric.back().maxMissionPower = 0.;
    }
    theConsumption->heatLoad.missionPower.clear();
    theConsumption->heatLoad.designPower = 0.;
    theConsumption->heatLoad.maxMissionPower = 0.;
    theConsumption->Hydraulic.clear();
    for (size_t n(0); n < data_->configuration.Hydraulic.circuits.size(); n++) {
        theConsumption->Hydraulic.push_back(powerProfile());
        theConsumption->Hydraulic.back().missionPower.clear();
        theConsumption->Hydraulic.back().designPower = 0.;
        theConsumption->Hydraulic.back().maxMissionPower = 0.;
    }
    //Mission Loads
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        theConsumption->Bleed.missionPower.push_back(missionLoads());
        for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
            theConsumption->Electric.at(n).missionPower.push_back(missionLoads());
        }
        theConsumption->heatLoad.missionPower.push_back(missionLoads());
        for (size_t n(0); n < data_->configuration.Hydraulic.circuits.size(); n ++) {
            theConsumption->Hydraulic.at(n).missionPower.push_back(missionLoads());
        }
    }
}

void standardSystemsDesign::getAccumulatedPower(const aircraftSystem& theSystem, standardData::accumulatedPowerConsumption *theConsumption) {
    /* Maximum performance over the mission **/
    theConsumption->Bleed.maxMissionPower       += theSystem.bleedAir.getMaxPower();
    for (size_t n(0); n < theConsumption->Electric.size(); n++) {
        theConsumption->Electric.at(n).maxMissionPower    += theSystem.Electric.at(n).getMaxPower();
    }
    for (size_t n(0); n < theConsumption->Hydraulic.size(); n++) {
        theConsumption->Hydraulic.at(n).maxMissionPower   += theSystem.Hydraulic.at(n).getMaxPower();
    }
    /* Design Power **/
    theConsumption->Bleed.designPower       += theSystem.bleedAir.designPower;
    for (size_t n(0); n < theConsumption->Electric.size(); n++) {
        theConsumption->Electric.at(n).designPower    += theSystem.Electric.at(n).designPower;
    }
    for (size_t n(0); n < theConsumption->Hydraulic.size(); n++) {
        theConsumption->Hydraulic.at(n).designPower   += theSystem.Hydraulic.at(n).designPower;
    }
    /* Summation of the performances per mission section **/
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        theConsumption->Bleed.missionPower.at(i).baseLoad += theSystem.bleedAir.missionPower.at(i).baseLoad;
        for (size_t j = 0; j < theSystem.bleedAir.missionPower.at(i).peakLoad.size(); j++) {
            theConsumption->Bleed.missionPower.at(i).peakLoad.push_back(theSystem.bleedAir.missionPower.at(i).peakLoad.at(j));
            theConsumption->Bleed.missionPower.at(i).timePeak.push_back(theSystem.bleedAir.missionPower.at(i).timePeak.at(j));
        }
        for (size_t n(0); n < theConsumption->Electric.size(); n++) {
            theConsumption->Electric.at(n).missionPower.at(i).baseLoad += theSystem.Electric.at(n).missionPower.at(i).baseLoad;
            for (size_t j = 0; j < theSystem.Electric.at(n).missionPower.at(i).peakLoad.size(); j++) {
                theConsumption->Electric.at(n).missionPower.at(i).peakLoad.push_back(theSystem.Electric.at(n).missionPower.at(i).peakLoad.at(j));
                theConsumption->Electric.at(n).missionPower.at(i).timePeak.push_back(theSystem.Electric.at(n).missionPower.at(i).timePeak.at(j));
            }
        }
        for (size_t n(0); n < theConsumption->Hydraulic.size(); n++) {
            theConsumption->Hydraulic.at(n).missionPower.at(i).baseLoad += theSystem.Hydraulic.at(n).missionPower.at(i).baseLoad;
            for (size_t j = 0; j < theSystem.Hydraulic.at(n).missionPower.at(i).peakLoad.size(); j++) {
                theConsumption->Hydraulic.at(n).missionPower.at(i).peakLoad.push_back(theSystem.Hydraulic.at(n).missionPower.at(i).peakLoad.at(j));
                theConsumption->Hydraulic.at(n).missionPower.at(i).timePeak.push_back(theSystem.Hydraulic.at(n).missionPower.at(i).timePeak.at(j));
            }
        }
        theConsumption->heatLoad.missionPower.at(i).baseLoad += theSystem.heatLoad.missionPower.at(i).baseLoad;
        for (size_t j = 0; j < theSystem.heatLoad.missionPower.at(i).peakLoad.size(); j++) {
            theConsumption->heatLoad.missionPower.at(i).peakLoad.push_back(theSystem.heatLoad.missionPower.at(i).peakLoad.at(j));
            theConsumption->heatLoad.missionPower.at(i).timePeak.push_back(theSystem.heatLoad.missionPower.at(i).timePeak.at(j));
        }
        // maximum performance in dependence of the mission section
//        double sum(0.0);
//        sum = theConsumption->Bleed.missionPower.at(i).baseLoad;
//        for(int j = 0; j <= int(theConsumption->Bleed.missionPower.at(i).peakLoad.size())-1; j++)
//        {
//            sum += theConsumption->Bleed.missionPower.at(i).peakLoad.at(j);
//        }
//        if(theConsumption->Bleed.maxPower < sum)
//        {
//            theConsumption->Bleed.maxPower = sum;
//        }
//
//        sum = theConsumption->Electric.missionPower.at(i).baseLoad;
//        for(int j = 0; j <= int(theConsumption->Electric.missionPower.at(i).peakLoad.size())-1; j++)
//        {
//            sum += theConsumption->Electric.missionPower.at(i).peakLoad.at(j);
//        }
//        if(theConsumption->Electric.maxPower < sum)
//        {
//            theConsumption->Electric.maxPower = sum;
//        }
//
//        sum = theConsumption->Hydraulic.missionPower.at(i).baseLoad;
//        for(int j = 0; j <= int(theConsumption->Hydraulic.missionPower.at(i).peakLoad.size())-1; j++)
//        {
//            sum += theConsumption->Hydraulic.missionPower.at(i).peakLoad.at(j);
//        }
//        if(theConsumption->Hydraulic.maxPower < sum)
//        {
//            theConsumption->Hydraulic.maxPower = sum;
//        }
//
//        sum = theConsumption->heatLoad.missionPower.at(i).baseLoad;
//        for(int j = 0; j <= int(theConsumption->heatLoad.missionPower.at(i).peakLoad.size())-1; j++)
//        {
//            sum += theConsumption->heatLoad.missionPower.at(i).peakLoad.at(j);
//        }
//        if(theConsumption->heatLoad.maxPower < sum)
//        {
//            theConsumption->heatLoad.maxPower = sum;
//        }
    }
}

void standardSystemsDesign::calculateVirtualSystemPower(int ID) {
    for (size_t n = 0; n < data_->mission.Steps.size(); n++) {
        int j(0);
        if (data_->mission.Steps.at(n).missionPhase == "Departure") {
            j = 0;
        } else if (data_->mission.Steps.at(n).missionPhase == "Cruise") {
            j = 1;
        } else if (data_->mission.Steps.at(n).missionPhase == "Approach") {
            j = 2;
        }
        /* Define baseLoad */
        data_->Systems.virtualSystem.at(ID)->bleedAir.missionPower.at(n).baseLoad = data_->configuration.virtualSystem.at(ID).powerBleed[j];
        for (size_t k(0); k < data_->configuration.Hydraulic.circuits.size(); k++) {
            data_->Systems.virtualSystem.at(ID)->Hydraulic.at(k).missionPower.at(n).baseLoad = data_->configuration.virtualSystem.at(ID).powerHydraulic[j];
            /*aigner: Temporarily equal distribution to all hydraulic circuits*/
            data_->Systems.virtualSystem.at(ID)->Hydraulic.at(k).missionPower.at(n).baseLoad *= 1 / static_cast<double>(data_->configuration.Hydraulic.circuits.size());
        }
        for (size_t k(0); k < data_->configuration.Electric.circuits.size(); k++) {
            data_->Systems.virtualSystem.at(ID)->Electric.at(k).missionPower.at(n).baseLoad = data_->configuration.virtualSystem.at(ID).powerElectric[j];
            /*aigner: Temporarily equal distribution to all electrical circuits*/
            data_->Systems.virtualSystem.at(ID)->Electric.at(k).missionPower.at(n).baseLoad *= 1 / static_cast<double>(data_->configuration.Electric.circuits.size());
        }
        data_->Systems.virtualSystem.at(ID)->heatLoad.missionPower.at(n).baseLoad = data_->configuration.virtualSystem.at(ID).heatLoad[j];
    }
    /* Define maxMissionPower */
    data_->Systems.virtualSystem.at(ID)->bleedAir.maxMissionPower = data_->Systems.virtualSystem.at(ID)->bleedAir.getMaxPower();
    for (size_t k(0); k < data_->configuration.Electric.circuits.size(); k++) {
        data_->Systems.virtualSystem.at(ID)->Electric.at(k).maxMissionPower = data_->Systems.virtualSystem.at(ID)->Electric.at(k).getMaxPower();
    }
    for (size_t k(0); k < data_->configuration.Hydraulic.circuits.size(); k++) {
        data_->Systems.virtualSystem.at(ID)->Hydraulic.at(k).maxMissionPower = data_->Systems.virtualSystem.at(ID)->Hydraulic.at(k).getMaxPower();
    }
    data_->Systems.virtualSystem.at(ID)->heatLoad.maxMissionPower = data_->Systems.virtualSystem.at(ID)->heatLoad.getMaxPower();
    /* Define designPower */
    data_->Systems.virtualSystem.at(ID)->bleedAir.designPower = data_->Systems.virtualSystem.at(ID)->bleedAir.maxMissionPower;
    for (size_t k(0); k < data_->configuration.Electric.circuits.size(); k++) {
        data_->Systems.virtualSystem.at(ID)->Electric.at(k).designPower = data_->Systems.virtualSystem.at(ID)->Electric.at(k).maxMissionPower;
    }
    for (size_t k(0); k < data_->configuration.Hydraulic.circuits.size(); k++) {
        data_->Systems.virtualSystem.at(ID)->Hydraulic.at(k).designPower = data_->Systems.virtualSystem.at(ID)->Hydraulic.at(k).maxMissionPower;
    }
    data_->Systems.virtualSystem.at(ID)->heatLoad.designPower = data_->Systems.virtualSystem.at(ID)->heatLoad.maxMissionPower;
}

/* Methods for converting power */
void standardSystemsDesign::transferElec2HydPower(std::vector<powerProfile> *myElectric, std::vector<powerProfile> *myHydraulic, std::vector<powerProfile> *transfHydraulic) {
    for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
        double electricDesignPower_tmp(myElectric->at(n).designPower);
        for (size_t k(0); k < data_->configuration.Electric.circuits.at(n).generators.size(); k++) {
            double genOperationFactor(data_->configuration.Electric.circuits.at(n).generators.at(k).operationFactor);
            double genEfficiency(data_->configuration.Electric.circuits.at(n).generators.at(k).efficiency);
            /*aigner: Run through the hydraulic circuits to allocate the energy to the correct circuit.*/
            for (size_t m(0); m < data_->configuration.Hydraulic.circuits.size(); m++) {
                if (data_->configuration.Electric.circuits.at(n).generators.at(k).powerSource.type == "Hydraulic"
                        && data_->configuration.Electric.circuits.at(n).generators.at(k).powerSource.sourceID == static_cast<int>(m + 1)) {
                    //Reassignment of the powerProfiles after calculating which part of the electrical energy is obtained from hydraulic generators.
                    double transferedPower(electricDesignPower_tmp * genOperationFactor);//Power transferred from hydraulic to electrical system
                    myElectric->at(n).designPower -= transferedPower;//Power is subtracted from Electric
                    myHydraulic->at(m).designPower += transferedPower * (1 / genEfficiency); //Power is added to Electric (with pump efficiency)
                    //Transferred hydraulic power
                    transfHydraulic->at(m).designPower = transferedPower * (1 / genEfficiency);
                }
            }
        }
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        /*aigner: Electrical energy provided by hydraulic generators --> baseLoad in the mission*/
        double peakLoad_tmp;
        for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
            double electricBaseLoad_tmp = myElectric->at(n).missionPower.at(i).baseLoad;
            for (size_t k(0); k < data_->configuration.Electric.circuits.at(n).generators.size(); k++) {
                double genOperationFactor(data_->configuration.Electric.circuits.at(n).generators.at(k).operationFactor);
                double genEfficiency(data_->configuration.Electric.circuits.at(n).generators.at(k).efficiency);
                /*aigner: Run through the electrical circuits to assign the energy to the correct circuit.*/
                for (size_t m(0); m < data_->configuration.Hydraulic.circuits.size(); m++) {
                    if (data_->configuration.Electric.circuits.at(n).generators.at(k).powerSource.type == "Hydraulic"
                            && data_->configuration.Electric.circuits.at(n).generators.at(k).powerSource.sourceID == static_cast<int>(m + 1)) {
                        double transferedPower(electricBaseLoad_tmp * genOperationFactor);//Power transferred from hydraulic to electrical system
                        //Reassignment of the powerProfiles after calculating which part of the electrical energy is obtained from hydraulic generators.
                        myElectric->at(n).missionPower.at(i).baseLoad -= transferedPower;
                        myHydraulic->at(m).missionPower.at(i).baseLoad += transferedPower * (1 / genEfficiency);
                        //Transferred hydraulic power
                        transfHydraulic->at(m).missionPower.at(i).baseLoad = transferedPower * (1 / genEfficiency);
                    }
                }
            }
            for (size_t j = 0; j < myElectric->at(n).missionPower.at(i).peakLoad.size(); j++) {
                /*aigner: electrical energy provided by hydraulic generators --> peakLoads in the mission*/
                peakLoad_tmp = myElectric->at(n).missionPower.at(i).peakLoad.at(j);
                for (size_t k(0); k < data_->configuration.Electric.circuits.at(n).generators.size(); k++) {
                    double genOperationFactor(data_->configuration.Electric.circuits.at(n).generators.at(k).operationFactor);
                    double genEfficiency(data_->configuration.Electric.circuits.at(n).generators.at(k).efficiency);
                    /*aigner: Run through the electrical circuits to assign the energy to the correct circuit.*/
                    for (size_t m(0); m < data_->configuration.Hydraulic.circuits.size(); m++) {
                        if (data_->configuration.Electric.circuits.at(n).generators.at(k).powerSource.type == "Hydraulic"
                                && data_->configuration.Electric.circuits.at(n).generators.at(k).powerSource.sourceID == static_cast<int>(m + 1)) {
                            //Reassignment of the powerProfiles after calculating which part of the hydraulic energy is obtained from electric pumps.
                            double transferedPower(peakLoad_tmp * genOperationFactor);//Power transferred from hydraulic to electrical system
                            myElectric->at(n).missionPower.at(i).peakLoad.at(j) -= transferedPower;
                            //Create a new peak, only if not already present
                            if (myHydraulic->at(m).missionPower.at(i).peakLoad.size() < j + 1) {
                                myHydraulic->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / genEfficiency));
                                myHydraulic->at(m).missionPower.at(i).timePeak.push_back(myElectric->at(n).missionPower.at(i).timePeak.at(j));
                                //Transferred hydraulic power
                                transfHydraulic->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / genEfficiency));
                                transfHydraulic->at(m).missionPower.at(i).timePeak.push_back(myElectric->at(n).missionPower.at(i).timePeak.at(j));
                            } else {
                                if (!accuracyCheck(myHydraulic->at(m).missionPower.at(i).timePeak.at(j), myElectric->at(n).missionPower.at(i).timePeak.at(j), ACCURACY_HIGH)) {
                                    myHydraulic->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / genEfficiency));
                                    myHydraulic->at(m).missionPower.at(i).timePeak.push_back(myElectric->at(n).missionPower.at(i).timePeak.at(j));
                                    //Transferred hydraulic power
                                    transfHydraulic->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / genEfficiency));
                                    transfHydraulic->at(m).missionPower.at(i).timePeak.push_back(myElectric->at(n).missionPower.at(i).timePeak.at(j));
                                } else {
                                    myHydraulic->at(m).missionPower.at(i).peakLoad.back() += transferedPower * (1 / genEfficiency);
                                    //Transferred hydraulic power
                                    //Create a new peak, only if not already present
                                    if (transfHydraulic->at(m).missionPower.at(i).peakLoad.size() < j + 1) {
                                        transfHydraulic->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / genEfficiency) * (1 / genEfficiency));
                                        transfHydraulic->at(m).missionPower.at(i).timePeak.push_back(myHydraulic->at(m).missionPower.at(i).timePeak.at(j));
                                    } else {
                                        if (!accuracyCheck(transfHydraulic->at(m).missionPower.at(i).timePeak.at(j),
                                                           myHydraulic->at(m).missionPower.at(i).timePeak.at(j), ACCURACY_HIGH)) {
                                            transfHydraulic->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / genEfficiency));
                                            transfHydraulic->at(m).missionPower.at(i).timePeak.push_back(myHydraulic->at(m).missionPower.at(i).timePeak.at(j));
                                        } else {
                                            transfHydraulic->at(m).missionPower.at(i).peakLoad.back() += transferedPower * (1 / genEfficiency);
                                        }
                                    }
                                    // TODO(aigner#1#06/26/20): Needs to be checked! Previous statements have no effect, if this line is uncommented. Which one is correct?
                                    //transfHydraulic.at(m).missionPower.at(i).peakLoad.back() = transferedPower*(1/genEfficiency);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    //Clean up peaks
    for (size_t n(0); n < myHydraulic->size(); n++) {
        cleanPowerPeaks(&myHydraulic->at(n));
    }
    for (size_t n(0); n < myElectric->size(); n++) {
        cleanPowerPeaks(&myElectric->at(n));
    }
    //Max Mission Power
    for (size_t n(0); n < data_->configuration.Hydraulic.circuits.size(); n++) {
        myHydraulic->at(n).maxMissionPower = myHydraulic->at(n).getMaxPower();
        //Transferred hydraulic power
        transfHydraulic->at(n).maxMissionPower = transfHydraulic->at(n).getMaxPower();
    }
    for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
        myElectric->at(n).maxMissionPower = myElectric->at(n).getMaxPower();
    }
}

void standardSystemsDesign::transferHyd2ElecPower(std::vector<powerProfile> *myHydraulic,
        std::vector<powerProfile> *myElectric,
        powerProfile *myHeatLoad,
        std::vector<powerProfile> *transfElectric) {
    for (size_t n(0); n < data_->configuration.Hydraulic.circuits.size(); n++) {
        double hydraulicDesignPower_tmp(myHydraulic->at(n).designPower);
        for (size_t k(0); k < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); k++) {
            double pumpOperationFactor(data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).operationFactor);
            double ePumpEfficiency(data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).efficiency);
            /*aigner: Run through the electrical circuits to assign the energy to the correct circuit.*/
            for (size_t m(0); m < data_->configuration.Electric.circuits.size(); m++) {
                if (data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).powerSource.type == "Electric"
                        && data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).powerSource.sourceID == static_cast<int>(m + 1)) {
                    //Reassignment of the powerProfiles after calculating which part of the hydraulic energy is obtained from electric pumps.
                    double transferedPower(hydraulicDesignPower_tmp * pumpOperationFactor);//Power fie is transferred from the hydraulic to the electrical system
                    myHydraulic->at(n).designPower -= transferedPower;//Power is subtracted from hydraulics
                    myElectric->at(m).designPower += transferedPower * (1 / ePumpEfficiency); //Power is added to Electric (with pump efficiency)
                    myHeatLoad->designPower += transferedPower * (1 / ePumpEfficiency - 1); //Heat generation due to pump efficiency
                    //Transferred electrical power
                    transfElectric->at(m).designPower = transferedPower * (1 / ePumpEfficiency);
                }
            }
        }
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        /*aigner: Hydraulic energy provided by electric pumps --> baseLoad in the mission*/
        double peakLoad_tmp;
        for (size_t n(0); n < data_->configuration.Hydraulic.circuits.size(); n++) {
            double hydraulicBaseLoad_tmp = myHydraulic->at(n).missionPower.at(i).baseLoad;
            for (size_t k(0); k < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); k++) {
                double pumpOperationFactor(data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).operationFactor);
                double ePumpEfficiency(data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).efficiency);
                /*aigner: Run through the electrical circuits to assign the energy to the correct circuit.*/
                for (size_t m(0); m < data_->configuration.Electric.circuits.size(); m++) {
                    if (data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).powerSource.type == "Electric"
                            && data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).powerSource.sourceID == static_cast<int>(m + 1)) {
                        double transferedPower(hydraulicBaseLoad_tmp * pumpOperationFactor);//Power transferred from hydraulic to electrical system
                        //Reassignment of the powerProfiles after calculating which part of the hydraulic energy is obtained from electric pumps.
                        myHydraulic->at(n).missionPower.at(i).baseLoad -= transferedPower;
                        myElectric->at(m).missionPower.at(i).baseLoad += transferedPower * (1 / ePumpEfficiency);
                        myHeatLoad->missionPower.at(i).baseLoad += transferedPower * (1 / ePumpEfficiency - 1);
                        //Transferred electrical power
                        transfElectric->at(m).missionPower.at(i).baseLoad = transferedPower * (1 / ePumpEfficiency);
                    }
                }
            }
            /*aigner: Hydraulic energy provided by electric pumps --> peakLoads in the mission*/
            for (size_t j = 0; j < myHydraulic->at(n).missionPower.at(i).peakLoad.size(); j++) {
                peakLoad_tmp = myHydraulic->at(n).missionPower.at(i).peakLoad.at(j);
                for (size_t k(0); k < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); k++) {
                    double pumpOperationFactor(data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).operationFactor);
                    double ePumpEfficiency(data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).efficiency);
                    /*aigner: Run through the electrical circuits to assign the energy to the correct circuit.*/
                    for (size_t m(0); m < data_->configuration.Electric.circuits.size(); m++) {
                        if (data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).powerSource.type == "Electric"
                                && data_->configuration.Hydraulic.circuits.at(n).pumps.at(k).powerSource.sourceID == static_cast<int>(m + 1)) {
                            //Reassignment of the powerProfiles after calculating which part of the hydraulic energy is obtained from electric pumps.
                            double transferedPower(peakLoad_tmp * pumpOperationFactor);//Power transferred from hydraulic to electrical system
                            myHydraulic->at(n).missionPower.at(i).peakLoad.at(j) -= transferedPower;
                            //Create a new peak, only if not already present
                            if (myHeatLoad->missionPower.at(i).peakLoad.size() < j + 1) {
                                myHeatLoad->missionPower.at(i).peakLoad.push_back(transferedPower * (1 / ePumpEfficiency - 1));
                                myHeatLoad->missionPower.at(i).timePeak.push_back(myHydraulic->at(n).missionPower.at(i).timePeak.at(j));
                            } else {
                                if (!accuracyCheck(myHeatLoad->missionPower.at(i).timePeak.at(j), myHydraulic->at(n).missionPower.at(i).timePeak.at(j), ACCURACY_HIGH)) {
                                    myHeatLoad->missionPower.at(i).peakLoad.push_back(transferedPower * (1 / ePumpEfficiency - 1));
                                    myHeatLoad->missionPower.at(i).timePeak.push_back(myHydraulic->at(n).missionPower.at(i).timePeak.at(j));
                                } else {
                                    myHeatLoad->missionPower.at(i).peakLoad.back() += transferedPower * (1 / ePumpEfficiency - 1);
                                }
                            }
                            //Create a new peak, only if not already present
                            if (myElectric->at(m).missionPower.at(i).peakLoad.size() < j + 1) {
                                myElectric->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / ePumpEfficiency));
                                myElectric->at(m).missionPower.at(i).timePeak.push_back(myHydraulic->at(n).missionPower.at(i).timePeak.at(j));
                                //Transferred hydraulic power
                                transfElectric->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / ePumpEfficiency));
                                transfElectric->at(m).missionPower.at(i).timePeak.push_back(myHydraulic->at(n).missionPower.at(i).timePeak.at(j));
                            } else {
                                if (!accuracyCheck(myElectric->at(m).missionPower.at(i).timePeak.at(j), myHydraulic->at(n).missionPower.at(i).timePeak.at(j), ACCURACY_HIGH)) {
                                    myElectric->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / ePumpEfficiency));
                                    myElectric->at(m).missionPower.at(i).timePeak.push_back(myHydraulic->at(n).missionPower.at(i).timePeak.at(j));
                                    //Transferred hydraulic power
                                    transfElectric->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / ePumpEfficiency));
                                    transfElectric->at(m).missionPower.at(i).timePeak.push_back(myHydraulic->at(n).missionPower.at(i).timePeak.at(j));
                                } else {
                                    myElectric->at(m).missionPower.at(i).peakLoad.back() += transferedPower * (1 / ePumpEfficiency);
                                    //Transferred hydraulic power
                                    //Create a new peak, only if not already present
                                    if (transfElectric->at(m).missionPower.at(i).peakLoad.size() < j + 1) {
                                        transfElectric->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / ePumpEfficiency));
                                        transfElectric->at(m).missionPower.at(i).timePeak.push_back(myElectric->at(m).missionPower.at(i).timePeak.at(j));
                                    } else {
                                        if (!accuracyCheck(transfElectric->at(m).missionPower.at(i).timePeak.at(j),
                                                           myElectric->at(m).missionPower.at(i).timePeak.at(j), ACCURACY_HIGH)) {
                                            transfElectric->at(m).missionPower.at(i).peakLoad.push_back(transferedPower * (1 / ePumpEfficiency));
                                            transfElectric->at(m).missionPower.at(i).timePeak.push_back(myElectric->at(m).missionPower.at(i).timePeak.at(j));
                                        } else {
                                            transfElectric->at(m).missionPower.at(i).peakLoad.back() += transferedPower * (1 / ePumpEfficiency);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    //Clean up peaks
    for (size_t n(0); n < myHydraulic->size(); n++) {
        cleanPowerPeaks(&myHydraulic->at(n));
    }
    for (size_t n(0); n < myElectric->size(); n++) {
        cleanPowerPeaks(&myElectric->at(n));
    }
    cleanPowerPeaks(myHeatLoad);
    //Max. Mission Power
    for (size_t n(0); n < data_->configuration.Hydraulic.circuits.size(); n++) {
        myHydraulic->at(n).maxMissionPower = myHydraulic->at(n).getMaxPower();
    }
    for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
        myElectric->at(n).maxMissionPower = myElectric->at(n).getMaxPower();
        //Transferred electrical power
        transfElectric->at(n).maxMissionPower = transfElectric->at(n).getMaxPower();
    }
    myHeatLoad->maxMissionPower = myHeatLoad->getMaxPower();
}

/* Methods for system masses */

void standardSystemsDesign::getSystemsMass() {
    for (size_t i = 0; i < data_->Systems.energySink.size(); i++) {
        data_->Systems.energySink.at(i)->calculateSinkMass(data_->Systems.energySink.at(i)->bleedAir, data_->Systems.energySink.at(i)->Electric,
                                                           data_->Systems.energySink.at(i)->Hydraulic);
        data_->data.accumulatedSystemsMass += data_->Systems.energySink.at(i)->Mass;
    }
    for (size_t i = 0; i < data_->configuration.virtualSystem.size(); i++) {
        data_->Systems.virtualSystem.at(i)->Mass = data_->configuration.virtualSystem.at(i).Mass;
        data_->data.accumulatedSystemsMass += data_->Systems.virtualSystem.at(i)->Mass;
        data_->Systems.virtualSystem.at(i)->refPoint[0] = data_->configuration.virtualSystem.at(i).refPoint[0];
        data_->Systems.virtualSystem.at(i)->refPoint[1] = data_->configuration.virtualSystem.at(i).refPoint[1];
        data_->Systems.virtualSystem.at(i)->refPoint[2] = data_->configuration.virtualSystem.at(i).refPoint[2];
    }
    clearInitializePowerConsumption(&data_->data.totalConsumption);
    getTotalConsumption(&data_->data.totalConsumption, data_->data.sinkConsumption, data_->data.ECSConsumption, data_->data.conductorConsumption);
    //Cleaning up the peaks
    for (size_t n(0); n < data_->data.totalConsumption.Hydraulic.size(); n++) {
        cleanPowerPeaks(&data_->data.totalConsumption.Hydraulic.at(n));
    }
    for (size_t n(0); n < data_->data.totalConsumption.Electric.size(); n++) {
        cleanPowerPeaks(&data_->data.totalConsumption.Electric.at(n));
    }
    cleanPowerPeaks(&data_->data.totalConsumption.heatLoad);
    cleanPowerPeaks(&data_->data.totalConsumption.Bleed);
    for (size_t i = 0; i < data_->configuration.Sources.size(); i++) {
        data_->Systems.energySource.at(i)->calculateSourceMass(data_->data.totalConsumption.Bleed, data_->data.totalConsumption.Electric, data_->data.totalConsumption.Hydraulic);
        data_->data.accumulatedSystemsMass += data_->Systems.energySource.at(i)->Mass;
    }
    for (size_t i = 0; i < data_->configuration.Conductors.size(); i++) {
        data_->Systems.energyConductor.at(i)->calculateConductorMass(data_->data.totalConsumption.Bleed, data_->data.totalConsumption.Electric,
                                                                  data_->data.totalConsumption.Hydraulic);
        data_->data.accumulatedSystemsMass += data_->Systems.energyConductor.at(i)->Mass;
    }
}

/* CHECK METHODS */

void standardSystemsDesign::checkArchitectureForConsistency() {
    /* Loop over all energy providers/mission phases */
    for (size_t i = 0; i < data_->energyProviders.size(); i++) {
        /* Check if energy source and carrier are known */
        if (data_->energyProviders.at(i).powerTrain == "propulsor") {
            /* check thorugh energy sources specified in config for conventionalPropulsion */
            bool sourceFound = checkEnergySource("conventionalPropulsion");
            if (sourceFound == true) {
                /* check that energy carrier matches energy carrier of a propulsor */
                // if energy carrier matches, then the propulsor exists
                // go through propulsors and check their energy carrier ID against the ID of this provider
                bool energyCarrierFound = false;
                for (size_t j = 0; j < data_->data.Propulsion.Propulsor.size(); j++) {
                    if (data_->data.Propulsion.Propulsor.at(j).energyCarrierID == data_->energyProviders.at(i).energyCarrier) {
                        energyCarrierFound = true;
                    }
                }
                if (energyCarrierFound == true) {
                    /* check that propulsor can be operated in the given mission phase */
                    // currently the systems are only sized for flight, thus this check is not implemented yet.
                } else {
                    myRuntimeInfo->err << "The energy carrier defined for the energy provider \"propulsor\" does not match the energy carrier of any propulsor of the aircraft."
                                       << std::endl;
                    exit(1);
                }
            } else {
                myRuntimeInfo->err << "The energy source " << data_->energyProviders.at(i).powerTrain << " is not specified in the config file." << std::endl;
                exit(1);
            }
        } else if (data_->energyProviders.at(i).powerTrain == "turbo") {
            /* check energy carrier to determine type of APU (conventional Kerosin or H2)*/
            if (data_->energyProviders.at(i).energyCarrier == 0) { // 0 = kerosene
                /* Check if energy source is specified in the config */
                bool sourceFound = checkEnergySource("conventionalAPU");
                // has to be done when the exact name of the system is known since names in acxml and config do not match
                if (sourceFound == true) {
                    // all checks cleared
                } else {
                    myRuntimeInfo->err << "The energy source " << data_->energyProviders.at(i).powerTrain << " is not specified in the config file." << std::endl;
                    exit(1);
                }
            } else {
                myRuntimeInfo->err << "The energy source with the powertrain " << data_->energyProviders.at(i).powerTrain << " and the energy carrier with ID "
                                   << data_->energyProviders.at(i).energyCarrier << " is not implemented yet." << std::endl;
                exit(1);
            }
        } else if (data_->energyProviders.at(i).powerTrain == "electric") {
            myRuntimeInfo->err << "Electric power source is not implemented yet for systemsDesign." << std::endl;
            exit(1);
        } else if (data_->energyProviders.at(i).powerTrain == "fuel_cell") {
            myRuntimeInfo->err << "Fuel cell power source is not implemented yet for systemsDesign." << std::endl;
            exit(1);
        } else {
            myRuntimeInfo->err << "Energy provider: " << data_->energyProviders.at(i).powerTrain << " is unknown." << std::endl;
            exit(1);
        }
    }
    /* Check if energy source provides all necessary energy forms */
}

bool standardSystemsDesign::checkEnergySource(std::string sourceName) {
    bool sourceFound;
    for (size_t i = 0; i < data_->configuration.Sources.size(); i++) {
        if (data_->configuration.Sources.at(i).systemName == sourceName) {
            sourceFound = true;
        }
    }
    return sourceFound;
}

void standardSystemsDesign::checkUserInputForConsistency() {
    /* Check hydraulic system */
    int ID = data_->Systems.getSpecificSystem("ATA-29");
    data_->Systems.allSystems.at(ID)->checkUserInput();
    /* Check energy sources for electric generators */
    ID = data_->Systems.getSpecificSystem("ATA-24");
    data_->Systems.allSystems.at(ID)->checkUserInput();
    /* Check flight controls */
    ID = data_->Systems.getSpecificSystem("ATA-27");
    data_->Systems.allSystems.at(ID)->checkUserInput();
}

void standardSystemsDesign::checkGeometryForConsistency() {
    bool aileronExists(false);
    for (size_t i(0); i < data_->theWing.at(0).controlSurface.size(); i++) {
        if (data_->theWing.at(0).controlSurface.at(i).name == "aileron" || data_->theWing.at(0).controlSurface.at(i).name == "Aileron"
            || data_->theWing.at(0).controlSurface.at(i).name == "droop_aileron" || data_->theWing.at(0).controlSurface.at(i).name == "DroopAileron")
            aileronExists = true;
    }
    if (!aileronExists) {
        myRuntimeInfo->warn << "There is no Aileron for the wing!" << std::endl;
    }
    /* SystemsDesign specific **/
    if (data_->theLandingGear.empty()) {
        myRuntimeInfo->err << "No landing gears integrated. Please integrate landing gears first" << std::endl;
        exit(1);
    } else {
        if (data_->theLandingGear.front().name != "nose_gear" && data_->theLandingGear.front().name != "nose_gear_assembly") {
            myRuntimeInfo->err << "First landing gear element is not nose landing gear. Abort!" << std::endl;
            exit(1);
        } else {
            if (data_->theLandingGear.at(0).strut_length <= 0.0 || data_->theLandingGear.at(0).strut_length >= 15.0) {
                myRuntimeInfo->err << "Gear.lengthStrut is not in the range of 0 and 15" << std::endl;
                exit(1);
            }
        }
    }
}
