"""Module providing functions for the preparation of user data."""
import pycoordinatesystemconversion as py11csc


def user_method_data_input_preparation(routing_dict):
    """Prepare necessary input data for the user method from aircraft exchange and module configuration files.

    In this function, the user is responsible for preparing the data needed for the user method. Relevant general data
    are obtained from the aircraft exchange file and calculation specific parameter from the module configuration file.
    The user must submit the data in the following format:
        dict = {'parameter_name': [path to parameter node, expected data type], ...}

    :param dict routing_dict: Dictionary containing information on necessary data from module configuration file
    :returns:
        - dict data_to_extract_from_aircraft_exchange_dict: Dictionary containing parameter name, path to parameter,
        and expected data type of parameters to be extracted from aircraft exchange file
        - dict data_to_extract_from_module_configuration_dict: Dictionary containing parameter name, path to parameter,
        and expected data type of parameters to be extracted from module configuration file
    """

    """Aircraft exchange file."""
    # Temporary paths for shorter lines.
    tmp_design_specs_path = './requirements_and_specifications/design_specification/'
    tmp_top_level_aircraft_requirements_path = (
        './requirements_and_specifications/requirements/top_level_aircraft_requirements/')
    tmp_fuselage_entity_path = './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/'
    tmp_fuselage_section_path = tmp_fuselage_entity_path + 'sections/section[@ID="0"]/'
    tmp_fuselage_accommodation_path = tmp_fuselage_entity_path + 'fuselage_accommodation/'
    tmp_fuselage_payload_deck_path = (tmp_fuselage_accommodation_path
                                      + 'payload_tube[@ID="0"]/payload_decks/payload_deck[@ID="0"]/')
    tmp_wing_path = './component_design/wing/specific/geometry/aerodynamic_surface[@ID="0"]/'
    tmp_wing_section_path = tmp_wing_path + 'parameters/sections/section[@ID="0"]/'
    tmp_spar_path = (
        './component_design/wing/specific/geometry/aerodynamic_surface[@ID="0"]/parameters/spars/spar[@ID="0"]/')
    tmp_empennage_path = './component_design/empennage/specific/geometry/aerodynamic_surface[@ID="0"]/'
    tmp_empennage_section_path = tmp_empennage_path + 'parameters/sections/section[@ID="0"]/'
    tmp_empennage_spar_path = (
        './component_design/empennage/specific/geometry/aerodynamic_surface[@ID="0"]/parameters/spars/spar[@ID="0"]/')

    # Enter all parameters to be extracted from the aircraft exchange file.
    data_to_extract_from_aircraft_exchange_dict = {
        # Data necessary for initial energy consumption estimation.
        'number_of_passengers':
            [tmp_design_specs_path + 'transport_task/passenger_definition/total_number_passengers', float],
        'range':
            [tmp_top_level_aircraft_requirements_path + 'design_mission/range', float],
        # Energy information.
        'mission_energy_amount':
            ['./analysis/mission/design_mission/mission_energy[@ID="0"]/consumed_energy', float],
        'mission_energy_ID':
            ['./analysis/mission/design_mission/mission_energy[@ID="0"]/energy_carrier_ID', int],
        'energy_carrier_name':
            [tmp_design_specs_path + 'energy_carriers/energy_carrier[@ID="0"]/type', str],
        # Tank configuration.
        'tank_energy_ID':
            [tmp_design_specs_path + 'configuration/tank_definition/tank[@ID="0"]/energy_carrier_ID', int],
        'tank_location':
            [tmp_design_specs_path + 'configuration/tank_definition/tank[@ID="0"]/location', str],
        'tank_position':
            [tmp_design_specs_path + 'configuration/tank_definition/tank[@ID="0"]/position', str],
        'energy_share':
            [tmp_design_specs_path + 'configuration/tank_definition/tank[@ID="0"]/energy_share', float],
        # Fuselage data.
        'fuselage_position_x': ['./component_design/fuselage/position/x', float],
        'fuselage_position_y': ['./component_design/fuselage/position/y', float],
        'fuselage_position_z': ['./component_design/fuselage/position/z', float],
        'fuselage_entity_position_x': [tmp_fuselage_entity_path + 'position/x', float],
        'fuselage_entity_position_y': [tmp_fuselage_entity_path + 'position/y', float],
        'fuselage_entity_position_z': [tmp_fuselage_entity_path + 'position/z', float],
        'fuselage_section_origin_x': [tmp_fuselage_section_path + 'origin/x', float],
        'fuselage_section_origin_y': [tmp_fuselage_section_path + 'origin/y', float],
        'fuselage_section_origin_z': [tmp_fuselage_section_path + 'origin/z', float],
        'fuselage_section_upper_height': [tmp_fuselage_section_path + 'upper_height', float],
        'fuselage_section_lower_height': [tmp_fuselage_section_path + 'lower_height', float],
        'fuselage_section_width': [tmp_fuselage_section_path + 'width', float],
        'fuselage_payload_deck_origin_x': [tmp_fuselage_payload_deck_path + 'position/x', float],
        'fuselage_payload_deck_origin_y': [tmp_fuselage_payload_deck_path + 'position/y', float],
        'fuselage_payload_deck_origin_z': [tmp_fuselage_payload_deck_path + 'position/z', float],
        'fuselage_payload_deck_structural_floor_thickness': [tmp_fuselage_payload_deck_path +
                                                             'payload_deck_structural_floor_thickness', float],
        # Wing data.
        'wing_name': [tmp_wing_path + 'name', str],
        'wing_position_x': ['./component_design/wing/position/x', float],
        'wing_position_y': ['./component_design/wing/position/y', float],
        'wing_position_z': ['./component_design/wing/position/z', float],
        'wing_symmetry': [tmp_wing_path + 'parameters/symmetric', bool],
        'wing_section_chord_origin_x': [tmp_wing_section_path + 'chord_origin/x', float],
        'wing_section_chord_origin_y': [tmp_wing_section_path + 'chord_origin/y', float],
        'wing_section_chord_origin_z': [tmp_wing_section_path + 'chord_origin/z', float],
        'wing_section_chord_length': [tmp_wing_section_path + 'chord_length', float],
        'wing_section_profile': [tmp_wing_section_path + 'profile', str],
        # Wing spar data.
        'spar_position_inner_spanwise': [tmp_spar_path + 'position/inner_position/spanwise', float],
        'spar_position_inner_chord': [tmp_spar_path + 'position/inner_position/chord/to', float],
        'spar_position_outer_spanwise': [tmp_spar_path + 'position/outer_position/spanwise', float],
        'spar_position_outer_chord': [tmp_spar_path + 'position/outer_position/chord/to', float],
        # Empennage data.
        'empennage_name': [tmp_empennage_path + 'name', str],
        'empennage_position_x': ['./component_design/empennage/position/x', float],
        'empennage_position_y': ['./component_design/empennage/position/y', float],
        'empennage_position_z': ['./component_design/empennage/position/z', float],
        'empennage_symmetry': [tmp_empennage_path + 'parameters/symmetric', bool],
        'empennage_entity_position_x': [tmp_empennage_path + 'position/x', float],
        'empennage_entity_position_y': [tmp_empennage_path + 'position/y', float],
        'empennage_entity_position_z': [tmp_empennage_path + 'position/z', float],
        'empennage_section_chord_origin_x': [tmp_empennage_section_path + 'chord_origin/x', float],
        'empennage_section_chord_origin_y': [tmp_empennage_section_path + 'chord_origin/y', float],
        'empennage_section_chord_origin_z': [tmp_empennage_section_path + 'chord_origin/z', float],
        'empennage_section_chord_length': [tmp_empennage_section_path + 'chord_length', float],
        'empennage_section_profile': [tmp_empennage_section_path + 'profile', str],
        # Empennage spar data.
        'empennage_spar_position_inner_spanwise': [tmp_empennage_spar_path +
                                                   'position/inner_position/spanwise', float],
        'empennage_spar_position_inner_chord': [tmp_empennage_spar_path + 'position/inner_position/chord/to', float],        
        'empennage_spar_position_outer_spanwise': [tmp_empennage_spar_path +
                                                   'position/outer_position/spanwise', float],
        'empennage_spar_position_outer_chord': [tmp_empennage_spar_path + 'position/outer_position/chord/to', float],  
    }

    """Module configuration file."""
    # Enter all general parameters to be extracted from the module configuration file. 'general parameters' means
    # parameters that do not differ according to the user layer. It should be noted that 'tmp_general' is only used to
    # shorten the path information in the 'general_data_to_extract_from_module_configuration_dict'.
    tmp_general = ('./program_settings/configuration[@ID="tube_and_wing"]/fidelity[@ID="empirical"]/'
                   + 'tank_design_tu_berlin/general/')
    general_data_to_extract_from_module_configuration_dict = {
        'mass_technology_factor': [tmp_general + 'mass_technology_factor', float]
    }

    # Enter all specific parameters to be extracted from the module configuration file. 'specific parameters' means
    # parameters that differ according to the user layer. It should be noted that 'tmp_specific' is only used to
    # shorten the path information in the 'specific_data_to_extract_from_module_configuration_dict'.
    tmp_specific = ('./program_settings/configuration[@ID="tube_and_wing"]/fidelity[@ID="empirical"]/'
                   + 'tank_design_tu_berlin/specific/')
    specific_data_to_extract_from_module_configuration_dict = {
        'kerosene_gravimetric_energy_density':
            [tmp_specific + 'kerosene_gravimetric_energy_density', float],
        'kerosene_volumetric_energy_density':
            [tmp_specific + 'kerosene_volumetric_energy_density', float],
        'liquid_hydrogen_gravimetric_energy_density':
            [tmp_specific + 'liquid_hydrogen_tank_structure/liquid_hydrogen_gravimetric_energy_density', float],
        'liquid_hydrogen_volumetric_energy_density':
            [tmp_specific + 'liquid_hydrogen_tank_structure/liquid_hydrogen_volumetric_energy_density', float],
        'material_wall':
            [tmp_specific + 'liquid_hydrogen_tank_structure/material_wall', str],
        'density_wall':
            [tmp_specific + 'liquid_hydrogen_tank_structure/density_wall', float],
        'thickness_wall_calculation_method':
            [tmp_specific + 'liquid_hydrogen_tank_structure/thickness_wall_calculation_method', str],
        'mass_pumps':
            [tmp_specific + 'liquid_hydrogen_tank_structure/mass_pumps', float],
        'mass_baffle':
            [tmp_specific + 'liquid_hydrogen_tank_structure/mass_baffle', float],
        'type_insulation':
            [tmp_specific + 'liquid_hydrogen_tank_insulation/type_insulation', str],
        'material_insulation':
            [tmp_specific + 'liquid_hydrogen_tank_insulation/material_insulation', str],
        'thickness_insulation':
            [tmp_specific + 'liquid_hydrogen_tank_insulation/thickness_insulation', float],
        'density_insulation':
            [tmp_specific + 'liquid_hydrogen_tank_insulation/density_insulation', float],
        'type_end_cap':
            [tmp_specific + 'liquid_hydrogen_tank_design_parameter/type_end_cap', str],
        'internal_pressure':
            [tmp_specific + 'liquid_hydrogen_tank_design_parameter/internal_pressure', float],
        'factor_usable_diameter':
            [tmp_specific + 'miscellaneous/factor_usable_diameter', float],
        'factor_volume_allowance':
            [tmp_specific + 'miscellaneous/factor_volume_allowance', float],
        'a_to_d_factor':
            [tmp_specific + 'kerosene_tank_design_parameter/a_to_d_factor', float],
        'factor_volume_usable':
            [tmp_specific + 'kerosene_tank_design_parameter/factor_volume_usable', float],
        'temperature_expansion_allowance':
            [tmp_specific + 'kerosene_tank_design_parameter/temperature_expansion_allowance', float],
        # Obelisk calculation method name.
        'obelisk_volume_calculation_method':
            [tmp_specific + 'kerosene_tank_design_parameter/obelisk_calculation_method', str],
        'buffer_inner_tank_segment':
            [tmp_specific + 'kerosene_tank_design_parameter/buffer_inner_tank_segment', float],
        'buffer_outer_tank_segment':
            [tmp_specific + 'kerosene_tank_design_parameter/buffer_outer_tank_segment', float],
        'buffer_center_tank_segment':
            [tmp_specific + 'kerosene_tank_design_parameter/buffer_center_tank_segment', float]
    }

    # Merge module configuration dictionaries.
    data_to_extract_from_module_configuration_dict = (general_data_to_extract_from_module_configuration_dict
                                                      | specific_data_to_extract_from_module_configuration_dict)

    return data_to_extract_from_aircraft_exchange_dict, data_to_extract_from_module_configuration_dict


def user_method_data_output_preparation(data_dict):
    """Prepare user-specific output data based on the calculation method results.

    This function is responsible for preparing the user-specific output data based on the results of the calculation
    method. The 'data_dict' input parameter contains the results of the module execution.
    The data for the key parameters output must be specified in the following format in order to be written correctly
    to the aircraft exchange file by the 'write_key_data_to_aircraft_exchange_file' function in the following step:
        dict = {'parameter_name': [path to parameter node, value, name (if needed)], ...}
    Important notes:
        (1) It should be noted that only key parameters may be written that have been previously defined by the module
        manager.
        (2) Attention must be paid to the proper path specifications, otherwise warnings may be issued or, in the worst
        case, errors may occur subsequently resulting in the write process and consequently the entire program being
        aborted.
        (3) If the path specifications contain IDs, these must start at '0' and be defined in ascending order without
        gaps.
    For the method-specific output, a path list and a dictionary is necessary to properly write the data to the
    method-specific XML file.
    the dictionary must be specified in the following format:
        dict = ...
    Note: If the user wants to export data from the design and study mission, two path lists and dictionaries are
    necessary.

    :param dict data_dict: Dictionary containing the results of the module execution
    :returns:
        - dict key_output_dict: Output dictionary containing key parameters that are written to aircraft XML file
        - dict method_specific_output_dict: Dictionary containing specific parameters that are written to
        method-specific output XML
    """

    """ Convert coordinates from aircraft coordinate system to CGAL."""
    # Implement function here...

    """Key parameters output."""
    total_position_path = './component_design/tank/position/'
    total_mass_path = './component_design/tank/mass_properties/'
    total_inertia_path = './component_design/tank/mass_properties/inertia/'
    total_cog_path = './component_design/tank/mass_properties/center_of_gravity/'
    # data_dict['total_tank_parameters']['position'] = \
        # coordinate_system_conversion(data_dict['total_tank_parameters']['position'])
    
    key_output_dict = {
        # Total tank parameter.
        'x':
            [total_position_path + 'x',
             data_dict['total_tank_parameters']['position']['x']],
        'y':
            [total_position_path + 'y',
             data_dict['total_tank_parameters']['position']['y']],
        'z':
            [total_position_path + 'z',
             data_dict['total_tank_parameters']['position']['z']],
        'mass':
            [total_mass_path + 'mass',
             data_dict['total_tank_parameters']['mass']],
        'j_xx':
            [total_inertia_path + 'j_xx',
             data_dict['total_tank_parameters']['inertia']['j_xx']],
        'j_yy':
            [total_inertia_path + 'j_yy',
             data_dict['total_tank_parameters']['inertia']['j_yy']],
        'j_zz':
            [total_inertia_path + 'j_zz',
             data_dict['total_tank_parameters']['inertia']['j_zz']],
        'j_xy':
            [total_inertia_path + 'j_xy',
             data_dict['total_tank_parameters']['inertia']['j_xy']],
        'j_xz':
            [total_inertia_path + 'j_xz',
             data_dict['total_tank_parameters']['inertia']['j_xz']],
        'j_yx':
            [total_inertia_path + 'j_yx',
             data_dict['total_tank_parameters']['inertia']['j_yx']],
        'j_yz':
            [total_inertia_path + 'j_yz',
             data_dict['total_tank_parameters']['inertia']['j_yz']],
        'j_zx':
            [total_inertia_path + 'j_zx',
             data_dict['total_tank_parameters']['inertia']['j_zx']],
        'j_zy':
            [total_inertia_path + 'j_zy',
             data_dict['total_tank_parameters']['inertia']['j_zy']],
        'cog_x':
            [total_cog_path + 'x',
             data_dict['total_tank_parameters']['center_of_gravity']['x']],
        'cog_y':
            [total_cog_path + 'y',
             data_dict['total_tank_parameters']['center_of_gravity']['y']],
        'cog_z':
            [total_cog_path + 'z',
             data_dict['total_tank_parameters']['center_of_gravity']['z']]
    }

    # Generate dictionary for each tank entity.
    j = 0
    while j < data_dict['number_of_tanks']:
        tank_path_id = './component_design/tank/specific/tank[@ID="' + str(j) + '"]/'
        tank_id = 'tank_' + str(j)
        data_dict[tank_id]['geometry']['total']['position'] = \
            coordinate_system_conversion(data_dict[tank_id]['geometry']['total']['position'])
        tank_output_dict = {
            'name_id' + str(j):
                [tank_path_id + 'name',
                 'tank_' + str(j)],
            'designator_id' + str(j):
                [tank_path_id + 'designator',
                 data_dict[tank_id]['tank_location'] + '_' + data_dict[tank_id]['tank_position']],
            'x_id' + str(j):
                [tank_path_id + 'position/x',
                 round(data_dict[tank_id]['geometry']['total']['position']['x'], 9)],
            'y_id' + str(j):
                [tank_path_id + 'position/y',
                 round(data_dict[tank_id]['geometry']['total']['position']['y'], 9)],
            'z_id' + str(j):
                [tank_path_id + 'position/z',
                 round(data_dict[tank_id]['geometry']['total']['position']['z'], 9)],
            'direction_x_id' + str(j):
                [tank_path_id + 'direction/x',
                 data_dict[tank_id]['geometry']['total']['direction']['x']],
            'direction_y_id' + str(j):
                [tank_path_id + 'direction/y',
                 data_dict[tank_id]['geometry']['total']['direction']['y']],
            'direction_z_id' + str(j):
                [tank_path_id + 'direction/z',
                 data_dict[tank_id]['geometry']['total']['direction']['z']],
            'mass_id' + str(j):
                [tank_path_id + 'mass_properties/mass',
                 data_dict[tank_id]['geometry']['total']['mass']],
            'j_xx_id' + str(j):
                [tank_path_id + 'mass_properties/inertia/j_xx',
                 data_dict[tank_id]['geometry']['total']['inertia']['j_xx']],
            'j_yy_id' + str(j):
                [tank_path_id + 'mass_properties/inertia/j_yy',
                 data_dict[tank_id]['geometry']['total']['inertia']['j_yy']],
            'j_zz_id' + str(j):
                [tank_path_id + 'mass_properties/inertia/j_zz',
                 data_dict[tank_id]['geometry']['total']['inertia']['j_zz']],
            'j_xy_id' + str(j):
                [tank_path_id + 'mass_properties/inertia/j_xy',
                 data_dict[tank_id]['geometry']['total']['inertia']['j_xy']],
            'j_xz_id' + str(j):
                [tank_path_id + 'mass_properties/inertia/j_xz',
                 data_dict[tank_id]['geometry']['total']['inertia']['j_xz']],
            'j_yx_id' + str(j):
                [tank_path_id + 'mass_properties/inertia/j_yx',
                 data_dict[tank_id]['geometry']['total']['inertia']['j_yx']],
            'j_yz_id' + str(j):
                [tank_path_id + 'mass_properties/inertia/j_yz',
                 data_dict[tank_id]['geometry']['total']['inertia']['j_yz']],
            'j_zx_id' + str(j):
                [tank_path_id + 'mass_properties/inertia/j_zx',
                 data_dict[tank_id]['geometry']['total']['inertia']['j_zx']],
            'j_zy_id' + str(j):
                [tank_path_id + 'mass_properties/inertia/j_zy',
                 data_dict[tank_id]['geometry']['total']['inertia']['j_zy']],
            'cog_x_id' + str(j):
                [tank_path_id + 'mass_properties/center_of_gravity/x',
                 data_dict[tank_id]['geometry']['total']['center_of_gravity']['x']],
            'cog_y_id' + str(j):
                [tank_path_id + 'mass_properties/center_of_gravity/y',
                 data_dict[tank_id]['geometry']['total']['center_of_gravity']['y']],
            'cog_z_id' + str(j):
                [tank_path_id + 'mass_properties/center_of_gravity/z',
                 data_dict[tank_id]['geometry']['total']['center_of_gravity']['z']],
            'energy_id' + str(j):
                [tank_path_id + 'maximum_energy_capacity',
                 data_dict[tank_id]['energy_available']],
            'centroid_x_id' + str(j):
                [tank_path_id + 'mass_properties/centroid/x',
                 round(data_dict[tank_id]['geometry']['total']['centroid']['x'], 9)],
            'centroid_y_id' + str(j):
                [tank_path_id + 'mass_properties/centroid/y',
                 round(data_dict[tank_id]['geometry']['total']['centroid']['y'], 9)],
            'centroid_z_id' + str(j):
                [tank_path_id + 'mass_properties/centroid/z',
                 round(data_dict[tank_id]['geometry']['total']['centroid']['z'], 9)],
        }

        # Iterate over tank sections of each tank entity.
        i = 0
        if data_dict[tank_id]['energy_carrier_name'] == 'kerosene':
            number_of_sections = 2
            section_names = ['inner_surface', 'outer_surface']
            if data_dict[tank_id]['tank_location'] == 'fuselage' and data_dict[tank_id]['tank_position'] == 'center':
                number_of_sections = sum(1 for k in data_dict[tank_id]['geometry'].keys()
                                     if k.startswith('cross_section_'))
                section_names = [k for k in data_dict[tank_id]['geometry'].keys() if k.startswith('cross_section_')]
                section_names = sorted(section_names, key=lambda x: int(x.split('_')[-1]))
            if data_dict[tank_id]['tank_location'] == 'horizontal_stabilizer' and (
                data_dict[tank_id]['tank_position'] == 'total'):
                number_of_sections = sum(1 for k in data_dict[tank_id]['geometry'].keys()
                                     if k.startswith('cross_section_'))
                section_names = [k for k in data_dict[tank_id]['geometry'].keys() if k.startswith('cross_section_')]
                section_names = sorted(section_names, key=lambda x: int(x.split('_')[-1]))
        else:
            number_of_sections = sum(1 for k in data_dict[tank_id]['geometry'].keys()
                                     if k.startswith('cross_section_'))
            section_names = [k for k in data_dict[tank_id]['geometry'].keys() if k.startswith('cross_section_')]
            section_names = sorted(section_names, key=lambda x: int(x.split('_')[-1]))
        while i < number_of_sections:
            tank_section_path = tank_path_id + 'geometry/cross_section[@ID="' + str(i) + '"]/'
            tank_section_name = 'tank_section_' + str(i)
            # data_dict[tank_id]['geometry'][section_names[i]]['position'] = \
               #  coordinate_system_conversion(data_dict[tank_id]['geometry'][section_names[i]]['position'])
            # Check if the energy carrier of current tank is kerosene 
            #  -> if true: -> use coordinate transformation for wing tanks.
            if data_dict['tank_' + str(j)]['energy_carrier_name'] == 'kerosene':
                tank_section_output_dict = {
                    'cross_section_name_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'name',
                        tank_section_name],
                    'cross_section_x_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'position/x',
                        round(data_dict[tank_id]['geometry'][section_names[i]]['position']['x'], 9)],
                    'cross_section_y_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'position/y',
                        round(data_dict[tank_id]['geometry'][section_names[i]]['position']['z'], 9)],
                    'cross_section_z_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'position/z',
                        round(data_dict[tank_id]['geometry'][section_names[i]]['position']['y'] * -1, 9)],
                    'shape_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'shape',
                        data_dict[tank_id]['geometry'][section_names[i]]['shape']],
                    'height_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'height',
                        data_dict[tank_id]['geometry'][section_names[i]]['height']],
                    'width_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'width',
                        data_dict[tank_id]['geometry'][section_names[i]]['width']],
                    'length_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'length',
                        data_dict[tank_id]['geometry'][section_names[i]]['length']]
                }
            # Else condition: The energy carrier of current tank is not kerosene 
            #  -> use coordinate transformation for fuselage tanks.
            else:
                tank_section_output_dict = {
                    'cross_section_name_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'name',
                        tank_section_name],
                    'cross_section_x_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'position/x',
                        round(data_dict[tank_id]['geometry'][section_names[i]]['position']['y'] * -1, 9)],
                    'cross_section_y_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'position/y',
                        round(data_dict[tank_id]['geometry'][section_names[i]]['position']['z'], 9)],
                    'cross_section_z_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'position/z',
                        round(data_dict[tank_id]['geometry'][section_names[i]]['position']['x'], 9)],
                    'shape_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'shape',
                        data_dict[tank_id]['geometry'][section_names[i]]['shape']],
                    'height_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'height',
                        data_dict[tank_id]['geometry'][section_names[i]]['height']],
                    'width_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'width',
                        data_dict[tank_id]['geometry'][section_names[i]]['width']],
                    'length_id' + str(j) + '_id' + str(i):
                        [tank_section_path + 'length',
                        data_dict[tank_id]['geometry'][section_names[i]]['length']]
                }
            # merge dictionary to one single output dictionary
            tank_output_dict.update(tank_section_output_dict)
            i += 1

        key_output_dict.update(tank_output_dict)
        j += 1

    # Add additional fuselage length.
    additional_length_dict = {
        'additional_fuselage_length': ['./component_design/tank/specific/additional_fuselage_length',
                                       data_dict['total_tank_parameters']['additional_fuselage_length']]
    }

    key_output_dict.update(additional_length_dict)

    """Method-specific output."""
    method_specific_output_dict = {}
    for key, value in key_output_dict.items():
        if 'mass_properties' not in value[0]:
            method_specific_output_dict[key] = value

    return key_output_dict, method_specific_output_dict


def coordinate_system_conversion(element):
    if len(element.keys()) > 3:
        element_3d = py11csc.Element3D(element["j_xx"], element["j_xy"], element["j_xz"],
                                       element["j_yx"], element["j_yy"], element["j_yz"],
                                       element["j_zx"], element["j_zy"], element["j_zz"])
        element_3d.AC2CGAL()
        for k in element.keys():
            element[k] = getattr(element_3d, k.split("_")[-1])()
        return element
    else:
        element_3d = py11csc.Element3D(element["x"], element["y"], element["z"])
        element_3d.AC2CGAL()
        for k in element.keys():
            element[k] = getattr(element_3d, k)()
        return element
