"""Module providing plotting functionalities for current calculation method."""
# Import standard libraries.
import os
from matplotlib import pyplot as plt
import numpy as np


def method_plot(paths_and_names, routing_dict, data_dict, method_specific_output_dict, runtime_output):
    """Plot function.

    This function is responsible for creating plots.
    [Add further information here...]

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameter
    :param dict data_dict: Dictionary containing results of module execution
    :return: None
    """

    # This is just a dummy code snippet. Insert your code here.
    tool_name = paths_and_names['tool_name']

    aircraft_group_masses = []
    ome = data_dict["operating_mass_empty"].mass
    mtom = data_dict["maximum_takeoff_mass"].mass
    for key, mass in data_dict["group_masses"].items():
        name = key.replace("_", " ").title()
        aircraft_group_masses.append(
            (name, mass, mass/ome * 100.0, mass/mtom * 100.0))

    # Extract components and their values from data_dict
    components = list(data_dict["group_masses"].keys())
    component_values = list(data_dict["group_masses"].values())  # Mass values in kg

    structure_components = list(data_dict["structure_components"].keys())
    structure_component_values = list(data_dict["structure_components"].values())  # Mass values in kg

    # Colors for each component
    colors = ['red', 'yellow', 'blue', 'orange', 'green']
    #colors_structure = ['darkgreen', 'chartreuse', 'teal', 'aquamarine', 'olive']
    colors_structure = ['0.35', '0.45', '0.57', '0.80', '0.9']

    # Setting color characteristics
    plt.rcParams['text.color'] = '#ecf0f1'
    plt.rcParams['xtick.color'] = '#ecf0f1'
    plt.rcParams['ytick.color'] = '#ecf0f1'

    # Plotting
    fig, ax = plt.subplots(figsize=(8, 5), facecolor='#494f4f')

    # Bar positions
    ind = np.arange(len(components) + 1)  # +1 for the total OEM bar

    # Create stacked bar for the total OEM
    bottom = 0
    for i in range(len(components)):
        ax.bar(0, component_values[i], bottom=bottom,
            width=0.5, color=colors[i], label=components[i])
        bottom += component_values[i]
    
    # Create bars for each individual component (not stacked)
    for i in range(len(components)):
        ax.bar(ind[i + 1], component_values[i], width=0.5, color=colors[i], label=f'{components[i]}')

    # Create stacked bar for the structure components
    bottom = 0
    for i in range(len(structure_components)):
        ax.bar(1, structure_component_values[i], bottom=bottom,
            width=0.5, color=colors_structure[i], label=structure_components[i])
        bottom += structure_component_values[i]

    # Configure the plot
    ax.set_ylabel('Mass (kg)', color='#ecf0f1')
    ax.set_title('Operating Mass Empty and its Components', color='#ecf0f1')
    ax.set_xticks(ind)
    ax.set_xticklabels(['OME'] + components)

    # Manually add legend entries for structure components and others
    handles, labels = ax.get_legend_handles_labels()
    legend_labels = components + structure_components
    handles = [plt.Rectangle((0, 0), 1, 1, color=color) for color in (colors + colors_structure)]
    legend = ax.legend(handles, legend_labels, loc='upper right', fontsize='small')

    # Set the text in the legend to black
    for text in legend.get_texts():
        text.set_color('black')

    # Adjusting the plot's edge color
    for spine in ax.spines.values():
        spine.set_edgecolor('#ecf0f1')

    # Save the figure
    plt.savefig("./aircraft_group_masses.svg", format="svg")

    # Loading chart

    plt.figure(figsize=(8, 5), facecolor='#494f4f')

    plt.plot(data_dict["Refueling CG change"],
             data_dict["Refueling mass change"], 'y')  # Refueling

    plt.plot(data_dict["Passengers boarding front back CG change"],
             data_dict["Passengers boarding front back mass change"], 'b--')
    plt.plot(data_dict["Passengers boarding back front CG change"],
             data_dict["Passengers boarding back front mass change"], 'b-.')

    plt.plot(data_dict["Cargo loading front back CG change"],
             data_dict["Cargo loading front back mass change"], 'c--')
    plt.plot(data_dict["Cargo loading back front CG change"],
             data_dict["Cargo loading back front mass change"], 'c-.')

    if data_dict["Defueling mode"] == 'mode_1':
        plt.plot(data_dict["Defueling CG change"],
                 data_dict["Defueling mass change"], 'm')  # Defueling
        legend_labels = ['Refueling', 'Pax from front', 'Pax from back', 'Cargo from front', 'Cargo from back',
                         'Defueling', 'OME', 'MTOM', 'Neutral point Ma=0,83',
                          'Most fwd CG', 'Most aft CG', 'NP-5%']
    else:
        legend_labels = ['Refueling', 'Pax from front', 'Pax from back', 'Cargo from front', 'Cargo from back',
                         'OME', 'MTOM', 'Neutral point Ma=0,83',
                          'Most fwd CG', 'Most aft CG', 'NP-5%']

    N = np.arange(data_dict["Most forward CG mac position"]-5, max(
        [data_dict["Most aft CG mac position"], -data_dict["Neutral point mac position"]*100])+5)
    plt.hlines(data_dict["operating_mass_empty"].mass,
               N.min(), N.max(), colors='gray', linestyles='--')
    plt.hlines(data_dict["maximum_takeoff_mass"].mass,
               N.min(), N.max(), colors='brown', linestyles='--')

    plt.xlim(data_dict["Most forward CG mac position"]-5, max(
        [data_dict["Most aft CG mac position"], -data_dict["Neutral point mac position"]*100])+5)

    M = np.arange(ome-100, mtom+100)
    plt.plot(-np.ones(len(M)) *
             data_dict["Neutral point mac position"]*100, M, 'r')

    plt.plot(np.ones(len(M)) *
             data_dict["Most forward CG mac position"], M, 'g')
    plt.plot(np.ones(len(M)) *
             data_dict["Most aft CG mac position"], M, 'limegreen')

    plt.plot(-np.ones(len(M)) *
             data_dict["Neutral point mac position"]*100 - 5, M, 'r-.')

    # Define legend labels and colors

    legend = plt.legend(legend_labels, loc='upper left', fontsize='small')
    for text in legend.get_texts():
        text.set_color('black')

    plt.grid(visible=False, color='#ecf0f1')

    plt.title("Loading Diagramm")
    plt.xlabel("CG x-position over MAC in %", color='#ecf0f1')
    plt.ylabel("Mass [kg]", color='#ecf0f1')

    plt.savefig("./weight_and_balance_diagramm.svg", format="svg")

    # Plot different mass configuration cases:

    plt.figure(figsize=(8, 5), facecolor='#494f4f')

    plt.plot(data_dict["operating_mass_empty"].cg_mac,
             data_dict["operating_mass_empty"].mass, '+')  

    plt.plot(data_dict["maximum_zero_fuel_mass"].cg_mac,
             data_dict["maximum_zero_fuel_mass"].mass, 'x')
    
    plt.plot(data_dict["mtow_with_max_payload"].cg_mac,
             data_dict["mtow_with_max_payload"].mass, '*')

    plt.plot(data_dict["mtow_with_max_fuel"].cg_mac,
             data_dict["mtow_with_max_fuel"].mass, 'D')
    
    plt.plot(data_dict["maximum_fuel_mass"].cg_mac,
             data_dict["maximum_fuel_mass"].mass,'o')
    
    plt.plot(data_dict["design_mass"].cg_mac,
             data_dict["design_mass"].mass, 's')    

    # plt.xlim(data_dict["Most forward CG mac position"]-5, max(
    #     [data_dict["Most aft CG mac position"], -data_dict["Neutral point mac position"]*100])+5)

    # Define legend labels and colors
    legend_labels2 = ['OME', 'MZFM', 'max Payload + Fuel', 'max Fuel + Payload', 'Maximum Fuel Mass',
                         'Design Fuel + Design Payload']
    legend = plt.legend(legend_labels2, loc='upper left', fontsize='small')
    for text in legend.get_texts():
        text.set_color('black')

    plt.grid(visible=False, color='#ecf0f1')

    plt.title("Loading Cases")
    plt.xlabel("CG x-position over MAC in %", color='#ecf0f1')
    plt.ylabel("Mass [kg]", color='#ecf0f1')
    #plt.show
    plt.savefig("./cg_cases.svg", format="svg")


    runtime_output.print("plotted")
