#include "cantileverLow.h"

#include <CGAL/Surface_mesh/IO/PLY.h>
#include <aircraftGeometry2/geometry/factory.h>
#include <aircraftGeometry2/io/convert.h>
#include <aircraftGeometry2/io/dat.h>
#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <atmosphere/atmosphere.h>
#include <moduleBasics/modeSelector.h>
#include <unitConversion/constants.h>

#include <map>
#include <numbers>
#include <string>
#include <tuple>

#include "lib/design_methods/aspectRatioMethods.h"
#include "lib/design_methods/dihedralMethods.h"
#include "lib/design_methods/sweepAngleMethods.h"
#include "lib/design_methods/taperRatioMethods.h"
#include "lib/design_methods/wingAreaMethods.h"
#include "lib/design_methods/wingProfileMethods.h"
#include "lib/general_methods/newtonAlgorithm.h"

namespace low {
namespace cantilever {
WingDimensioner::WingDimensioner() {}

void WingDimensioner::design() {}

void WingDimensioner::ruwisch_method(const double mach_ica, const double mach_mmo, const double design_altitude,
                                     const std::string& airfoildatadir, std::vector<std::string>& wing_profile_names,
                                     const double user_defined_taper_ratio, const double user_defined_wing_area,
                                     const double maximum_takeoff_mass, const double wing_loading) {
  bool wing_position_low = true;
  bool wing_mounted_landing_gear = true;

  bool kinked_wing = wing_position_low && wing_mounted_landing_gear ? true : false;
  bool taper_ratio_auto_flag = true;

  double max_thickness_to_chord_ratio = 0.15;  // get from airfoil profile
  // Setup values current
  double delta_drag_divergence_to_cruise = 0.02;  // See wing design script
  double user_aspect_ratio = 10.5;
  double user_quarter_chord_sweep = 25 * geom2::detail::to_radians;
  double max_trailing_edge_inner_sweep = 10 * geom2::detail::to_radians;
  double technology_factor_dd = 0.96;
  double quarter_chord_sweep =
      get_wing_quarter_chord_sweep(1, user_quarter_chord_sweep, mach_ica, wing_loading, design_altitude,
                                   max_thickness_to_chord_ratio, delta_drag_divergence_to_cruise, technology_factor_dd);

  double span_limit_max = 34;
  double aspect_ratio_max = get_design_aspect_ratio(0, user_aspect_ratio, quarter_chord_sweep);

  double installed_wing_area =
      get_design_wing_area(1, user_defined_wing_area, maximum_takeoff_mass, wing_loading);  // add mode
  double installed_span = span_limit_max;
  double installed_aspect_ratio = pow(installed_span, 2.) / installed_wing_area;
  // Check for aspect ratio
  if (installed_aspect_ratio > aspect_ratio_max) {
    std::cout << "Aspect ratio with max span > than pitch up aspect ratio\n";
    std::cout << "Switch to max aspect ratio\n";
    installed_aspect_ratio = aspect_ratio_max;
    installed_span = sqrt(installed_aspect_ratio * installed_wing_area);
  } else {
    std::cout << "Span limit check ... [OK]\n";
  }
  double taper_ratio = get_design_taper_ratio(1, user_defined_taper_ratio, installed_aspect_ratio, quarter_chord_sweep);

  // Prepare geometry parameter
  const double DUMMY_fuselage_width = 5.;
  double half_fuselage_width = DUMMY_fuselage_width / 2.;  // Parameter from acexchange file
  double half_span = installed_span / 2.;
  double half_wing_area = installed_wing_area / 2.;
  double eta_kink = 0.3;
  double eta_fuselage = half_fuselage_width / half_span;
  double eta_tip = half_span / half_span;  // 1 - might change if wing tip devices will be added

  // Geometry computation functions
  auto condition_trailing_edge_max = [=](const double inner_taper_ratio) {
    return (4. / 3. * (tan(quarter_chord_sweep) - tan(max_trailing_edge_inner_sweep)) * half_span *
            (eta_kink - eta_fuselage) / (1 - inner_taper_ratio));
  };
  auto root_chord_function = [=](const double inner_taper_ratio) {
    return installed_wing_area / (2. * half_span *
                                  (eta_fuselage + (1 + inner_taper_ratio) * 0.5 * (eta_kink - eta_fuselage) +
                                   (inner_taper_ratio + taper_ratio) * 0.5 * (eta_tip - eta_kink)));
  };

  // Maximum possible taper ratio
  const double eta_ratio = (eta_tip - eta_kink) / (eta_kink - eta_fuselage);
  const double taper_ratio_inner_max = (eta_ratio + taper_ratio) / (1 + eta_ratio);
  auto taper_ratio_inner_searcher = [=](const double inner_taper_ratio) {
    return root_chord_function(inner_taper_ratio) - condition_trailing_edge_max(inner_taper_ratio);
  };
  // Find root_chord which fulfills - half_wing_area = half_inner_wing_area + half_outer_wing_area + half_fuselage_area
  double taper_ratio_inner = newton_algorithm(taper_ratio_inner_searcher, taper_ratio_inner_max);
  if (taper_ratio_inner > taper_ratio_inner_max) {
    taper_ratio_inner = taper_ratio_inner_max;
  }
  const double taper_ratio_outer = taper_ratio / taper_ratio_inner;
  const double root_chord = root_chord_function(taper_ratio_inner);
  // Compute values for wing geometry
  const double kink_chord = root_chord * taper_ratio_inner;
  const double tip_chord = root_chord * taper_ratio;
  const double fuselage_area = 2. * eta_fuselage * half_span * root_chord;
  const double inner_wing_area = (eta_kink - eta_fuselage) * (1 + taper_ratio_inner) * half_span * root_chord;
  const double outer_wing_area = 2. * (eta_tip - eta_kink) * (taper_ratio_inner + taper_ratio) * half_span * root_chord;

  // Inner wing
  const double tan_inner_wing_quarter_chord_sweep = tan(quarter_chord_sweep);
  const double tan_inner_wing_leading_edge_sweep =
      tan_inner_wing_quarter_chord_sweep +
      0.25 * root_chord / (half_span * (eta_kink - eta_fuselage)) * (1 - taper_ratio_inner);
  const double tan_inner_wing_trailing_edge_sweep =
      tan_inner_wing_quarter_chord_sweep -
      0.75 * root_chord / (half_span * (eta_kink - eta_fuselage)) * (1 - taper_ratio_inner);

  // Outer wing
  const double tan_outer_wing_quarter_chord_sweep = tan_inner_wing_quarter_chord_sweep;
  const double tan_outer_wing_leading_edge_sweep =
      tan_outer_wing_quarter_chord_sweep +
      0.25 * (root_chord * taper_ratio_inner) / (half_span * (eta_tip - eta_kink)) * (1 - taper_ratio_outer);
  const double tan_outer_wing_trailing_edge_sweep =
      tan_inner_wing_quarter_chord_sweep -
      0.75 * root_chord * taper_ratio_inner / (half_span * (eta_tip - eta_kink)) * (1 - taper_ratio_outer);

  // Check different conditions

  // Check sweeps leading edge and trailing edge by single equation
  // Condition 1: leading edge inner wing >= leading edge outer wing
  // Condition 2: trailing edge inner wing <= trailing edge outer wing
  // K_eta * (1-taper_ratio_2) >= taper_ratio_2 - taper_ratio -> OK
  const double K_eta = (eta_tip - eta_kink) / (eta_kink - eta_fuselage);
  std::cout << "Leading edge check ... ";
  if ((K_eta * (1 - taper_ratio_inner)) >= (taper_ratio_inner - taper_ratio)) {
    std::cout << "[OK]\n";
    // Check if trailing edge is none negative
    // s*(eta_kink-eta_fuselage)*tan(sweep_leading_edge_inner) >= 1 - taper_ratio_2 -> ok
    std::cout << "Trailing edge check ... ";
    if ((half_span * (eta_kink - eta_fuselage) * tan_inner_wing_leading_edge_sweep) >= (1 - taper_ratio_inner)) {
      std::cout << "[OK]\n";
    } else {
      throw("Error - trailing edge condition not fulfilled (TE Sweep < 0)");
    }
  } else {
    throw("Error - leading edge condition not fulfilled (LE Sweep inner < LE Sweep outer).");
  }

  // Set dihedral
  double dihedral = get_design_dihedral(1, "low", 0.0, quarter_chord_sweep, mach_ica);

  // Design wing geometry
  std::vector<geom2::Polygon_2> profiles;
  // Load Airfoils
  geom2::Polygon_2 shape = geom2::io::read_airfoil(airfoildatadir + "/" + wing_profile_names[0] + ".dat");
  std::vector<geom2::AirfoilSection> sections;
  for (const std::string& profile_name : wing_profile_names) {
    profiles.push_back(geom2::io::read_airfoil(airfoildatadir + "/" + profile_name + ".dat"));
    sections.push_back(geom2::AirfoilSection(profiles.back()));
  }
  sections[0].set_chord_length(root_chord);
  sections[1].set_chord_length(root_chord);
  sections[2].set_chord_length(kink_chord);
  sections[3].set_chord_length(tip_chord);

  sections[0].origin = geom2::Point_3(0., 0., 0.);
  sections[1].origin = geom2::Point_3(0., 0., -eta_fuselage * half_span);
  sections[2].origin = geom2::Point_3(tan_inner_wing_leading_edge_sweep * (eta_kink - eta_fuselage) * half_span,
                                      tan(dihedral) * (eta_kink - eta_fuselage) * half_span, -eta_kink * half_span);
  sections[3].origin = geom2::Point_3(tan_inner_wing_leading_edge_sweep * (eta_kink - eta_fuselage) * half_span +
                                          tan_outer_wing_leading_edge_sweep * (eta_tip - eta_kink) * half_span,
                                      tan(dihedral) * (eta_tip - eta_kink) * half_span, -eta_tip * half_span);
  geom2::Point_3 reference_position(15., 0., 0.);

  geom2::MultisectionSurface<geom2::AirfoilSection> wing;
  for (auto section : sections) {
    wing.sections.push_back(section);
  }
  wing.origin = reference_position;
  wing.normal = geom2::Direction_3(0., -1., 0.);

  geom2::Mesh wing_mesh = geom2::transform::to_mesh(wing);
  CGAL::IO::write_STL("wing.stl", wing_mesh);

  print_wing_data(installed_aspect_ratio, installed_span, taper_ratio, root_chord, kink_chord, tip_chord,
                  quarter_chord_sweep, atan(tan_inner_wing_leading_edge_sweep),
                  atan(tan_inner_wing_trailing_edge_sweep), atan(tan_outer_wing_leading_edge_sweep),
                  atan(tan_outer_wing_trailing_edge_sweep), installed_wing_area);
}

double WingDimensioner::get_wing_quarter_chord_sweep(modi quarter_chord_sweep_mode,
                                                     const double user_quarter_chord_sweep, const double mach_ica,
                                                     const double wing_loading, const double design_altitude,
                                                     const double max_thickness_to_chord_ratio,
                                                     const double delta_drag_divergence,
                                                     const double technology_factor_drag_divergence) {
  std::map<modi, std::tuple<std::string, double>> quarter_chord_sweep_selection = {
      {0, {"user_defined", user_defined::sweep_angle_at_quarter_chord(user_quarter_chord_sweep)}},
      {1,
       {"drag_divergence", drag_divergence::sweep_angle_at_quarter_chord(
                               mach_ica, design_altitude, delta_drag_divergence, technology_factor_drag_divergence,
                               max_thickness_to_chord_ratio, wing_loading)}}};
  const auto [method, quarter_chord_sweep] = quarter_chord_sweep_selection[quarter_chord_sweep_mode];
  std::cout << "Quarter chord sweep [method]: " << method << "\n";
  std::cout << "Quarter chord sweep  [value]: " << quarter_chord_sweep * geom2::detail::to_degrees << " [deg]\n";
  return quarter_chord_sweep;
}

double WingDimensioner::get_design_relative_kink_position(modi relative_kink_position_mode,
                                                          const double user_defined_relative_kink_position) {
  std::map<modi, std::tuple<std::string, double>> relative_kink_position_selection = {
      {0, {"user_defined", user_defined_relative_kink_position}}};
  const auto [method, relative_kink_position] = relative_kink_position_selection[relative_kink_position_mode];
  std::cout << "Relative kink position [method]: " << method << "\n";
  std::cout << "Relative kink position  [value]: " << relative_kink_position << " [1]\n";
  return relative_kink_position;
}

double WingDimensioner::get_design_aspect_ratio(modi aspect_ratio_mode, const double user_aspect_ratio,
                                                const double quarter_chord_sweep) {
  std::map<modi, std::tuple<std::string, double>> aspect_ratio_selection = {
      {0, {"user_defined", user_defined::aspect_ratio(user_aspect_ratio)}},
      {1, {"by_pitch_up_limit_function", by_pitch_up_limit_function::aspect_ratio(quarter_chord_sweep)}}};
  const auto [method, aspect_ratio] = aspect_ratio_selection[aspect_ratio_mode];
  std::cout << "Wing aspect ratio [method]: " << method << "\n";
  std::cout << "Wing aspect ratio  [value]: " << aspect_ratio << " [1]\n";
  return aspect_ratio;
}

double WingDimensioner::get_design_taper_ratio(modi taper_ratio_mode, const double user_taper_ratio,
                                               const double aspect_ratio, const double quarter_chord_sweep) {
  std::map<modi, std::tuple<std::string, double>> taper_ratio_selection = {
      {0, {"user_defined", user_defined::taper_ratio(user_taper_ratio)}},
      {1, {"howe", howe::taper_ratio(aspect_ratio, quarter_chord_sweep)}}};
  const auto [method, taper_ratio] = taper_ratio_selection[taper_ratio_mode];
  std::cout << "Wing taper ratio [method]: " << method << "\n";
  std::cout << "Wing taper ratio  [value]: " << taper_ratio << " [1]\n";
  return taper_ratio;
}

double WingDimensioner::get_design_wing_area(modi wing_area_mode, const double user_wing_area,
                                             const double maximum_takeoff_mass, const double wing_loading) {
  std::map<modi, std::tuple<std::string, double>> wing_area_selection = {
      {0, {"user_defined", user_defined::wing_area(user_wing_area)}},
      {1, {"by_loading_and_mtom", loading_and_mass::wing_area(maximum_takeoff_mass, wing_loading)}}};
  const auto [method, area] = wing_area_selection[wing_area_mode];
  std::cout << "Wing area [method]: " << method << "\n";
  std::cout << "Wing area  [value]: " << area << " [m^2]\n";
  return area;
}

double WingDimensioner::get_design_dihedral(modi dihedral_mode, const std::string& wing_position,
                                            const double user_dihedral, const double quarter_chord_sweep,
                                            const double mach_design) {
  std::map<modi, std::tuple<std::string, std::tuple<double, double>>> dihedral_selection = {
      {0, {"user_defined", user_defined::dihedral(user_dihedral)}},
      {1, {"raymer", ::raymer::dihedral_limitation(wing_position, quarter_chord_sweep, mach_design)}},
      {2, {"howe", ::howe::dihedral_limitation(wing_position, quarter_chord_sweep)}}};

  const auto [method, limits] = dihedral_selection[dihedral_mode];
  const auto [minimum, maximum] = limits;
  const double value = 0.5 * (minimum + maximum);
  std::cout << "Dihedral [method]: " << method << "\n";
  std::cout << "Dihedral  [value]: " << value * geom2::detail::to_degrees << " [deg]\n";
  return value;
}

std::vector<std::tuple<double, double, std::string>> WingDimensioner::get_wing_profile_and_thickness(
    modi wing_profile_and_thickness,
    const std::vector<std::tuple<double, double, std::string>> user_wing_profile_and_thickess,
    const std::string& wing_profile, const double maximum_thickness_to_chord_ratio,
    const double technology_factor_airfoil, const double mach_design, const double design_altitude,
    const double quarter_chord_sweep, const double mach_drag_divergence, const double wing_loading,
    const double eta_fuselage, std::vector<double> eta_positions) {
  std::map<modi, std::tuple<std::string, std::vector<std::tuple<double, double, std::string>>>>
      wing_profile_and_thickness_selection = {
          {0,
           {"user_defined", user_defined::wing_profile_and_thickness_distribution(user_wing_profile_and_thickess,
                                                                                  eta_fuselage, eta_positions)}},
          {1,
           {"torenbeek_jenkinson",
            torenbeek_jenkinson::wing_profile_and_thickness_distribution(
                wing_profile, maximum_thickness_to_chord_ratio, technology_factor_airfoil, mach_design, design_altitude,
                quarter_chord_sweep, mach_drag_divergence, wing_loading, eta_fuselage, eta_positions)}}};
  const auto [method, wing_profile_and_thickness_vector] =
      wing_profile_and_thickness_selection[wing_profile_and_thickness];
  std::cout << "Wing profile and thickness [method]: " << method << "\n";
  std::cout << "Wing profile and thickness [values]: "
            << "\n";
  for (auto item : wing_profile_and_thickness_vector) {
    const auto [profile_name, t_to_c, eta] = item;
    std::cout << "\tProfile           : " << profile_name << "\n";
    std::cout << "\tThickness to chord: " << t_to_c << "\n";
    std::cout << "\tEta               : " << eta << "\n";
  }
  return wing_profile_and_thickness_vector;
}

void WingDimensioner::print_wing_data(
    const double installed_aspect_ratio, const double installed_wing_span, const double installed_taper_ratio,
    const double installed_root_chord, const double installed_kink_chord, const double installed_tip_chord,
    const double installed_quarter_chord_sweep, const double installed_leading_edge_sweep_inner,
    const double installed_trailing_edge_sweep_inner, const double installed_leading_edge_sweep_outer,
    const double installed_trailing_edge_sweep_outer, const double installed_wing_area) {
  std::cout << "Wing data:\n";
  std::cout << "Aspect Ratio   ... " << std::setprecision(4) << installed_aspect_ratio << "\n";
  std::cout << "Taper ratio    ... " << std::setprecision(4) << installed_taper_ratio << "\n";
  std::cout << "Wing span      ... " << std::setprecision(4) << installed_wing_span << "\n";
  std::cout << "Wing area      ... " << std::setprecision(4) << installed_wing_area << "\n";
  std::cout << "Root chord     ... " << std::setprecision(4) << installed_root_chord << "\n";
  std::cout << "Kink chord     ... " << std::setprecision(4) << installed_kink_chord << "\n";
  std::cout << "Tip chord      ... " << std::setprecision(4) << installed_tip_chord << "\n";
  std::cout << "Sweep 25       ... " << std::setprecision(4)
            << installed_quarter_chord_sweep * geom2::detail::to_degrees << "\n";
  std::cout << "Sweep LE inner ... " << std::setprecision(4)
            << installed_leading_edge_sweep_inner * geom2::detail::to_degrees << "\n";
  std::cout << "Sweep TE inner ... " << std::setprecision(4)
            << installed_trailing_edge_sweep_inner * geom2::detail::to_degrees << "\n";
  std::cout << "Sweep LE outer ... " << std::setprecision(4)
            << installed_leading_edge_sweep_outer * geom2::detail::to_degrees << "\n";
  std::cout << "Sweep TE outer ... " << std::setprecision(4)
            << installed_trailing_edge_sweep_outer * geom2::detail::to_degrees << "\n";
}
}  // namespace cantilever
}  // namespace low
