#include "dihedralMethods.h"

#include <aircraftGeometry2/processing/transform.h>
#include <unitConversion/constants.h>

#include <map>
#include <iostream>
#include <numbers>
#include <tuple>
#include <functional>


constexpr double to_degree = 180.0 / std::numbers::pi;
constexpr double to_radian = std::numbers::pi / 180.0;

std::tuple<double, double> user_defined::dihedral(const double user_dihedral) { return {user_dihedral, user_dihedral}; }

double by_wing_position_and_quarter_chord_sweep::dihedral(const std::string& wing_position,
                                                          const double quarter_chord_sweep, const double mach_design) {
  const auto [minimum_dihedral, maximum_dihedral] =
      raymer::dihedral_limitation(wing_position, quarter_chord_sweep, mach_design);
  return 0.5 * (minimum_dihedral + maximum_dihedral);
}

std::tuple<double, double> raymer::dihedral_limitation(const std::string& wing_position,
                                                       const double quarter_chord_sweep, const double mach_design) {
  std::map<std::string, std::tuple<double, double>> unswept = {{"low", {5. * geom2::detail::to_radians, 7. * geom2::detail::to_radians}},
                                                               {"mid", {2. * geom2::detail::to_radians, 4. * geom2::detail::to_radians}},
                                                               {"high", {0. * geom2::detail::to_radians, 2. * geom2::detail::to_radians}}};

  std::map<std::string, std::tuple<double, double>> swept_subsonic = {{"low", {3. * geom2::detail::to_radians, 7. * geom2::detail::to_radians}},
                                                                      {"mid", {-2. * geom2::detail::to_radians, 2. * geom2::detail::to_radians}},
                                                                      {"high", {-5. * geom2::detail::to_radians, -2. * geom2::detail::to_radians}}};

  std::map<std::string, std::tuple<double, double>> swept_supersonic = {{"low", {0. * geom2::detail::to_radians, 5. * geom2::detail::to_radians}},
                                                                        {"mid", {-5. * geom2::detail::to_radians, 0. * geom2::detail::to_radians}},
                                                                        {"high", {-5. * geom2::detail::to_radians, 0. * geom2::detail::to_radians}}};

  if (mach_design < 1.) {
    if (fabs(quarter_chord_sweep) < ACCURACY_MEDIUM) {
      return unswept[wing_position];
    } else {
      return swept_subsonic[wing_position];
    }
  } else {
    return swept_supersonic[wing_position];
  }
}

std::tuple<double, double> howe::dihedral_limitation(const std::string& wing_position,
                                                     const double quarter_chord_sweep) {
  std::map<std::string, std::tuple<double, double>> unswept = {{"low", {3. * geom2::detail::to_radians, 5. * geom2::detail::to_radians}},
                                                               {"mid", {0. * geom2::detail::to_radians, 0. * geom2::detail::to_radians}},
                                                               {"high", {0. * geom2::detail::to_radians, 0. * geom2::detail::to_radians}}};
  std::map<std::string, std::tuple<double, double>> swept = {
      {"low", {0. * to_radian, std::numeric_limits<double>::max()}},
      {"mid", {std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max()}},
      {"high", {std::numeric_limits<double>::lowest(), 0. * geom2::detail::to_radians}}};
  if (fabs(quarter_chord_sweep) < ACCURACY_MEDIUM) {
    return unswept[wing_position];
  } else {
    return swept[wing_position];
  }
}

double get_design_dihedral(modes dihedral_mode, const std::string& wing_mounting, const double user_dihedral,
                                 const double quarter_chord_sweep, const double mach_design) {
  std::map<modes, std::tuple<std::string, std::function<std::tuple<double, double>(void)>>> dihedral_selection = {
      {"mode_0", {"user_defined", [=]() { return user_defined::dihedral(user_dihedral); }}},
      {"mode_1",
       {"raymer", [=]() { return ::raymer::dihedral_limitation(wing_mounting, quarter_chord_sweep, mach_design); }}},
      {"mode_2", {"howe", [=]() { return ::howe::dihedral_limitation(wing_mounting, quarter_chord_sweep); }}}};

  const auto [method, limits_method] = dihedral_selection[dihedral_mode];
  std::tuple<double, double> limits = limits_method();
  const auto [minimum, maximum] = limits;
  const double value = 0.5 * (minimum + maximum);
  std::cout << "Dihedral [method]: " << method << "\n";
  std::cout << "Dihedral  [value]: " << value * geom2::detail::to_degrees << " [deg]\n";
  return value;
}