/**
 * \file lowWingDesignTaw.h
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief
 * \version 0.1
 * \date 2023-12-05
 *
 * @copyright Copyright (c) 2023
 *
 */
#ifndef LOW_WINGDESIGN_TAW_H_
#define LOW_WINGDESIGN_TAW_H_

#include <airfoils/airfoils.h>
#include <moduleBasics/module.h>
#include <moduleBasics/strategySelector.h>
#include <moduleBasics/report.h>

#include "src/taw/tawWingDesignConfig.h"
#include "src/taw/tawWingDesignData.h"

using modes = std::string;
namespace taw {
namespace cantilever {
class Wing : public Strategy {
 public:
  Wing(const std::shared_ptr<RuntimeIO>& rtIO);
  ~Wing() = default;

  /**
   * \brief Initialize routine
   * 
   */
  void initialize();
  
  /**
   * \brief Run routine
   * 
   */
  void run();
  
  /**
   * \brief Update routine
   * 
   */
  void update();

  /**
   * \brief Report routine
   * 
   */
  void report();

  /**
   * \brief Save routine
   * 
   */
  void save();

  /**
   * \brief Standard cantilever design method [kinked or unkinked dependency on wing position]
   * 
   */
  void standard_design();

  /**
   * \brief mass method -> flops
   * 
   */
  void flops_mass();

  /**
   * \brief mass method -> chiozzotto weight estimation relationship
   * 
   */
  void chiozzotto_wer_mass();

  /**
   * \brief Chiozzotto utility check function
   * 
   * \param min_max 
   * \param value_to_check 
   * \return true 
   * \return false 
   */
  bool chiozzotto_wer_check_limits(std::tuple<double, double>& min_max, double value_to_check);

  /**
   * \brief Check if designed wing is a kinked or not
   * 
   * \return true if kinked 
   * \return false if not kinked
   */
  bool is_wing_kinked_check();

  /**
   * \brief Get the max thickness to chord ratio value 
   * 
   * \return double 
   */
  double get_max_thickness_to_chord_ratio();

  /**
   * \brief Get the maximum takeoff mass value
   * 
   * \return double 
   */
  double get_maximum_takeoff_mass();

  /**
   * \brief Get the control device area in fraction of wing area value
   * 
   * \return double 
   */
  double get_control_device_area_in_fraction_of_wing_area();

  /**
   * \brief Calculates the wing position based on the fuselage length
   * 
   * \return geom2::Point_3 
   */
  geom2::Point_3 calculate_wing_position();

  /**
   * \brief Calculation method for kinked wing geometry
   * 
   * \param taper_ratio 
   * \param aspect_ratio 
   * \param quarter_chord_sweep 
   * \param dihedral 
   * \param wing_area 
   * \param span 
   * \return geom2::MultisectionSurface<geom2::AirfoilSection> object 
   */
  geom2::MultisectionSurface<geom2::AirfoilSection> calculate_kinked_wing_geometry(
      const double taper_ratio, const double aspect_ratio, const double quarter_chord_sweep, const double dihedral,
      const double wing_area, const double span);

  /**
   * \brief Get the rescaled airfoil geometry object
   * 
   * \param wing_profile_and_thickness tuple
   * \return geom2::AirfoilSection 
   */
  geom2::AirfoilSection get_rescaled_airfoil_geometry(std::tuple<double, double, std::string> wing_profile_and_thickness);

  /**
   * \brief Calculation method for unkinked wing geometry
   * 
   * \param taper_ratio 
   * \param aspect_ratio 
   * \param quarter_chord_sweep 
   * \param dihedral 
   * \param wing_area 
   * \param span 
   * \return geom2::MultisectionSurface<geom2::AirfoilSection> object 
   */
  geom2::MultisectionSurface<geom2::AirfoilSection> calculate_unkinked_wing_geometry(
      const double taper_ratio, const double aspect_ratio, const double quarter_chord_sweep, const double dihedral,
      const double wing_area, const double span);

  /**
   * \brief Set the html body 
   * 
   */
  void set_html_body(void);

  /**
   * \brief Generate plot for wing planform
   * 
   * \return fs::path relative path between reportpage and wing planform plot
   */
  fs::path plot_wing_planform();

  /**
   * \brief Generate plots for thickness and twist 
   * 
   * \return fs::path relative path between reportpage and thickness and twist distribution plot
   */
  fs::path plot_thickness_and_twist_distribution();

  /**
   * \brief Generate airfoil plots
   * 
   * \return std::vector<fs::path> vector of relative paths between reportpage and thickness and twist distribution plot
   */
  std::vector<fs::path> plot_airfoils();

  std::string selected_design_mode;
  std::string selected_mass_mode;
  std::map<std::string, std::function<void(void)>> design_mode_runner;
  std::map<std::string, std::function<void(void)>> mass_mode_runner;

  std::shared_ptr<Airfoils> airfoils_library;
  const std::shared_ptr<RuntimeIO>& rtIO;
  const std::shared_ptr<taw::cantilever::WingDesignData> data;
  const std::shared_ptr<taw::cantilever::WingDesignConfig> config;
  Report reporter;
};

}  // namespace cantilever
}  // namespace taw

#endif  // LOW_WINGDESIGN_TAW_H_